/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 2001-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

bsearchbody.c is for searching e-mail bodies for site domain names and
IP addresses, which are compared against the names and addresses in a
database file.

This program uses mmap(2) to map the database file into the Unix VM
system. The names and addresses from the e-mail body are compared
against the database file using a binary search. The database file
name is a required command line argument.

The database is a standard Unix text file, one domain name or IP
address per line, in lexical order, constructed with "sort -u infile >
outfile", or equivalent.  The database is a standard Unix text file,
one IP address, or site domain name, per line, in lexical order,
constructed with "sort -u infile > outfile", or equivalent. An IP
address range can be represented as a Class A, B, or C range.  For
example, the IP address "123.210." in the database file would match
"123.210.1.0" in a "Received: " e-mail header record.

The database mechanism is conservative with machine resources,
requiring about 12.5 micro-seconds of machine time to lookup a word in
the Unix system dictionary, (2.5 MB, quarter of a million words,
single 466 MHz., Pentium, lightly loaded, Linux 2.2, time(1) command
to lookup every word in the dictionary, divided by the number of
words.) Conceptually, the database mechanism is implemented similar to
the the technique used in the look(1) command, but requires exact
matches, as opposed to partial key matches.

The input e-mail file name(s) may be supplied as additional optional
command line arguments, or redirected to the program via stdin for
compatibility with procmail(1), and other e-mail scripting agents.

A suitable procmail(1) recipe example might be:

    :0 wfh
    * ? bsearchbody blacklist.db
    | formail -A "X-Notice: Message in blacklist.db database"

A suitable way of making a database is by something like:

    bsearchbody -c -P blacklist.db < spam.1 spam.2 ...> temp

where the spam.(n) files are spam messages. The temp file will have to
be hand edited for appropriate spam addresses, and have the current
blacklist.db database added to it, and then sorted in lexical order as
per sort(1) to make a new blacklist.db standard Unix text file
database.

The program contains less than 1000 lines of declarations and
statements, all of which are documented with in line comments.

The program has been compiled and tested on SunOS, Solaris, and Linux,
and may work on other brands of Unix.

The program returns 0 if no error and a match was found in the
database file for the site domain names or IP addresses, 1 if no error
and no match found; else returns a unique error code greater than 1
representing the error encountered-which will, also, print an error
diagnostic to stderr.

The -r option is useful for controlling the return value under error
conditions-for example, the program return can be preempted if the
database file can not be opened, (or read,) with a return value of
match, or no match, depending on environmental requirements.

John Conover
January 16, 2005
john@email.johncon.com

$Revision: 0.0 $
$Date: 2005/01/17 19:36:00 $
$Id: bsearchbody.c,v 0.0 2005/01/17 19:36:00 john Exp $
$Log: bsearchbody.c,v $
Revision 0.0  2005/01/17 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* for getopt() */
#include <ctype.h> /* for tolower(), isspace() */
#include <sys/types.h> /* for open() */
#include <sys/stat.h> /* for open() */
#include <fcntl.h> /* for open() */
#include <unistd.h> /* for mmap() */
#include <sys/mman.h> /* for mmap() */

static char rcsid[] = "$Id: bsearchbody.c,v 0.0 2005/01/17 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 2001-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define MAX_ALPHABET_SIZE 256 /* maximum size of alphabet, characters numbered 0 .. 255 */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Compare domain names and IP addresses found in the body of an e-mail\n",
    "against a database file:\n",
    "Usage: bsearchbody [-c] [-P] [-r m|n] [-v] dbfilename [filename(s)]\n",
    "    -c, print all addresses, not just the first\n",
    "    -P, print the addresses not in the database\n",
    "    -r m|n, on file error, exit return = match for m, no match for n\n",
    "    -v, print the program's version information\n",
    "    dbfilename, database file name\n",
    "    filename(s), e-mail file name(s), (defaults to stdin)\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error, match found\n",
    "No error, match not found\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error in status of database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error in database file\n"
};

#define MATCH 0 /* error values, one for each index in the error message array */
#define NOMATCH 1
#define EARGS 2
#define EOPEN 3
#define ECLOSE 4
#define ESTAT 5
#define EMAP 6
#define EUNMAP 7
#define EDB 8

/*

All possible top level domains of site address URLs, and the possible
last digits in a dotted quad IP address; in lexical order; used for
fast binary lookup to see if a word is a URL or IP address; the array
is made with the integers 0-256, and the lowercase TLDs, and in the
lexical order of the strcmp() function or sort(1) command.

The current TLDs are listed at:

    http://www.iana.org/domain-names.htm
    http://www.iana.org/cctld/cctld-whois.htm

Note that the scheme will not work if TLDs ever have digits; an index,
(implicit address,) between 0 and 255 is used to detect the last quad
of an IP address.

*/

#ifdef __STDC__

static const char *tld[] =

#else

static char *tld[] =

#endif

{
    "0",
    "1",
    "10",
    "100",
    "101",
    "102",
    "103",
    "104",
    "105",
    "106",
    "107",
    "108",
    "109",
    "11",
    "110",
    "111",
    "112",
    "113",
    "114",
    "115",
    "116",
    "117",
    "118",
    "119",
    "12",
    "120",
    "121",
    "122",
    "123",
    "124",
    "125",
    "126",
    "127",
    "128",
    "129",
    "13",
    "130",
    "131",
    "132",
    "133",
    "134",
    "135",
    "136",
    "137",
    "138",
    "139",
    "14",
    "140",
    "141",
    "142",
    "143",
    "144",
    "145",
    "146",
    "147",
    "148",
    "149",
    "15",
    "150",
    "151",
    "152",
    "153",
    "154",
    "155",
    "156",
    "157",
    "158",
    "159",
    "16",
    "160",
    "161",
    "162",
    "163",
    "164",
    "165",
    "166",
    "167",
    "168",
    "169",
    "17",
    "170",
    "171",
    "172",
    "173",
    "174",
    "175",
    "176",
    "177",
    "178",
    "179",
    "18",
    "180",
    "181",
    "182",
    "183",
    "184",
    "185",
    "186",
    "187",
    "188",
    "189",
    "19",
    "190",
    "191",
    "192",
    "193",
    "194",
    "195",
    "196",
    "197",
    "198",
    "199",
    "2",
    "20",
    "200",
    "201",
    "202",
    "203",
    "204",
    "205",
    "206",
    "207",
    "208",
    "209",
    "21",
    "210",
    "211",
    "212",
    "213",
    "214",
    "215",
    "216",
    "217",
    "218",
    "219",
    "22",
    "220",
    "221",
    "222",
    "223",
    "224",
    "225",
    "226",
    "227",
    "228",
    "229",
    "23",
    "230",
    "231",
    "232",
    "233",
    "234",
    "235",
    "236",
    "237",
    "238",
    "239",
    "24",
    "240",
    "241",
    "242",
    "243",
    "244",
    "245",
    "246",
    "247",
    "248",
    "249",
    "25",
    "250",
    "251",
    "252",
    "253",
    "254",
    "255",
    "26",
    "27",
    "28",
    "29",
    "3",
    "30",
    "31",
    "32",
    "33",
    "34",
    "35",
    "36",
    "37",
    "38",
    "39",
    "4",
    "40",
    "41",
    "42",
    "43",
    "44",
    "45",
    "46",
    "47",
    "48",
    "49",
    "5",
    "50",
    "51",
    "52",
    "53",
    "54",
    "55",
    "56",
    "57",
    "58",
    "59",
    "6",
    "60",
    "61",
    "62",
    "63",
    "64",
    "65",
    "66",
    "67",
    "68",
    "69",
    "7",
    "70",
    "71",
    "72",
    "73",
    "74",
    "75",
    "76",
    "77",
    "78",
    "79",
    "8",
    "80",
    "81",
    "82",
    "83",
    "84",
    "85",
    "86",
    "87",
    "88",
    "89",
    "9",
    "90",
    "91",
    "92",
    "93",
    "94",
    "95",
    "96",
    "97",
    "98",
    "99",
    "ac",
    "ad",
    "ae",
    "aero",
    "af",
    "ag",
    "ai",
    "al",
    "am",
    "an",
    "ao",
    "aq",
    "ar",
    "as",
    "at",
    "au",
    "aw",
    "az",
    "ba",
    "bb",
    "bd",
    "be",
    "bf",
    "bg",
    "bh",
    "bi",
    "biz",
    "bj",
    "bm",
    "bn",
    "bo",
    "br",
    "bs",
    "bt",
    "bv",
    "bw",
    "by",
    "bz",
    "ca",
    "cc",
    "cd",
    "cf",
    "cg",
    "ch",
    "ci",
    "ck",
    "cl",
    "cm",
    "cn",
    "co",
    "com",
    "coop",
    "cr",
    "cu",
    "cv",
    "cx",
    "cy",
    "cz",
    "de",
    "dj",
    "dk",
    "dm",
    "do",
    "dz",
    "ec",
    "edu",
    "ee",
    "eg",
    "eh",
    "er",
    "es",
    "et",
    "fi",
    "fj",
    "fk",
    "fm",
    "fo",
    "fr",
    "ga",
    "gd",
    "ge",
    "gf",
    "gg",
    "gh",
    "gi",
    "gl",
    "gm",
    "gn",
    "gov",
    "gp",
    "gq",
    "gr",
    "gs",
    "gt",
    "gu",
    "gw",
    "gy",
    "hk",
    "hm",
    "hn",
    "hr",
    "ht",
    "hu",
    "id",
    "ie",
    "il",
    "im",
    "in",
    "info",
    "int",
    "io",
    "iq",
    "ir",
    "is",
    "it",
    "je",
    "jm",
    "jo",
    "jp",
    "ke",
    "kg",
    "kh",
    "ki",
    "km",
    "kn",
    "kp",
    "kr",
    "kw",
    "ky",
    "kz",
    "la",
    "lb",
    "lc",
    "li",
    "lk",
    "lr",
    "ls",
    "lt",
    "lu",
    "lv",
    "ly",
    "ma",
    "mc",
    "md",
    "mg",
    "mh",
    "mil",
    "mk",
    "ml",
    "mm",
    "mn",
    "mo",
    "mp",
    "mq",
    "mr",
    "ms",
    "mt",
    "mu",
    "museum",
    "mv",
    "mw",
    "mx",
    "my",
    "mz",
    "na",
    "name",
    "nc",
    "ne",
    "net",
    "nf",
    "ng",
    "ni",
    "nl",
    "no",
    "np",
    "nr",
    "nu",
    "nz",
    "om",
    "org",
    "pa",
    "pe",
    "pf",
    "pg",
    "ph",
    "pk",
    "pl",
    "pm",
    "pn",
    "pr",
    "pro",
    "ps",
    "pt",
    "pw",
    "py",
    "qa",
    "re",
    "ro",
    "ru",
    "rw",
    "sa",
    "sb",
    "sc",
    "sd",
    "se",
    "sg",
    "sh",
    "si",
    "sj",
    "sk",
    "sl",
    "sm",
    "sn",
    "so",
    "sr",
    "st",
    "sv",
    "sy",
    "sz",
    "tc",
    "td",
    "tf",
    "tg",
    "th",
    "tj",
    "tk",
    "tm",
    "tn",
    "to",
    "tp",
    "tr",
    "tt",
    "tv",
    "tw",
    "tz",
    "ua",
    "ug",
    "uk",
    "um",
    "us",
    "uy",
    "uz",
    "va",
    "vc",
    "ve",
    "vg",
    "vi",
    "vn",
    "vu",
    "wf",
    "ws",
    "ye",
    "yt",
    "yu",
    "za",
    "zm",
    "zw"
};

static size_t tld_size = (sizeof (tld) / sizeof (char *)) - (size_t) 1; /* element in the tld array */

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap);
static int str_search (const char *array[], int elements, char *string);
static int quad_verify (char *string);

#else

static void print_message (); /* print any error messages */
static char *binarysearch ();
static int str_search ();
static int quad_verify ();

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char word[BUFLEN], /* word space */
         temp[BUFLEN], /* temporary buffer space */
         *cptr, /* temporary character reference */
         *domain_ref, /* character reference into a possible domain name or IP address */
         *next_domain_ref, /* next character reference into a possible domain name or IP address */
         *word_end = word + BUFLEN, /* buffer overlow detection */
         *begin, /* begining of memory mapped file */
         *end, /* end of memory mapped file */
         *last_period, /* reference to last period in a word, i.e., reference to the TLD of a site address, or fourth quad in an IP address */
         *middle_period, /* reference to the third quad in an IP address */
         *first_period, /* reference to the second quad in an IP address */
         first_char; /* first character in a word */

    unsigned char translit[MAX_ALPHABET_SIZE]; /* array of characters for transliteration */

    int retval = NOMATCH, /* return value, assume no error */
        fd = -1, /* input file descriptor */
        i, /* general counter */
        c, /* command line switch */
        match_count = 0, /* count of the number of matches-the number of addresses found in the database */
        print_flag = 0, /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match */
        continue_flag = 0, /* continue to print all matching addresses, not just the first, 0 = no, 1 = yes */
        continue_reading = 1, /* containue to read the file, even if a match has been found flag, 0 = no, 1 = yes */
        error_return_change = -1, /* change exit return on error, -1 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
        fc, /* input character */
        tc, /* input character, transliterated */
        last_c = -1, /* last input character */
        eoheader = 0, /* end of header flag, 0 = record from header, 1 = record from body */
        local_match, /* local match flag, 0 = no local match, 1 = local match */
        tld_element; /* implicit address/index of a tld */

    FILE *infile = stdin; /* reference to input file; defaults to stdin */

    size_t filesize; /* size of file */

    struct stat fileinfo; /* structure for file status */

    caddr_t mapbegin; /* reference to beginning of mmap'ed file */

    while ((c = getopt (argc, argv, "cPr:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'c': /* request for continue to print all matching addresses, not just the first? */

                continue_flag = 1; /* set the continue to print all matching addresses, not just the first, 0 = no, 1 = yes */
                break;

            case 'P': /* request for verbose print? */

                print_flag = 1; /* set the verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match */
                break;

            case 'r': /* request for change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH? */

                switch (*optarg) /* first character of argument object, only */
                {

                    case 'm':

                        error_return_change = MATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    case 'n':

                        error_return_change = NOMATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    default:

                        optind = argc; /* force argument error */
                        retval = EARGS; /* assume not enough arguments */
                        break;
                }

                break;

            case 'v': /* request for program's version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    translit[0] = (unsigned char) ' '; /* the null character is reserved as an end of search sentinel */

    for (i = 1; i < MAX_ALPHABET_SIZE; i++) /* for each remaining character in the transliteration array */
    {

        if (isspace (i)) /* character whitespace? */
        {
            translit[i] = (unsigned char) ' '; /* yes, make sure the character is a space */
        }

        else
        {
            translit[i] = (unsigned char) tolower (i); /* no, convert the character to lowercase */
        }

    }

    /*

    Characters that are considered word delimiters, and are replaced
    with white space; these characters are not valid in domain names
    and IP addresses.

    */

    translit[(int) '!'] = (unsigned char) ' ';
    translit[(int) '"'] = (unsigned char) ' ';
    translit[(int) '#'] = (unsigned char) ' ';
    translit[(int) '$'] = (unsigned char) ' ';
    translit[(int) '%'] = (unsigned char) ' ';
    translit[(int) '&'] = (unsigned char) ' ';
    translit[(int) '\''] = (unsigned char) ' ';
    translit[(int) '('] = (unsigned char) ' ';
    translit[(int) ')'] = (unsigned char) ' ';
    translit[(int) '*'] = (unsigned char) ' ';
    translit[(int) '+'] = (unsigned char) ' ';
    translit[(int) ','] = (unsigned char) ' ';
    translit[(int) '/'] = (unsigned char) ' ';
    translit[(int) ':'] = (unsigned char) ' ';
    translit[(int) ';'] = (unsigned char) ' ';
    translit[(int) '<'] = (unsigned char) ' ';
    translit[(int) '>'] = (unsigned char) ' ';
    translit[(int) '?'] = (unsigned char) ' ';
    translit[(int) '@'] = (unsigned char) ' ';
    translit[(int) '['] = (unsigned char) ' ';
    translit[(int) '\\'] = (unsigned char) ' ';
    translit[(int) ']'] = (unsigned char) ' ';
    translit[(int) '^'] = (unsigned char) ' ';
    translit[(int) '`'] = (unsigned char) ' ';
    translit[(int) '{'] = (unsigned char) ' ';
    translit[(int) '|'] = (unsigned char) ' ';
    translit[(int) '}'] = (unsigned char) ' ';
    translit[(int) '~'] = (unsigned char) ' ';
    translit[(int) '\t'] = (unsigned char) ' ';

    cptr = &word[0]; /* character reference to first character of the word space */
    *cptr = '\0'; /* terminate the word */

    if (retval == NOMATCH) /* any errors? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (argc > optind) /* enough arguments? */
        {
            retval = ESTAT; /* assume error in status of file */

            if (stat (argv[optind], &fileinfo) == 0) /* get the status of the file */
            {
                retval = EDB;  /* assume error in database file */

                if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of file */
                {
                    retval = EOPEN; /* assume error opening file */

                    if ((fd = open (argv[optind], O_RDONLY)) != -1) /* open the file */
                    {
                        retval = EMAP; /* assume error mapping file */

                        if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, fd, (off_t) 0)) != (caddr_t) 0) /* map the file */
                        {
                            retval = EDB;  /* assume error in database file */

                            if (mapbegin[filesize - 1] == '\n') /* verify that the last record of the memory mapped file has a terminating '\n' character as required for binarysearch() */
                            {
                                retval = NOMATCH; /* yes, return value, assume no error */
                                optind ++; /* file argument handled */
                                begin = (char *) mapbegin; /* reference the begining of memory mapped file */
                                end = &mapbegin[filesize - 1]; /* reference the end of memory mapped file */

                                do /* while there are input file(s), if any, listed on the command line */
                                {
                                    retval = EOPEN; /* assume error opening file */

                                    if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* open the input file; if none specified, default to stdin */
                                    {
                                        retval = NOMATCH; /* yes, return value, assume no error */
                                        eoheader = 0; /* reset the end of header flag, 0 = record from header, 1 = record from body */
                                        continue_reading = 1; /* containue to read the file, even if a match has been found flag, 0 = no, 1 = yes */

                                        while (((fc = fgetc (infile)) != EOF) && (continue_reading != 0)) /* read the characters from the input file until EOF */
                                        {

                                            if (eoheader == 1) /* passed header? end of header flag, 0 = record from header, 1 = record from body */
                                            {
                                                tc = translit[fc]; /* input character, transliterated */

                                                switch (tc) /* which transliterated character? */
                                                {

                                                    /*

                                                    Characters that are not word delimibers; a list of all
                                                    the characters which may appear in a domain name or IP
                                                    address.

                                                    */

                                                    case '-':
                                                    case '.':
                                                    case '0':
                                                    case '1':
                                                    case '2':
                                                    case '3':
                                                    case '4':
                                                    case '5':
                                                    case '6':
                                                    case '7':
                                                    case '8':
                                                    case '9':
                                                    case '=':
                                                    case 'a':
                                                    case 'b':
                                                    case 'c':
                                                    case 'd':
                                                    case 'e':
                                                    case 'f':
                                                    case 'g':
                                                    case 'h':
                                                    case 'i':
                                                    case 'j':
                                                    case 'k':
                                                    case 'l':
                                                    case 'm':
                                                    case 'n':
                                                    case 'o':
                                                    case 'p':
                                                    case 'q':
                                                    case 'r':
                                                    case 's':
                                                    case 't':
                                                    case 'u':
                                                    case 'v':
                                                    case 'w':
                                                    case 'x':
                                                    case 'y':
                                                    case 'z':
                                                    case '_':

                                                        if (cptr >= word_end) /* buffer overflow? */
                                                        {
                                                            cptr = &word[0]; /* yes, trash the word; a domain name or IP address can not be greater than BUFLEN = BUFSIZ, this word is something else */
                                                            *cptr = '\0'; /* terminate the word */
                                                        }

                                                        *cptr = (char) tc; /* store the transliterated character in the word space */
                                                        cptr ++; /* next character in word space */
                                                        break;

                                                    case ' ':
                                                    default :

                                                        /*

                                                        word delimiter from the translit array; this is
                                                        the end of the word in the word space

                                                        */

                                                        if (last_c == '=' && fc == '\n') /* a "=\n" character sequence is a line continuation */
                                                        {

                                                            if (cptr > word) /* not at beginning of word? */
                                                            {
                                                                cptr --; /* step backwards over the '=' character, and continue */
                                                            }

                                                            *cptr = '\0'; /* terminate the word */
                                                        }

                                                        else
                                                        {
                                                            *cptr = '\0'; /* terminate the word */

                                                            while (*(cptr - 1) == '.' && cptr > &word[0]) /* remove trailing '.' characters; these were probably a period at the end of a sentence */
                                                            {
                                                                cptr --; /* previous character at the end of the word was a '.' */
                                                                *cptr = '\0'; /* remove it, and step backwards */
                                                            }

                                                            if (cptr - word >= 3) /* last three characters of word "=20"? */
                                                            {

                                                                if ((*(cptr - 1) == '0') && (*(cptr - 2) == '2') && (*(cptr - 3) == '='))
                                                                {
                                                                    cptr = cptr - 3; /* yes, step backwards over the "=20" */
                                                                    *cptr = '\0';
                                                                }

                                                            }

                                                            if ((cptr - &word[0]) > 3) /* the smallest site domain name would be somthing like "a.uk", which is 4 characters */
                                                            {

                                                                if (strchr (&word[0], '=') == (char *) 0) /* a sites domain site or IP address containing a '=' character is neither */
                                                                {

                                                                    if (strstr (&word[0], "..") == (char *) 0) /* null fields in a site domain name or IP address means its neither */
                                                                    {
                                                                        first_char = word[0]; /* save the first character in a word */

                                                                        if ((first_char >= 'a' && first_char <= 'z') || (first_char >= '0' && first_char <= '9')) /* first character of a domain name must be [a-z] or [0-9] */
                                                                        {

                                                                            if ((last_period = strrchr (word, '.')) != (char *) 0) /* word contain a period? (site domain names and IP addresses all must contain at least one period) */
                                                                            {

                                                                                /*

                                                                                Found a valid word; the beginning of the
                                                                                word is starts in the word[] space, cptr
                                                                                points at the '\0' word termination
                                                                                character, and last_period points at the
                                                                                last period character.

                                                                                It could be a word with an international
                                                                                domain, like somedomain.co.uk, or a US
                                                                                domain, like somedomin.com, or an IP
                                                                                address, like 123.123.123.123?

                                                                                cptr - last_period must be:

                                                                                    2 for something.1
                                                                                    3 for something.uk, or somthing.12
                                                                                    4 for something.com, or something.123
                                                                                    5 for something.name
                                                                                    7 for something.museum

                                                                                See:
                                                                                    http://www.iana.org/domain-names.htm
                                                                                    http://www.iana.org/cctld/cctld-whois.htm

                                                                                for particulars.

                                                                                */

                                                                                switch (cptr - last_period)
                                                                                {

                                                                                    case 2:
                                                                                    case 3:
                                                                                    case 4:
                                                                                    case 5:
                                                                                    case 7:

                                                                                        if ((tld_element = str_search (tld, (int) tld_size, last_period + 1)) > 0) /* search the tld array for the tld or last dotted quad element */
                                                                                        {

                                                                                            if (tld_element > 255) /* dotted quad IP address or TLD? */
                                                                                            {
                                                                                                local_match = 0; /* TLD, reset the local match flag, 0 = no local match, 1 = local match */
                                                                                                domain_ref = word; /* reference the first possible domain name or IP address */

                                                                                                while (((next_domain_ref = strchr (domain_ref, '.')) != (char *) 0) && (continue_reading != 0)) /* reference the next '.' in the FQDN */
                                                                                                {
                                                                                                    next_domain_ref ++; /* step over the '.' character */

                                                                                                    if (binarysearch (begin, end, domain_ref, strlen (domain_ref), begin, end) != (char *) 0) /* search for the domain name */
                                                                                                    {

                                                                                                        if (print_flag == 0) /* found in the database, and verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, reset? */
                                                                                                        {
                                                                                                            (void) printf ("%s\n", domain_ref); /* print the match */
                                                                                                        }

                                                                                                        if (continue_flag == 0) /* continue to print all matching addresses, not just the first, 0 = no, 1 = yes, set? */
                                                                                                        {
                                                                                                            continue_reading = 0; /* yes, do not containue to read the file, even if a match has been found flag, 0 = no, 1 = yes */
                                                                                                        }

                                                                                                        retval = MATCH; /* return value, assume match found in database */
                                                                                                        local_match = 1; /* set the local match flag, 0 = no local match, 1 = local match */
                                                                                                        match_count ++; /* increment the count of the number of matches-the number e-mail addresses found in the database */
                                                                                                    }

                                                                                                    domain_ref = next_domain_ref; /* next sub domain in the FQDN */
                                                                                                }

                                                                                                if (local_match == 0) /* set the local match flag, 0 = no local match, 1 = local match */
                                                                                                {

                                                                                                    if (print_flag != 0) /* match not found in the database, and verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, reset? */
                                                                                                    {
                                                                                                        (void) printf ("%s\n", word); /* print the non-match */
                                                                                                    }

                                                                                                }

                                                                                            }

                                                                                            else
                                                                                            {

                                                                                                if (last_period > word) /* period not first character of quad? */
                                                                                                {
                                                                                                    *last_period = '\0'; /* temporarily replace period with a null string terminator */

                                                                                                    if ((middle_period = strrchr (word, '.')) != (char *) 0) /* reference the third quad in an IP address */
                                                                                                    {

                                                                                                        if (middle_period > word) /* period not first character of quad? */
                                                                                                        {
                                                                                                            *middle_period = '\0'; /* temporarily replace period with a null string terminator */

                                                                                                            if ((first_period = strrchr (word, '.')) != (char *) 0) /* reference to the second quad in an IP address */
                                                                                                            {

                                                                                                                if (first_period > word) /* period not first character of quad? */
                                                                                                                {
                                                                                                                    *first_period = '\0'; /* temporarily replace period with a null string terminator */

                                                                                                                    if (strrchr (word, '.') == (char *) 0) /* three periods found in dotted quad IP address, no more permitted */
                                                                                                                    {

                                                                                                                        if (quad_verify (middle_period + 1) > 0) /* last quad has already been verified, verify the next to the last */
                                                                                                                        {

                                                                                                                            if (quad_verify (first_period + 1) > 0) /* verify the second quad */
                                                                                                                            {

                                                                                                                                if (quad_verify (word) > 0) /* verify the first quad */
                                                                                                                                {
                                                                                                                                    temp[0] = '\0'; /* null the temporary buffer space */
                                                                                                                                    (void) strcat (temp, word); /* re-assemble a copy of the dotted quad IP address */
                                                                                                                                    (void) strcat (temp, "."); /* include the dot */
                                                                                                                                    local_match = 1; /* set the local match flag, 0 = no local match, 1 = local match */

                                                                                                                                    if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                                                                                    {
                                                                                                                                        (void) strcat (temp, first_period + 1); /* re-assemble a copy of the dotted quad IP address */
                                                                                                                                        (void) strcat (temp, "."); /* include the dot */

                                                                                                                                        if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                                                                                        {
                                                                                                                                            (void) strcat (temp, middle_period + 1); /* re-assemble a copy of the dotted quad IP address */
                                                                                                                                            (void) strcat (temp, "."); /* include the dot */

                                                                                                                                            if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                                                                                            {
                                                                                                                                                (void) strcat (temp, last_period + 1); /* re-assemble a copy of the dotted quad IP address */

                                                                                                                                                if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                                                                                                {
                                                                                                                                                    local_match = 0; /* no match, reset the local match flag, 0 = no local match, 1 = local match */

                                                                                                                                                    if (print_flag != 0) /* match not found in the database, and verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, reset? */
                                                                                                                                                    {
                                                                                                                                                        *first_period = '.'; /* replace the first period in the dotted quad IP address */
                                                                                                                                                        *middle_period = '.'; /* replace the second period in the dotted quad IP address */
                                                                                                                                                        *last_period = '.'; /* replace the third period in the dotted quad IP address */
                                                                                                                                                        (void) printf ("%s\n", word); /* print the non-match */
                                                                                                                                                    }

                                                                                                                                                }

                                                                                                                                            }

                                                                                                                                        }

                                                                                                                                    }

                                                                                                                                    if (local_match == 1) /* local match flag, 0 = no local match, 1 = local match? */
                                                                                                                                    {
                                                                                                                                        *first_period = '.'; /* replace the first period in the dotted quad IP address */
                                                                                                                                        *middle_period = '.'; /* replace the second period in the dotted quad IP address */
                                                                                                                                        *last_period = '.'; /* replace the third period in the dotted quad IP address */

                                                                                                                                        if (print_flag == 0) /* found in the database, and verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, reset? */
                                                                                                                                        {
                                                                                                                                            (void) printf ("%s\n", word); /* print the match */
                                                                                                                                        }

                                                                                                                                        if (continue_flag == 0) /* continue to print all matching addresses, not just the first, 0 = no, 1 = yes, set? */
                                                                                                                                        {
                                                                                                                                            continue_reading = 0; /* yes, do not containue to read the file, even if a match has been found flag, 0 = no, 1 = yes */
                                                                                                                                        }

                                                                                                                                        retval = MATCH; /* return value, assume match found in database */
                                                                                                                                        match_count ++; /* increment the count of the number of matches-the number e-mail addresses found in the database */
                                                                                                                                    }

                                                                                                                                }

                                                                                                                            }

                                                                                                                        }

                                                                                                                    }

                                                                                                                }

                                                                                                            }

                                                                                                        }

                                                                                                    }

                                                                                                }

                                                                                            }

                                                                                        }

                                                                                        break;

                                                                                    default:

                                                                                        break;

                                                                                }

                                                                            }

                                                                        }

                                                                    }

                                                                }

                                                            }

                                                            cptr = &word[0]; /* character reference to first character */
                                                            *cptr = '\0'; /* terminate the word */
                                                        }

                                                }

                                            }

                                            else
                                            {

                                                if (last_c == '\n' && fc == '\n') /* a first record with no characters is first line of the message body */
                                                {
                                                    eoheader = 1; /* passed header, set the end of header flag, 0 = record from header, 1 = record from body */
                                                }

                                            }

                                            last_c = fc; /* save the input character */
                                        }

                                        if (argc > optind) /* using stdin as input? */
                                        {

                                            if (fclose (infile) == EOF) /* no, close the input file */
                                            {
                                                retval = ECLOSE; /* error closing file */
                                            }

                                        }

                                    }

                                }
                                while ((argc > ++ optind) && ((retval == MATCH) || (retval == NOMATCH))); /* while there are any remaining input file(s), if any, listed on the command line */

                            }

                            if (munmap ((void *) mapbegin, filesize) != 0) /* unmap the file */
                            {
                                retval = EUNMAP; /* error unmapping file */
                            }

                        }

                        if (close (fd) != 0) /* close the file */
                        {
                            retval = ECLOSE; /* error closing file */
                        }

                    }

                }

            }

        }

    }

    retval = (((retval == MATCH) || (retval == NOMATCH)) && (match_count > 0)) ? MATCH : retval; /* if any matches, but no errors, set the return value to MATCH */
    print_message (retval); /* print any error messages */

    if ((error_return_change != -1) && (retval >= EARGS)) /* change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH, and error? */
    {
        retval = error_return_change; /* yes, set the error value to MATCH or NOMATCH */
    }

    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit-match found.

        2) Else if retval is unity, print nothing-normal/successful
           program exit-match not found.

        3) Else if retval is two, print help.

        4) Else retval is not zero, unity, or two, it is an error
           code, print the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, match found, print nothing */
        case 1: /* program ended without errors, match not found, print nothing */

            break;

        case 2: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Recursively binary search a memory area.

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap);

I) Binary search a memory area, referenced by the begin and end
variables, for string. The variable begin references the beginning of
the area to be searched; the variable end references the end of the
area to be searched, and will always reference the '\n' terminating
character of the last record of the memory area, (which is a
requirment); the variable string is the null terminated string to
search for, and has a length represented by the variable string_len.

II) Requirements:

    A) The memory area size must be greater than zero, (this is a
    requirement for mmap(1), also.)

    B) The last character of the memory area must be a '\n' text file
    record terminator, e.g., the last record, (which, minimally, may
    be the only record,) must be terminated.

    C) The records in the memory area must be in lexical sequence,
    (consistent with strcmp() and strncmp().)

III) The general strategy is:

    A) For each iteration in the recursion:

        1) Divide the memory area exactly in two.

        2) Work from this middle, in both directions, to find the end
        of the preceding record, (or beginning of file, whichever
        comes first) and, the end of the current record:

            a) The current record contains the first instance of the
            the terminating character at, or past, the middle.

            b) The current record's beginning is the next character
            after the preceding records's terminating character, or
            the beginning of the memory area.

        3) Compare this record with the string for a match.

        4) If not a match, decide which direction to move for the next
        iteration, and call this function, again.

        5) The recursive binary search terminates when a match is
        found, or when there is no more memory area to search.

    B) Return a reference to the matched area in memory, or null if no
    match.

Returns a reference to the matched string, null if no match.

*/

#ifdef __STDC__

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap)

#else

static char *binarysearch (begin, end, string, string_len, BOFmap, EOFmap)
char *begin;
char *end;
char *string;
size_t string_len;
char *BOFmap;
char *EOFmap;

#endif

{
    int retval; /* return for string comparisons */

    char *middle, /* reference to the middle between begin and end */
         *BOFmiddle_record, /* reference to the beginning of the record in the middle */
         *EOFmiddle_record, /* reference to the end of the record in the middle */
         *BOFnext_record, /* beginning of next record */
         *EOFprevious_record, /* end of previous record */
         *match = (char *) 0; /* reference to the beginning of the record in the middle if a match is found; null if no match */

    size_t middle_record_len; /* length record in the middle */

/* #define QA */ /* comment out to disable QA/Debuging */

#ifdef QA

    /*

    The Unix dictionary file, "sort -u /usr/local/lib/dict/websters >
    testfile", is a good QA test candidate.

    Also "thisprogram testfile < testfile > out" without QA #defined,
    and "cmp testfile out" is a good QA test candidate, (to verify
    that all words are successfully found,) where thisprogram is the
    name of the program containing this function.

    Test for the pathological cases, too; duplicate records, null
    records, single null records, etc.

    */

    char QAbuf[BUFSIZ], /* QA print buffer */
         *QAbegin, /* beginning of record to be printed */
         *QAend; /* end of record to be printed */

    size_t QAreclen; /* QA record length in print buffer */

#endif

    middle = BOFmiddle_record = EOFmiddle_record = begin + ((end - begin) / 2); /* reference the middle between begin and end */

    while ((BOFmap < BOFmiddle_record) && (*(BOFmiddle_record - 1) != '\n')) /* step backwards to the beginning of the record in the middle */
    {
        BOFmiddle_record --;
    }

    while ((EOFmiddle_record < EOFmap) && (*EOFmiddle_record != '\n')) /* step forwards to the end of the record in the middle */
    {
        EOFmiddle_record ++;
    }

    begin = (BOFmiddle_record < begin) ? BOFmiddle_record : begin; /* beginning of record in the middle outside of begin? */
    end = (end < EOFmiddle_record) ? EOFmiddle_record : end; /* end of record in the middle outside of end? */
    BOFnext_record = (EOFmiddle_record < EOFmap) ? EOFmiddle_record + 1 : EOFmap; /* save the beginning of next record */
    EOFprevious_record = (BOFmap < BOFmiddle_record) ? BOFmiddle_record - 1 : BOFmap; /* end of previous record */
    middle_record_len = (size_t) (EOFmiddle_record - BOFmiddle_record); /* save the length of the record in the middle, minus the terminating '\n' */
    retval = strncmp (string, BOFmiddle_record, string_len < middle_record_len ? string_len : middle_record_len); /* compare the record in the middle with the search string */

#ifdef QA

    (void) printf ("%s, ", string); /* print the string being searched for */

    QAreclen = strchr (BOFmiddle_record, '\n') - BOFmiddle_record; /* save the length of the record being searched */
    (void) strncpy (QAbuf, BOFmiddle_record, QAreclen); /* copy the record being searched to the print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the record being searched in the print buffer */
    (void) printf ("%s, ", QAbuf); /* print the copy of the record being searched */

    QAreclen = strchr (begin, '\n') - begin; /* length of first record in begin/end */
    (void) strncpy (QAbuf, begin, QAreclen); /* copy first record in begin/end to QA print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the first record in begin/end */
    (void) printf ("(%s - ", QAbuf); /* print the copy of the first record in begin/end */

    QAend = QAbegin = end; /* end of last record in begin/end */

    while ((BOFmap < QAbegin) && (*(QAbegin - 1) != '\n')) /* step backwards to the beginning of the last record in begin/end */
    {
        QAbegin --;
    }

    (void) strncpy (QAbuf, QAbegin, (size_t) (QAend - QAbegin)); /* copy last record in begin/end to QA print buffer */
    QAbuf[QAend - QAbegin] = '\0'; /* terminate the copy of the last record in begin/end */
    (void) printf ("%s)\n", QAbuf); /* print the copy of the last record in begin/end */

#endif

    if (retval < 0) /* string < record? */
    {

        if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
        {
            match = binarysearch (begin, EOFprevious_record, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
        }

    }

    else if (retval > 0) /* string > record? */
    {

        if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
        {
            match = binarysearch (BOFnext_record, end, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
        }

    }

    else /* match, string = record? */
    {

        if (string_len < middle_record_len) /* partial key match; string shorter than record? */
        {

            if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
            {
                match = binarysearch (begin, EOFprevious_record, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
            }

        }

        else if (string_len > middle_record_len) /* partial key match; string longer than record? */
        {

            if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
            {
                match = binarysearch (BOFnext_record, end, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
            }

        }

        else /* exact match */
        {
            match = BOFmiddle_record; /* return a reference to this record */
        }

    }

    return (match); /* return a reference to the match found between begin and end; null if no match */
}

/*

Binary search a null terminated character string array for a string.

static int str_search (const char *array[], int elements, char *string);

I) The array of character strings, each referenced indirectly by an
element in the char *array[] of elements many elements, is in lexical
order.

    A) Note: on entry, "elements" should be the index, (e.g., the
    implicit address,) of the last array element, ie., if the array
    contains n many elements, then the call would be:

        str_search (array, n - 1, string);

II) If the character string, string, is found in the char *array[],
its index, (implicit address,) in the array is returned, else -1 if
not.

*/

#define MIDWAY(a,b) (int)(((unsigned int)(a)+(unsigned int)(b))/(unsigned int)2)

#ifdef __STDC__

static int str_search (const char *array[], int elements, char *string)

#else

static int str_search (array, elements, string)
char *array[];
int elements;
char *string;

#endif

{
    int bottom = 0, /* bottom index of array */
        middle, /* middle index of array */
        top = elements, /* top index of array */
        i;

    while (bottom <= top) /* while the bottom element is lexically less than or equal to the top element */
    {
        middle = MIDWAY(bottom,top); /* starting in the middle of the array */
        i = strcmp (string,array[middle]); /* compare the key with the referenced character string */

        if (i < 0) /* if lexically less, move the top moving array index */
        {
            top = middle - 1;
        }

        else if (i > 0) /* if lexically greater, move the bottom moving array index */
        {
            bottom = middle + 1;
        }

        else  /* if lexically equal, return the element */
        {
            return (middle);
        }

    }

    return (-1); /* match not found, return -1 */
}

/*

Determine if a string is a valid dotted quad from an IP address.

static int quad_verify (char *string);

I) The quad, string, must be null terminated, and, if valid, will
consist of one, two, or three digits, which can have values of 0
through 9, 10 through 99, or 100 through 255, respectively.

II) Returns 1 if valid, else, returns -1 if not.

*/

#ifdef __STDC__

static int quad_verify (char *string)

#else

static int quad_verify (string)
char *string;

#endif

{
    int retval = -1; /* return value, assume quad is not valid */

    if (string[0] != '\0') /* null string can not be a quad */
    {

        if (string[1] == '\0') /* one digit quad, 1 - 9 */
        {

            if (string[0] >= '1' && string[0] <= '9')
            {
                retval = 1; /* valid one digit quad */
            }

        }

        else if (string[2] == '\0') /* two digit quad, 10 - 99 */
        {

            if (string[0] >= '1' && string[0] <= '9' && string[1] >= '0' && string[1] <= '9')
            {
                retval = 1; /* valid two digit quad */
            }

        }

        else if (string[3] == '\0') /* three digit quad, 100 - 255 */
        {

            if (string[0] == '1' && string[1] >= '0' && string[1] <= '9' && string[2] >= '0' && string[2] <= '9')
            {
                retval = 1; /* valid three digit quad */
            }

            else if (string[0] == '2' && string[1] >= '0' && string[1] <= '4' && string[2] >= '0' && string[2] <= '9')
            {
                retval = 1; /* valid three digit quad */
            }

            else if (string[0] == '2' && string[1] == '5' && string[2] >= '0' && string[2] <= '5')
            {
                retval = 1; /* valid three digit quad */
            }

        }

        /* fall through with error if string is longer than 3 characters */
    }

    return (retval);
}
