/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 2001-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

bsorttext.c is for binary sorting an unordered text file of character
strings.

The program requires mmap(2) to map the database file into the Unix VM
system. The database file name is a required command line argument.

The database is a standard Unix text file, one string per line.

The database mechanism is conservative with machine resources,
requiring about 12.5 micro-seconds of machine time to lookup a word in
the Unix system dictionary, (2.5 MB, quarter of a million words,
single 466 MHz., Pentium, lightly loaded, Linux 2.2, time(1) command
to lookup every word in the dictionary, divided by the number of
words.)

The most common usage is where the input file, example.db, is an
unordered list of character strings, and the output file is a uniquely
sorted list of unique character strings, with no null strings, out.db.

    bsorttext -o example.db > out.db.

The sort mechanism is implemented as a quicksort.

The program, optionally, eliminates duplicate records, (i.e., records
that are lexically equal,) removes null records, (i.e., "^$",)
converts all characters to lowercase, and issues a warning if
whitespace is found in the file.

Alternatively, the program can be used to binary search queries. The
strings to be searched for may be supplied as additional optional
command line arguments, or redirected to the program via stdin for
compatibility with procmail(1), and other e-mail scripting agents.

A suitable procmail(1) recipe example might be:

    :0 wfh
    * ? something | bsorttext reject.db
    | formail -A "X-Notice: Word in reject.db database"

which could be, if necessary, overridden, on a case-by-case basis,
with the example recipe:

    :0 wfh
    * ^X-Notice: +Word +in +reject.db +database
    * ? something | bsorttext accept.db
    | formail -I "X-Notice: Word in reject.db database"

or similar construct, where the databases contain e-mail addresses or
domain names, etc.

Similarly, the look(1) program could be used:

    :0
    * ? look -f "Word" "${HOME}/reject.db"
    | formail -I "X-Notice: Word in reject.db database"

which would provide the same functionality, but with partial key
matches.

Since the database file is read-only memory mapped, using mmap(2), and
the database file closed immediately after the mmap call, the
unstructured/unordered database file can be appended from the output
of the bsorttext(1) program, i.e., for example, constructs like:

    bsorttext -P example.db "this" "and" "that" > example.db

are permitted, (which, for example, would add the words "this", "and",
"that" to the unstructured/unordered database file, example.db, but
only if the words were not already in the file.)

Additionally, it is not required that the database file exist, and/or
be consistent with the requirements of mmap(2) for the -P
option. Specifically, the file does not have to exist, and can have a
size of zero-but only for the -P option. (If the file exists, and has
a length greater than zero, the file must be readable, and the last
character in the file must be a '\n', however.)

The program contains less than 300 lines of declarations and
statements, all of which are documented with in line comments.

The program has been compiled and tested on SunOS, Solaris, and Linux,
and may work on other brands of Unix.

If used for binary sorting an unordered text file of character
strings, the program returns 0 if no error; else returns a unique
error code greater than 1 representing the error encountered-which
will, also, print an error diagnostic to stderr.

If used for querying an unordered text file of character strings, the
program returns 0 if no error and any of the specified strings were
found in the database file, 1 if no error and no strings were found;
else returns a unique error code greater than 1 representing the error
encountered-which will, also, print an error diagnostic to stderr.

The -r option is useful for controlling the return value under error
conditions-for example, the program return can be preempted if the
database file can not be opened, (or read,) with a return value of
match, or no match, depending on environmental requirements.

John Conover
January 17, 2005
john@email.johncon.com

$Revision: 0.0 $
$Date: 2005/01/17 19:36:00 $
$Id: bsorttext.c,v 0.0 2005/01/17 19:36:00 john Exp $
$Log: bsorttext.c,v $
Revision 0.0  2005/01/17 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* for getopt() */
#include <ctype.h> /* for toupper(), tolower() */
#include <sys/types.h> /* for open() */
#include <sys/stat.h> /* for open() */
#include <fcntl.h> /* for open() */
#include <unistd.h> /* for mmap() */
#include <sys/mman.h> /* for mmap() */

static char rcsid[] = "$Id: bsorttext.c,v 0.0 2005/01/17 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 2001-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define ALPHABET_SIZE 256 /* 256 characters in the alphabet */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Binary sort an unordered text file of character strings:\n",
    "Usage: bsorttext [-I] [-N] [-o] [-P] [-r m|n] [-S] [-U] [-v] filename [string(s)]\n",
    "    -I, case sensitive alphabet\n",
    "    -N, include null records\n",
    "    -o, output the sorted file, filename, disable querying\n",
    "    -P, print the string(s) not in the database\n",
    "    -r, m|n, on file error, exit return = match for m, no match for n\n",
    "    -S, disable whitespace in file warning\n",
    "    -U, records to not have to be unique\n",
    "    -v, print the program's version information\n",
    "    filename, database file name\n",
    "    string(s), character string(s) to be searched for, (defaults to stdin)\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error, match found\n",
    "No error, match not found\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error in status of database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error allocating memory\n",
    "Warning whitespace in file\n",
    "Error in buffer\n",
    "Error in database file\n"
};

#define MATCH 0 /* error values, one for each index in the error message array */
#define NOERROR 0
#define NOMATCH 1
#define EARGS 2
#define EOPEN 3
#define ECLOSE 4
#define ESTAT 5
#define EMAP 6
#define EUNMAP 7
#define EALLOC 8
#define WSPACE 9
#define EBOF 10
#define EDB 11

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static void str_sort (char *array[],int bottom,int top);
static int str_search (char *array[], int elements, char *string);
static int strcmpeol (char *s, char *t);

#else

static void print_message (); /* print any error messages */
static void str_sort ();
static int str_search ();
static int strcmpeol ();

#endif

static char alphabet[ALPHABET_SIZE]; /* character table for the alphabet */
static int slash_n = (int) '\n'; /* the line return, from the alphabet; for expediency */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char ch, /* output character */
         buffer[BUFLEN + 1], /* i/o buffer */
         *cptr, /* temporary character reference */
         *end, /* end of memory mapped file */
         *last_record = (char *) 0, /* reference to the last record output */
         *eobuffer, /* reference to end of buffer character */
         **ref_array; /* reference to array of pointers to records in memory mapped file */

    int retval = NOERROR, /* return value, assume no error */
        fd = -1, /* input file descriptor */
        argctr, /* argument counter of search strings supplied from the command line */
        c, /* command line switch */
        match_count = 0, /* count of the number of matches-the number of IP addresses found in the database */
        line_counter, /* the count of lines in the file */
        i, /* line counter */
        i_flag = 0, /* case sensitive, 0 = no, 1 = yes */
        n_flag = 0, /* include null records, 0 = no, 1 = yes */
        o_flag = 0, /* output the sorted file, filename, disable querying, 0 = no, 1 = yes */
        print_flag = 0, /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match */
        s_flag = 0, /* disable whitespace in file warning, 0 = no, 1 = yes */
        u_flag = 0, /* records do not have to be unique, 0 = no, 1 = yes */
        error_return_change = -1; /* change exit return on error, -1 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */

    FILE *infile = stdin; /* reference to stdin */

    size_t filesize; /* size of file */

    struct stat fileinfo; /* structure for file status */

    caddr_t mapbegin; /* reference to beginning of mmap'ed file */

    while ((c = getopt (argc, argv, "INoPr:SUv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'I': /* request for case sensitive */

                i_flag = 1; /* yes, set the case sensitive, 0 = no, 1 = yes, flag */
                break;

            case 'N': /* request to include null records? */

                n_flag = 1; /* yes, set the include null records, 0 = no, 1 = yes, flag */
                break;

            case 'o': /* request for output the sorted file, filename, disable querying? */

                o_flag = 1; /* yes, set the output the sorted file, filename, disable querying, 0 = no, 1 = yes, flag */
                break;

            case 'P': /* request for verbose print? */

                print_flag = 1; /* set the verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match */
                break;

            case 'r': /* request for change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH? */

                switch (*optarg) /* first character of argument object, only */
                {

                    case 'm':

                        error_return_change = MATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    case 'n':

                        error_return_change = NOMATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    default:

                        optind = argc; /* force argument error */
                        retval = EARGS; /* assume not enough arguments */
                        break;
                }

                break;

            case 'S':  /* request for disable whitespace in file warning? */

                s_flag = 1; /* yes, set the disable whitespace in file warning, 0 = no, 1 = yes, flag */
                break;

            case 'U': /* request for records do not have to be unique? */

                u_flag = 1; /* yes, set the records do not have to be unique, 0 = no, 1 = yes, flag */
                break;

            case 'v': /* request for program's version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* any errors? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (argc > optind) /* enough arguments? */
        {

            if (i_flag == 0) /* case sensitive, 0 = no, 1 = yes, set? */
            {

                for (i = 0; i < ALPHABET_SIZE; i ++) /* for each character in the alphabet */
                {
                    alphabet[i] = (char) tolower ((char) i); /* load the character table for the alphabet */
                }

            }

            else
            {

                for (i = 0; i < ALPHABET_SIZE; i ++) /* for each character in the alphabet */
                {
                    alphabet[i] = (char) i; /* load the character table for the alphabet */
                }

            }

            slash_n = alphabet[(int) '\n']; /* save the line return, from the alphabet; for expediency */
            retval = ESTAT; /* assume error in status of file */

            if (stat (argv[optind], &fileinfo) == 0) /* get the status of the file */
            {
                retval = EDB;  /* assume error in database file */

                if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of file */
                {
                    retval = EOPEN; /* assume error opening file */

                    if ((fd = open (argv[optind], O_RDONLY)) != -1) /* open the file */
                    {
                        retval = EMAP; /* assume error mapping file */

                        if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, fd, (off_t) 0)) != (caddr_t) 0) /* map the file */
                        {
                            retval = ECLOSE; /* error closing file */

                            if (close (fd) == 0) /* close the file */
                            {
                                fd = -1; /* reset the input file descriptor */
                                retval = EDB;  /* assume error in database file */

                                if (mapbegin[filesize - 1] == slash_n) /* verify that the last record of the memory mapped file has a terminating '\n' character as required */
                                {
                                    retval = NOERROR; /* assume no error */
                                    optind ++; /* file argument handled */
                                    cptr = mapbegin; /* reference the beginning of the map file */
                                    end = &mapbegin[filesize - 1]; /* reference the end of the map file */
                                    line_counter = 0; /* zero the count of lines in the file */

                                    do  /* for each EOL in the file */
                                    {
                                        line_counter ++; /* increment the count of lines in the file */
                                        cptr ++; /* step over the EOL */
                                    }
                                    while ((cptr < end) && (cptr = strchr (cptr, slash_n)));

                                    retval = EALLOC; /* assume allocating memory */

                                    if ((ref_array = (char **) malloc (((size_t) line_counter) * sizeof (char *))) != (char **) 0) /* allocate space for the references to the array of pointers to records in memory mapped file */
                                    {
                                        retval = NOERROR; /* assume no error */
                                        cptr = mapbegin - 1; /* reference the preceeding character from the beginning of the map file */
                                        line_counter = 0; /* zero the count of lines in the file */

                                        do /* for each EOL in the file */
                                        {
                                            cptr ++; /* step over the EOL */
                                            ref_array[line_counter] = cptr; /* save the reference to the record in the memory mapped file */
                                            line_counter ++; /* increment the count of lines in the file */
                                        }
                                        while ((cptr = strchr (cptr, slash_n)) && cptr < end);

                                        str_sort (ref_array, 0, line_counter - 1); /* sort the array of pointers to records in memory mapped file */

                                        if (o_flag == 1) /* output the sorted file, filename, disable querying, 0 = no, 1 = yes flag set? */
                                        {

                                            for (i = 0; i < line_counter; i ++) /* for each record in the file */
                                            {
                                                cptr = ref_array[i]; /* reference the record */

                                                if ((last_record == (char *) 0) || (strcmpeol (cptr, last_record) != 0) || (u_flag == 1)) /* if the record is the same as the last record output, records do not have to be unique, 0 = no, 1 = yes flag set? */
                                                {
                                                    ch = alphabet[(int) *cptr]; /* save the output character */

                                                    if ((ch != slash_n) || (n_flag == 1)) /* null record means the first character of the record is '\n', include null records, 0 = no, 1 = yes flag set? */
                                                    {

                                                        do /* for each character in the record, up to, and including the EOL of the record */
                                                        {
                                                            ch = alphabet[(int) *cptr]; /* save the output character */

                                                            if ((s_flag == 0) && (ch != slash_n) && isspace (ch)) /* character whitespace, disable whitespace in file warning, 0 = no, 1 = yes, flag reset? */
                                                            {
                                                                s_flag = 1; /* reset the disable whitespace in file warning, 0 = no, 1 = yes, flag; notify only once */
                                                                retval = WSPACE; /* assume warning whitespace in file */
                                                            }

                                                            (void) putc ((int) ch, stdout); /* print the character */
                                                        }
                                                        while (alphabet[(int) *cptr ++] != slash_n);

                                                        last_record = ref_array[i]; /* save the reference to the last record output */
                                                    }

                                                }

                                            }

                                        }

                                        else
                                        {
                                            retval = NOMATCH; /* return value, assume no error */

                                            if (optind < argc) /* more arguments are strings to search for; if none, assume strings to search for come from stdin */
                                            {

                                                for (argctr = optind; (argctr < argc) && (retval == NOMATCH); argctr ++) /* for each string to search for supplied on the command line, but only while there are no errors */
                                                {
                                                    retval = EBOF; /* assume error in buffer */

                                                    if (strlen (argv[argctr]) < BUFLEN) /* string to search for supplied on the command line fit in the buffer? */
                                                    {
                                                        retval = NOMATCH; /* return value, assume no error */
                                                        (void) strcpy (buffer, argv[argctr]); /* make a copy of the string to search for supplied on the command line */
                                                        (void) strcat (buffer, "\n"); /* add a EOL charcter */

                                                        if ((i = str_search (ref_array, line_counter - 1, buffer)) >= 0) /* search for matching record */
                                                        {
                                                            match_count ++; /* match found, increment the count of the number of matches-the number of IP addresses found in the database */

                                                            if (print_flag == 0) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, reset? */
                                                            {
                                                                cptr = ref_array[i]; /* reference the record */

                                                                do /* for each character in the record, up to, and including the EOL of the record */
                                                                {
                                                                    (void) putc ((int) alphabet[(int) *cptr], stdout); /* print the character */
                                                                }
                                                                while (alphabet[(int) *cptr ++] != slash_n);

                                                            }

                                                        }

                                                        else
                                                        {

                                                            if (print_flag == 1) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, set? */
                                                            {
                                                                cptr = buffer; /* reference the copy of the record */

                                                                do /* for each character in the record, up to, and including the EOL of the record */
                                                                {
                                                                    (void) putc ((int) alphabet[(int) *cptr], stdout); /* print the character */
                                                                }
                                                                while (alphabet[(int) *cptr ++] != slash_n);

                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                            else
                                            {

                                                while ((fgets (buffer, BUFLEN, infile) != (char *) 0) && (retval == NOMATCH)) /* read the strings to search for from stdin, but only while there are no errors */
                                                {
                                                    retval = EBOF; /* assume error in buffer */

                                                    if ((eobuffer = strchr (buffer, '\n')) != (char *) 0) /* a record with no carriage return was too long for the fgets () buffer, and an overflow condition exists; line contain a carriage return? */
                                                    {
                                                        retval = NOMATCH; /* return value, assume no error */

                                                        if ((i = str_search (ref_array, line_counter - 1, buffer)) >= 0) /* search for matching record */
                                                        {
                                                            match_count ++; /* match found, increment the count of the number of matches-the number of IP addresses found in the database */

                                                            if (print_flag == 0) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, reset? */
                                                            {
                                                                cptr = ref_array[i]; /* reference the record */

                                                                do /* for each character in the record, up to, and including the EOL of the record */
                                                                {
                                                                    (void) putc ((int) alphabet[(int) *cptr], stdout); /* print the character */
                                                                }
                                                                while (alphabet[(int) *cptr ++] != slash_n);

                                                            }

                                                        }

                                                        else
                                                        {

                                                            if (print_flag == 1) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, set? */
                                                            {
                                                                cptr = buffer; /* reference the string */

                                                                do /* for each character in the record, up to, and including the EOL of the record */
                                                                {
                                                                    (void) putc ((int) alphabet[(int) *cptr], stdout); /* print the character */
                                                                }
                                                                while (alphabet[(int) *cptr ++] != slash_n);

                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                        }

                                        free (ref_array); /* free the space for the references to the array of pointers to records in memory mapped file */
                                    }

                                }

                            }

                            if (munmap ((void *) mapbegin, filesize) != 0) /* unmap the file */
                            {
                                retval = EUNMAP; /* error unmapping file */
                            }

                        }

                        if (fd != -1) /* file open? */
                        {

                            if (close (fd) != 0) /* close the file */
                            {
                                retval = ECLOSE; /* error closing file */
                            }

                        }

                    }

                }

            }

            if ((print_flag == 1) && ((retval == EMAP || retval == ESTAT))) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, and, error mapping file or error in status of file, set? */
            {
                optind ++; /* file argument handled, but didn't work out */
                retval = NOMATCH; /* return value, assume no error */

                if (optind < argc) /* more arguments are strings to search for; if none, assume strings to search for come from stdin */
                {

                    for (argctr = optind; (argctr < argc) && (retval == NOMATCH); argctr ++) /* for each string to search for supplied on the command line, but only while there are no errors */
                    {
                        retval = EBOF; /* assume error in buffer */

                        if (strlen (argv[argctr]) < BUFLEN) /* string to search for supplied on the command line fit in the buffer? */
                        {
                            retval = NOMATCH; /* return value, assume no error */
                            (void) strcpy (buffer, argv[argctr]); /* make a copy of the string to search for supplied on the command line */
                            (void) strcat (buffer, "\n"); /* add a EOL charcter */
                            cptr = buffer; /* reference the copy of the record */

                            do /* for each character in the record, up to, and including the EOL of the record */
                            {
                                (void) putc ((int) alphabet[(int) *cptr], stdout); /* print the character */
                            }
                            while (alphabet[(int) *cptr ++] != slash_n);

                        }

                    }

                }

                else
                {

                    while ((fgets (buffer, BUFLEN, infile) != (char *) 0) && (retval == NOMATCH)) /* read the strings to search for from stdin, but only while there are no errors */
                    {
                        retval = EBOF; /* assume error in buffer */

                        if ((eobuffer = strchr (buffer, '\n')) != (char *) 0) /* a record with no carriage return was too long for the fgets () buffer, and an overflow condition exists; line contain a carriage return? */
                        {
                            retval = NOMATCH; /* return value, assume no error */
                            cptr = buffer; /* reference the string */

                            do /* for each character in the record, up to, and including the EOL of the record */
                            {
                                (void) putc ((int) alphabet[(int) *cptr], stdout); /* print the character */
                            }
                            while (alphabet[(int) *cptr ++] != slash_n);

                        }

                    }

                }

            }

        }

    }

    if (o_flag == 1) /* output the sorted file, filename, disable querying, 0 = no, 1 = yes, flag set? */
    {
        print_message (retval); /* print any error messages */

        if (retval == WSPACE) /* return value a warning of whitespace in file? */
        {
            retval = NOERROR; /* assume no error */
        }

    }

    else
    {
        retval = (((retval == MATCH) || (retval == NOMATCH)) && (match_count > 0)) ? MATCH : retval; /* if any matches, but no errors, set the return value to MATCH */
        print_message (retval); /* print any error messages */

        if ((error_return_change != -1) && (retval >= EARGS)) /* change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH, and error? */
        {
            retval = error_return_change; /* yes, set the error value to MATCH or NOMATCH */
        }

    }

    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit-match found.

        2) Else if retval is unity, print nothing-normal/successful
           program exit-match not found.

        3) Else if retval is two, print help.

        4) Else retval is not zero, unity, or two, it is an error
           code, print the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, match found, print nothing */
        case 1: /* program ended without errors, match not found, print nothing */

            break;

        case 2: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Quicksort an EOL, ('\n'), terminated array of character strings.

static void str_sort (char *array[], int bottom, int top);

Note: this is a recursive proceedure, requiring three references, and
4 integers to be allocated on the stack for each recursion.

Note: on entry, argument "bottom" should be zero, and argument "top"
should be the index of the last array element, ie., if the array
contains n many elements, then the call would be:

    str_sort (array, 0, n - 1);

Returns nothing.

*/

#define MIDWAY(a, b) (int)(((unsigned int)(a)+(unsigned int)(b))/(unsigned int)2)

#ifdef __STDC__

static void str_sort (char *array[], int bottom, int top)

#else

static void str_sort (array, bottom, top)
char *arrag[];
int bottom;
int top;

#endif

{
    int i = bottom, /* bottom index of the array */
        j = top; /* top index of the array */

    char *x,
         *y;

    x = array[MIDWAY(bottom, top)]; /* starting in the middle of the array */

    do /* while the bottom moving index of the array is less than or equal to the top moving index of the array */
    {

        while (strcmpeol (array[i], x) < 0 && i < top) /* increment the bottom moving index of the array while the elements referenced are lexically "less" */
        {
            i ++;
        }

        while (strcmpeol (array[j], x) > 0 && j > bottom) /* decrement the top moving index of the array while the elements referenced are lexically "greater" */
        {
            j --;
        }

        if (i <= j) /* if the bottom moving index is less than or equal to the top moving index of the array */
        {
            y = array[i]; /* swap the element references, increment the bottom moving index, decrement the top moving index */
            array[i] = array[j];
            array[j] = y;
            i ++;
            j --;
        }

    }
    while (i <= j);

    if (bottom < j) /* if the top moving index is not at the bottom of the array elements, sort the remaining elements */
    {
        str_sort (array, bottom, j);
    }

    if (i < top) /* if the bottom moving index is not at the top of the array elements, sort the remaining elements */
    {
        str_sort (array, i, top);
    }

}

/*

Binary search an EOL, ('\n'), terminated array of character strings.

Search an array of referenced character strings and returns the array
index of the reference of a character string that matches a key-the
array of referenced character strings must be in lexical sequence, -1
if the matching character string reference was not found

Note: on entry, "elements" should be the index of the last array
element, ie., if the array contains n many elements, then the call
would be:

    str_search (array, n - 1, string);

Returns the index of the match into array, -1 if not found.

*/

#ifdef __STDC__

int str_search (char *array[] ,int elements, char *string)

#else

int str_search (array, elements, string)
    char *array[];
    int elements;
    char *string;

#endif

{
    int bottom = 0, /* bottom index of array */
        middle, /* middle index of array */
        top = elements, /* top index of array */
        i;

    while (bottom <= top) /* while the bottom element is lexically less than or equal to the top element */
    {
        middle = MIDWAY(bottom,top); /* starting in the middle of the array */
        i = strcmpeol (string, array[middle]); /* compare the key with the referenced character string */

        if (i < 0) /* if lexically less, move the top moving array index */
        {
            top = middle - 1;
        }

        else if (i > 0) /* if lexically greater, move the bottom moving array index */
        {
            bottom = middle + 1;
        }

        else  /* if lexically equal, return the element */
        {
            return (middle);
        }

    }

    return (-1); /* match not found, return -1 */
}

/*

Compare two EOL, ('\n'), terminated character strings, possibly using
a case insensitive character comparison.

static int strcmpeol (char *s, char *t);

Exactly the same as strcmp (), but the character strings are EOL,
'\n', instead of null, '\0', terminated.

The global character array "alphabet" contains the character set, for
example, if case insensitive comparison is desired, the A'th character
in the array, alphabet['A'], would return 'a', and so on.

Returns an integer less than, equal to, or greater than zero to
indicate whether string, s, is less than, equal to, or greater than
string, t.

*/

#ifdef __STDC__

static int strcmpeol (char *s, char *t)

#else

static int strcmpeol (s, t)
char *s;
char *t;

#endif

{

    for ( ; alphabet[(int) *s] == alphabet[(int) *t]; s ++, t ++) /* for each character */
    {

        if (alphabet[(int) *s] == slash_n) /* if the source is at EOL */
        {
            return (0); /* match, return zero */
        }

    }

    return ((int) (alphabet[(int) *s] - alphabet[(int) *t])); /* no match, return the difference between the last two characters compared */
}
