/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 2001-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

hsearchtext.c is for hash searching an unordered text file of
character strings.

The program requires mmap(2) to map the database file into the Unix VM
system. The database file name is a required command line argument.

The database is a standard Unix text file, one string per line.

The database mechanism is conservative with machine resources,
requiring about 17.5 micro-seconds of machine time to lookup a word in
the Unix system dictionary, (2.5 MB, quarter of a million words,
single 466 MHz., Pentium, lightly loaded, Linux 2.2, time(1) command
to lookup every word in the dictionary, divided by the number of
words.)

The program, optionally, eliminates duplicate records, (i.e., records
that are lexically equal,) removes null records, (i.e., "^$",)
converts all characters to lowercase, and parses records with
whitespace, leaving only the last token as the record.

The program can be used to hash queries. The strings to be searched
for may be supplied as additional optional command line arguments, or
redirected to the program via stdin for compatibility with
procmail(1), and other e-mail scripting agents.

A suitable procmail(1) recipe example might be:

    :0 wfh
    * ? something | hsearchtext reject.db
    | formail -A "X-Notice: Word in reject.db database"

which could be, if necessary, overridden, on a case-by-case basis,
with the example recipe:

    :0 wfh
    * ^X-Notice: +Word +in +reject.db +database
    * ? something | hsearchtext accept.db
    | formail -I "X-Notice: Word in reject.db database"

or similar construct, where the databases contain e-mail addresses or
domain names, etc.

Since the database file is read-only memory mapped, using mmap(2), and
the database file closed immediately after the mmap call, the
unstructured/unordered database file can be appended from the output
of the hsearchtext(1) program, i.e., for example, constructs like:

    hsearchtext -P example.db "this" "and" "that" > example.db

are permitted, (which, for example, would add the words "this", "and",
"that" to the unstructured/unordered database file, example.db, but
only if the words were not already in the file.)

Additionally, it is not required that the database file exist, and/or
be consistent with the requirements of mmap(2).  Specifically, the
file does not have to exist, and/or can have a size of zero.

The program contains less than 700 lines of declarations and
statements, all of which are documented with in line comments.

The program has been compiled and tested on SunOS, Solaris, and Linux,
and may work on other brands of Unix.

If used for querying an unordered text file of character strings, the
program returns 0 if no error and any of the specified strings were
found in the database file, 1 if no error and no strings were found;
else returns a unique error code greater than 1 representing the error
encountered-which will, also, print an error diagnostic to stderr.

The -r option is useful for controlling the return value under error
conditions-for example, the program return can be preempted if the
database file can not be opened, (or read,) with a return value of
match, or no match, depending on environmental requirements.

John Conover
May 17, 2005
john@email.johncon.com

$Revision: 0.0 $
$Date: 2005/01/17 19:36:00 $
$Id: hsearchtext.c,v 0.0 2005/01/17 19:36:00 john Exp $
$Log: hsearchtext.c,v $
Revision 0.0  2005/01/17 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* for getopt() */
#include <ctype.h> /* for toupper(), tolower() */
#include <sys/types.h> /* for open() */
#include <sys/stat.h> /* for open() */
#include <fcntl.h> /* for open() */
#include <unistd.h> /* for mmap() */
#include <sys/mman.h> /* for mmap() */

#ifndef __STDC__

#include <malloc.h>

#endif

static char rcsid[] = "$Id: hsearchtext.c,v 0.0 2005/01/17 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 2001-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define TOKEN_SEPARATORS " \t\n\r\b" /* file record whitespace */

#define ALPHABET_SIZE 256 /* 256 characters in the alphabet */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Hash search an unordered text file of character strings:\n",
    "Usage: hsearchtext [-A] [-h] [-I] [-N] [-P] [-r m|n] [-S] [-v] filename\n",
    "    -A, return = match if all strings found, (match if any string found)\n",
    "    -h number, hash table size = prime number (99871)\n",
    "    -I, case sensitive alphabet\n",
    "    -N, include null records\n",
    "    -P, print the string(s) not in the database\n",
    "    -r, m|n, on file error, exit return = match for m, no match for n\n",
    "    -S, disable whitespace in file warning\n",
    "    -v, print the program's version information\n",
    "    filename, database file name\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error, match found\n",
    "No error, match not found\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error in status of database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error allocating memory\n",
    "Error in buffer\n",
    "Error in database file\n",
    "Error allocating hash table\n",
    "Error inserting record in hash table\n",
    "Error finding record in hash table\n"
};

#define MATCH 0 /* error values, one for each index in the error message array */
#define NOERROR 0
#define NOMATCH 1
#define EARGS 2
#define EOPEN 3
#define ECLOSE 4
#define ESTAT 5
#define EMAP 6
#define EUNMAP 7
#define EALLOC 8
#define EBOF 10
#define EDB 11
#define EHASHINIT 12
#define EHASHINSERT 13
#define EHASHFIND 14

/* Internal hash errors, will end up as EHASHINIT or EHASHINSERT */

#define HASH_INI_ERR 15 /* hash table already initilaized */
#define HASH_DUP_ERR 16 /* duplicate key when inserting key into hash table */
#define HASH_MK_ERR 17 /* hash table mkhash () failure */
#define HASH_KEY_ERR 18 /* hash table key not found */

#define UNIQUE_KEYS /* multiple keys not be permitted in the hash table, remove this line if multiple keys permitted */

typedef struct hash_struct /* hash table element, the required elements are struct hash_struct *previous; and struct hash_struct *next; other appropriate elements may be added, consistent with the capabilities of the functions in HASHTABLE */
{
    struct hash_struct *previous; /* reference to next element in hash's doubly, circular linked list */
    void *hash_data; /* reference to element's data */
    int hash_data_printed; /* hash data printed, 0 = no, 1 = yes */
    struct hash_struct *next; /* reference to previous element in hash's doubly, circular linked list */
} HASH;

#ifdef __STDC__

typedef struct hashtable_struct /* hash table descriptor */
{
    size_t hash_size; /* size of hash table */
    HASH *table; /* reference to hash table index */
    HASH *(*mkhash) (void *data); /* reference to the function that allocates hash elements */
    int (*cmphash) (void *data, struct hash_struct *element); /* reference to the function that compares element's keys */
    void (*rmhash) (struct hash_struct *element); /* reference to the function that deallocates hash elements */
    int (*comphash) (struct hashtable_struct *hash_table, void *key); /* reference to the function that computes the hash value */
} HASHTABLE;

#else

typedef struct hashtable_struct /* hash table descriptor */
{
    size_t hash_size; /* size of hash table */
    HASH *table; /* reference to hash table */
    HASH *(*mkhash) (); /* reference to the function that allocates data and hash elements */
    int (*cmphash) (); /* reference to the function that compares element's keys */
    void (*rmhash) (); /* reference to the function that deallocates data hash elements */
    int (*comphash) (); /* reference to the function that computes the hash value */
} HASHTABLE;

#endif

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static int hash_init (HASHTABLE *hash_table);
static int hash_insert (HASHTABLE *hash_table, void *data);
static HASH *hash_find (HASHTABLE *hash_table, void *data);
static void hash_term (HASHTABLE *hash_table);
static int hash_text (HASHTABLE *hash_table, void *key);
static int text_cmphash (void *data, HASH *element); /* function to compare a text key with a hash table's element key */
static HASH *text_mkhash (void *data); /* function to allocate a text hash table element and data */
static void text_rmhash (HASH *element); /* function to deallocate a text hash table element and data */
static int strtoken (char *string, char *parse_array, char **parse, const char *delim);

#else

static void print_message (); /* print any error messages */
static int hash_init ();
static int hash_insert ();
static HASH *hash_find ();
static void hash_term ();
static int hash_text ();
static int text_cmphash (); /* function to compare a text key with a hash table's element key */
static HASH *text_mkhash (); /* function to allocate a text hash table element and data */
static void text_rmhash (); /* function to deallocate a text hash table element and data */
static int strtoken ();

#endif

static char alphabet[ALPHABET_SIZE]; /* character table for the alphabet */
static int slash_n = (int) '\n'; /* the line return, from the alphabet; for expediency */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN + 1], /* i/o buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         *cptr, /* temporary character reference */
         *bptr, /* temporary character reference */
         *end, /* end of memory mapped file */
         *eobuffer; /* reference to end of buffer character */

    int retval = NOERROR, /* return value, assume no error */
        tretval, /* temporary return value */
        fd = -1, /* input file descriptor */
        fields, /* number of fields in a record */
        argctr, /* argument counter of search strings supplied from the command line */
        c, /* command line switch */
        match_count = 0, /* count of the number of matches-the number of IP addresses found in the database */
        match, /* match in lookup of record, 0 = no, 1 = yes */
        len, /* length of a record in buffer */
        i, /* line counter */
        i_flag = 0, /* case sensitive, 0 = no, 1 = yes */
        n_flag = 0, /* include null records, 0 = no, 1 = yes */
        print_flag = 0, /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match */
        s_flag = 0, /* disable whitespace in file warning, 0 = no, 1 = yes */
        error_return_change = -1, /* change exit return on error, -1 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
        search_count = 0, /* number of searches */
        all_match = 0; /* all match flag, 0 = no, 1 = yes */

    FILE *infile = stdin; /* reference to stdin */

    size_t filesize; /* size of file */

    struct stat fileinfo; /* structure for file status */

    caddr_t mapbegin = (caddr_t) 0; /* reference to beginning of mmap'ed file */

    HASH *element; /* reference to hash table element */

    HASHTABLE text_table = {99871, (HASH *) 0, text_mkhash, text_cmphash, text_rmhash, hash_text}; /* declare the hash table descriptor for text keys*/

    while ((c = getopt (argc, argv, "Ah:INPr:Sv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'A': /* request for all match? */

                all_match = 1; /* set the all match flag, 0 = no, 1 = yes */
                break;

            case 'h': /* request for hash table size */

                if ((tretval = atoi (optarg)) > 1) /* cursory check on hash table size */
                {
                    text_table.hash_size = (size_t) tretval; /* save the hash table size */
                }

                break;

            case 'I': /* request for case sensitive */

                i_flag = 1; /* yes, set the case sensitive, 0 = no, 1 = yes, flag */
                break;

            case 'N': /* request to include null records? */

                n_flag = 1; /* yes, set the include null records, 0 = no, 1 = yes, flag */
                break;

            case 'P': /* request for verbose print? */

                print_flag = 1; /* set the verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match */
                break;

            case 'r': /* request for change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH? */

                switch (*optarg) /* first character of argument object, only */
                {

                    case 'm':

                        error_return_change = MATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    case 'n':

                        error_return_change = NOMATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    default:

                        optind = argc; /* force argument error */
                        retval = EARGS; /* assume not enough arguments */
                        break;
                }

                break;

            case 'S':  /* request for disable whitespace in file warning? */

                s_flag = 1; /* yes, set the disable whitespace in file warning, 0 = no, 1 = yes, flag */
                break;

            case 'v': /* request for program's version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* any errors? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (argc > optind) /* enough arguments? */
        {

            if (i_flag == 0) /* case sensitive, 0 = no, 1 = yes, set? */
            {

                for (i = 0; i < ALPHABET_SIZE; i ++) /* for each character in the alphabet */
                {
                    alphabet[i] = (char) tolower ((char) i); /* load the character table for the alphabet */
                }

            }

            else
            {

                for (i = 0; i < ALPHABET_SIZE; i ++) /* for each character in the alphabet */
                {
                    alphabet[i] = (char) i; /* load the character table for the alphabet */
                }

            }

            slash_n = alphabet[(int) '\n']; /* save the line return, from the alphabet; for expediency */
            retval = EHASHINIT;  /* assume error allocating hash table */

            if (hash_init (&text_table) == NOERROR) /* initialize the hash table */
            {
                retval = ESTAT; /* assume error in status of file */

                if (stat (argv[optind], &fileinfo) == 0) /* get the status of the file */
                {
                    retval = EDB;  /* assume error in database file */

                    if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of file */
                    {
                        retval = EOPEN; /* assume error opening file */

                        if ((fd = open (argv[optind], O_RDONLY)) != -1) /* open the file */
                        {
                            retval = EMAP; /* assume error mapping file */

                            if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, fd, (off_t) 0)) != (caddr_t) 0) /* map the file */
                            {
                                retval = EDB;  /* assume error in database file */

                                if (mapbegin[filesize - 1] == slash_n) /* verify that the last record of the memory mapped file has a terminating '\n' character as required */
                                {
                                    retval = NOERROR; /* assume no error */
                                }

                            }

                            if (close (fd) != 0) /* close the file */
                            {
                                retval = ECLOSE; /* error closing file */
                            }

                            fd = -1; /* reset the input file descriptor */
                        }

                    }

                }

                if (retval == NOERROR) /* file open and mapped? */
                {

                    cptr = bptr = mapbegin; /* reference the beginning of the map file */
                    end = &mapbegin[filesize - 1]; /* reference the end of the map file */

                    while ((cptr < end) && (cptr = strchr (cptr, slash_n))) /* for each EOL in the file */
                    {

                        if ((len = (cptr - bptr)) >= BUFLEN) /* length of a record in buffer create a buffer overflow? */
                        {
                            retval = EBOF; /* error in buffer */
                            break;
                        }

                        (void) strncpy (buffer, bptr, (size_t) len); /* save the record */
                        buffer[len] = '\0';

                        if (i_flag == 0) /* case sensitive, 0 = no, 1 = yes, flag set? */
                        {
                            eobuffer = buffer; /* reference the beginning of the buffer character */

                            while ((*eobuffer = alphabet[(int) *eobuffer]) != '\0') /* for each character of the string, convert it to lower case */
                            {
                                eobuffer ++; /* next character */
                            }

                        }

                        if (s_flag == 0) /* disable whitespace in file, 0 = no, 1 = yes, flag set? */
                        {

                            if ((fields = strtoken (buffer, parsebuffer, token, TOKEN_SEPARATORS)) > 0) /* parse the record to remove whitespace */
                            {
                                (void) strcpy (buffer, token[fields - 1]); /* and copy the last token */
                            }

                        }

                        if (n_flag != 0 || buffer[0] != '\0') /* include null records, 0 = no, 1 = yes, flag set? */
                        {

                            if ((retval = hash_insert (&text_table, buffer)) != NOERROR) /* add a record to the hash table */
                            {

                                if (retval != HASH_DUP_ERR) /* record already in hash table; this record a duplicate? */
                                {
                                    retval = EHASHINSERT; /* error inserting record in hash table */
                                    break;
                                }

                            }

                        }

                        cptr ++; /* step over the record's EOL */
                        bptr = cptr;
                    }

                }

                if (mapbegin != (caddr_t) 0) /* mapped file exist? */
                {

                    if (munmap ((void *) mapbegin, filesize) != 0) /* yes, unmap the file */
                    {
                        retval = EUNMAP; /* error unmapping file */
                    }

                }

                optind ++; /* file argument handled */
                retval = NOMATCH; /* return value, assume no error */

                if (optind < argc) /* more arguments are strings to search for; if none, assume strings to search for come from stdin */
                {

                    for (argctr = optind; (argctr < argc) && (retval == NOMATCH); argctr ++) /* for each string to search for supplied on the command line, but only while there are no errors */
                    {
                        retval = EBOF; /* assume error in buffer */

                        if (strlen (argv[argctr]) < BUFLEN) /* string to search for supplied on the command line fit in the buffer? */
                        {
                            retval = NOMATCH; /* return value, assume no error */
                            (void) strcpy (buffer, argv[argctr]); /* make a copy of the string to search for supplied on the command line */
                            match = 0; /* match in lookup of record, 0 = no, 1 = yes */

                            if (i_flag == 0) /* case sensitive, 0 = no, 1 = yes, flag set? */
                            {
                                eobuffer = buffer; /* reference the beginning of the buffer character */

                                while ((*eobuffer = alphabet[(int) *eobuffer]) != '\0') /* for each character of the string, convert it to lower case */
                                {
                                    eobuffer ++; /* next character */
                                }

                            }

                            if (s_flag == 0) /* disable whitespace in file, 0 = no, 1 = yes, flag set? */
                            {

                                if ((fields = strtoken (buffer, parsebuffer, token, TOKEN_SEPARATORS)) > 0) /* parse the record to remove whitespace */
                                {
                                    (void) strcpy (buffer, token[fields - 1]); /* and copy the last token */
                                }

                                else
                                {
                                    buffer[0] = '\0';
                                }

                            }

                            if (n_flag != 0 || buffer[0] != '\0') /* include null records, 0 = no, 1 = yes, flag set? */
                            {

                                if ((tretval = hash_insert (&text_table, buffer)) != NOERROR) /* add a record to the hash table */
                                {

                                    if (tretval != HASH_DUP_ERR) /* record already in hash table; this record a duplicate? */
                                    {
                                        retval = EHASHINSERT; /* error inserting record in hash table */
                                        break;
                                    }

                                    match = 1; /* match in lookup of record, 0 = no, 1 = yes */
                                }

                                else
                                {

                                    if ((element = hash_find (&text_table, buffer)) != NOERROR) /* find a hash table element with the record in the hash table */
                                    {

                                        if (element->hash_data_printed == 0) /* hash data printed, 0 = no, 1 = yes, reset? */
                                        {
                                            element->hash_data_printed =1; /* set the hash data printed, 0 = no, 1 = yes */
                                        }

                                    }

                                    else
                                    {
                                        retval = EHASHFIND; /* error finding record in hash table */
                                        break;
                                    }

                                }

                            }

                            search_count ++; /* increment the number of searches */

                            if (match > 0) /* search for matching record found? */
                            {
                                match_count ++; /* match found, increment the count of the number of matches-the number of IP addresses found in the database */

                                if (print_flag == 0) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, reset? */
                                {

                                    if ((element = hash_find (&text_table, buffer)) != NOERROR) /* find a hash table element with the record in the hash table */
                                    {

                                        if (element->hash_data_printed == 0) /* hash data printed, 0 = no, 1 = yes, reset? */
                                        {
                                            (void) printf ("%s\n", buffer); /* yes, print the record */
                                            element->hash_data_printed =1; /* set the hash data printed, 0 = no, 1 = yes */
                                        }

                                    }

                                    else
                                    {
                                        retval = EHASHFIND; /* error finding record in hash table */
                                        break;
                                    }

                                }

                            }

                            else
                            {

                                if (print_flag == 1) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, set? */
                                {

                                    if (n_flag != 0 || buffer[0] != '\0') /* include null records, 0 = no, 1 = yes, flag set? */
                                    {
                                        (void) printf ("%s\n", buffer); /* yes, print the record */
                                    }

                                }

                            }

                        }

                    }

                }

                else
                {

                    while ((fgets (buffer, BUFLEN, infile) != (char *) 0) && (retval == NOMATCH)) /* read the strings to search for from stdin, but only while there are no errors */
                    {
                        retval = EBOF; /* assume error in buffer */

                        if ((eobuffer = strchr (buffer, '\n')) != (char *) 0) /* a record with no carriage return was too long for the fgets () buffer, and an overflow condition exists; line contain a carriage return? */
                        {
                            *eobuffer = '\0'; /* null the carriage return */
                            retval = NOMATCH; /* return value, assume no error */
                            match = 0; /* match in lookup of record, 0 = no, 1 = yes */

                            if (i_flag == 0) /* case sensitive, 0 = no, 1 = yes, flag set? */
                            {
                                eobuffer = buffer; /* reference the beginning of the buffer character */

                                while ((*eobuffer = alphabet[(int) *eobuffer]) != '\0') /* for each character of the string, convert it to lower case */
                                {
                                    eobuffer ++; /* next character */
                                }

                            }

                            if (s_flag == 0) /* disable whitespace in file, 0 = no, 1 = yes, flag set? */
                            {

                                if ((fields = strtoken (buffer, parsebuffer, token, TOKEN_SEPARATORS)) > 0) /* parse the record to remove whitespace */
                                {
                                    (void) strcpy (buffer, token[fields - 1]); /* and copy the last token */
                                }

                                else
                                {
                                    buffer[0] = '\0';
                                }

                            }

                            if (n_flag != 0 || buffer[0] != '\0') /* include null records, 0 = no, 1 = yes, flag set? */
                            {

                                if ((tretval = hash_insert (&text_table, buffer)) != NOERROR) /* add a record to the hash table */
                                {

                                    if (tretval != HASH_DUP_ERR) /* record already in hash table; this record a duplicate? */
                                    {
                                        retval = EHASHINSERT; /* error inserting record in hash table */
                                        break;
                                    }

                                    match = 1; /* match in lookup of record, 0 = no, 1 = yes */
                                }

                                else
                                {

                                    if ((element = hash_find (&text_table, buffer)) != NOERROR) /* find a hash table element with the record in the hash table */
                                    {

                                        if (element->hash_data_printed == 0) /* hash data printed, 0 = no, 1 = yes, reset? */
                                        {
                                            element->hash_data_printed =1; /* set the hash data printed, 0 = no, 1 = yes */
                                        }

                                    }

                                    else
                                    {
                                        retval = EHASHFIND; /* error finding record in hash table */
                                        break;
                                    }

                                }

                            }

                            search_count ++; /* increment the number of searches */

                            if (match > 0) /* search for matching record found? */
                            {
                                match_count ++; /* match found, increment the count of the number of matches-the number of IP addresses found in the database */

                                if (print_flag == 0) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, reset? */
                                {

                                    if ((element = hash_find (&text_table, buffer)) != NOERROR) /* find a hash table element with the record in the hash table */
                                    {

                                        if (element->hash_data_printed == 0) /* hash data printed, 0 = no, 1 = yes, reset? */
                                        {
                                            (void) printf ("%s\n", buffer); /* yes, print the record */
                                            element->hash_data_printed =1; /* set the hash data printed, 0 = no, 1 = yes */
                                        }

                                    }

                                    else
                                    {
                                        retval = EHASHFIND; /* error finding record in hash table */
                                        break;
                                    }

                                }

                            }

                            else
                            {

                                if (print_flag == 1) /* verbose print flag, 0 = print string(s) that match, 1 = print string(s) that don't match, set? */
                                {

                                    if (n_flag != 0 || buffer[0] != '\0') /* include null records, 0 = no, 1 = yes, flag set? */
                                    {
                                        (void) printf ("%s\n", buffer); /* yes, print the record */
                                    }

                                }

                            }

                        }

                    }

                }

                hash_term (&text_table); /* terminate the hash table */
            }

        }

    }

    retval = (((retval == MATCH) || (retval == NOMATCH)) && (match_count > 0)) ? MATCH : retval; /* if any matches, but no errors, set the return value to MATCH */

    if ((retval == MATCH) && (all_match != 0)) /* any matches, and all matches required? */
    {

        if (match_count != search_count) /* all searches match? */
        {
            retval = NOMATCH; /* no, return no matches */
        }

    }

    print_message (retval); /* print any error messages */

    if ((error_return_change != -1) && (retval >= EARGS)) /* change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH, and error? */
    {
        retval = error_return_change; /* yes, set the error value to MATCH or NOMATCH */
    }

    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit-match found.

        2) Else if retval is unity, print nothing-normal/successful
           program exit-match not found.

        3) Else if retval is two, print help.

        4) Else retval is not zero, unity, or two, it is an error
           code, print the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, match found, print nothing */
        case 1: /* program ended without errors, match not found, print nothing */

            break;

        case 2: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

hash.c, hash table functions

static int hash_init (HASHTABLE *hash_table);
static int hash_insert (HASHTABLE *hash_table, void *data);
static HASH *hash_find (HASHTABLE *hash_table, void *data);
static int hash_delete (HASHTABLE *hash_table, void *data);
static void hash_term (HASHTABLE *hash_table);

I) the objective of the hash functions are to provide a means of
building look up tables in an expedient fashion

    A) each hash table consists of elements

        1) a hash table index, consisting of an array of hash
        elements, of type HASH

            a) the number of elements in the hash table index is
            determined by the hash table's descriptor structure
            element, hash_size

            b) each of the elements in the hash table index are the
            head elements in a doubly linked list

                i) each doubly linked list contains the hash table
                index at its head, and hash table elements, also of
                type HASH, that reference data stored in the hash
                table

        2) each element in the hash table is referenced via a key
        value

            a) there can be no assumptions made about the data, data's
            structure, or data's complexity

            b) this means that functions, unique to each hash table,
            must be implemented to:

                i) allocate any data space required

                ii) compare key values that reference the data

                iii) deallocate any data allocated

            c) references to these functions are stored in the hash
            table's descriptor structure, which has the following
            elements:

                i) hash_size is the size of the hash table

                ii) table is a reference to the hash table

                iii) mkhash is a reference to the function that
                allocates hash table elements and data space

                iv) cmphash is a reference to the function that
                compares hash table element keys

                v) rmhash is a reference to the function that
                deallocates hash table elements and data space

                vi) comphash is a reference to the function that
                computes the hash value of a key

        3) the hash table element structure, of type HASH, has the
        following elements:

            a) struct hash_struct *previous, which is a reference to
            next element in hash's doubly, circular linked list, and
            used by the hash system's internal list operations

            b) struct hash_struct *next, which is a reference to
            previous element in hash's doubly, circular linked list,
            and used by the hash system's internal list operations

            c) any collection of other elements as defined
            appropriately by the user, and can include:

                i) references to other data and data structures

                ii) numerical data, etc.

                iii) data referenced by these elements are allocated
                and deallocated by the user definable functions,
                mkhash () and rmhash (), respectively

    B) the hash table operations are performed by the following
    functions:

        1) int hash_init (HASHTABLE *hash_table), which initializes
        the hash table

        2) int hash_insert (HASHTABLE *hash_table, void *data);, which
        inserts elements, keys, and data in the hash table

        3) HASH *hash_find (HASHTABLE *hash_table, void *data), which
        fetches elements from the hash table

        4) int hash_delete (HASHTABLE *hash_table, void *data), which
        deletes elements from the hash table

        5) void hash_term (HASHTABLE *hash_table), which terminates
        use of the hash table, calling hash_delete () for each element
        remaining in the hash table

    C) All hash table functions return success or failure:

        1) functions that return an integer success/failure error code
        set the integer, hash_error to the return value of the
        function

        2) functions that return an indirect reference return
        success/failure error code (NULL corresponds to an error,) and
        set the integer, hash_error to a unique error value

        3) all hash functions set hash_error (numerical assignments
        are made in hash.h:)

            a) NOERROR if no error

            b) EALLOC if error allocating memory

            c) HASH_DUP_ERR if a duplicate key when inserting key into
            hash table

            d) HASH_MK_ERR if hash table mkhash () failure, and mkhash
            () did not set hash_error to NOERROR

            e) HASH_KEY_ERR if hash table key not found

II) static int hash_init (HASHTABLE *hash_table) initializes the hash
tables data structures, and must be called prior to any operations on
a hash table:

    A) the single argument:

        1) hash_table is a reference to the hash table descriptor
        structure

    B) the return value:

        1) returns NOERROROR if successful

        2) returns EALLOC if a memory allocation error

        3) returns HASH_INI_ERR if the hash table was already
        initialized

III) static int hash_insert (HASHTABLE *hash_table, void *data)
inserts a new hash element into a hash table:

     A) the arguments:

         1) hash_table is a reference to the hash table descriptor
         structure

         2) data is a reference to the key value

             a) this reference is passed to the key comparison routine
             and the hash element construction routines specified in
             the hash descriptor structure

     B) the return value is dependent on whether UNIQUE_KEYS was
     defined at compile time:

        1) the return value if UNIQUE_KEYS was defined at compile time
        is:

            a) returns NOERROROR if successful

            b) returns HASH_DUP_ERR if a duplicate key was found

            c) if an error occured in mkhash ()

                i) if mkhash () set hash_error to other than
                NOERROR, returns the value in hash_error

                ii) if mkhash () did not set hash_error to other than
                NOERROR, returns HASH_MK_ERR

        2) the return value if UNIQUE_KEYS was not defined at compile
        time is:

            a) returns NOERROROR if successful

                i) the element is always inserted as the first element
                in the hash index's linked list, and will thus be the
                first element found when the hash table is searched
                for the key (ie., if multiple elements exist with the
                same key in a hash table, the most recently inserted
                element will be the one found by any hash table
                search)

            b) if an error occured in mkhash ()

                i) if mkhash () set hash_error to other than
                NOERROR, returns the value in hash_error

                ii) if mkhash () did not set hash_error to other than
                NOERROR, returns HASH_MK_ERR

IV) static HASH *hash_find (HASHTABLE *hash_table, void *data)
searches the hash table for an element that matches a given key,
according to the key comparison routine specified in the hash table
descriptor structure

    A) the arguments:

        1) hash_table is a reference to the hash table descriptor

        2) data is a reference to the element's key value:

            a) this reference is passed to the key comparison routine
            specified in the hash table descriptor structure

    B) the return value:

        1) returns a reference to the hash element found in the hash
        table

        2) returns NULL if the element was not found in the hash table

V) static int hash_delete (HASHTABLE *hash_table, void *data) deletes
an element from a hash table:

    A) the arguments:

        1) hash_table is a reference to the hash table descriptor

        2) data is a reference to the element's key value:

            a) this reference is passed to the key comparison routine
            specified in the hash table descriptor structure

    B) the return value:

        1) returns NOERROROR if successful

        2) returns HASH_DEL_ERR if key not found

VI) static void hash_term (HASHTABLE *hash_table) deletes a hash
table, including all remaining elements, and data space referenced by
the elements:

    A) the single argument:

        1) hash_table is a reference to the hash table descriptor
        structure

    B) there is no return value:

VII) hash table descriptor structure:

    A) before any use of hash routines, the handler functions, mkhash
    (), cmphash (), rmhash (), and comphash (), must be specified
    along with the hash table size, hash_size, in a hash table
    descriptor of type HASHTABLE

        1) the HASHTABLE element, table, which references the hash
        table, should be initialized to zero before calling
        hash_init ()

        2) for example:

            HASHTABLE my_table =
                {2729, 0, my_mkhash, my_cmphash, my_rmhash, my_comphash};

    B) the hash table descriptor structure is defined in hash.h, with
    the following elements:

        1) hash_size, size of hash table

            a) hash_size should be a prime number for optimal
            distribution of the hash elements in the hash array

        2) table, which is a reference to hash to the table

            a) this element is, generally, used only by the hash
            algorithms, but should be initialized to zero before
            calling hash_init ()

        3) mkhash, which is a reference to the function that allocates
        hash elements and data space

            a) mkhash () creates a hash element and its allocated data
            for insertion into a hash table

            b) struct HASH *(*mkhash) (void *data)

            c) the single argument:

                i) data is the address of the key associated with the
                element

            d) the return value:

                i) the address of the element constructed

                ii) returns NULL if no element was constructed

            e) mkhash must initialize the link elements, next and
            previous, both, to reference the element on return

        4) cmphash, which is a reference to the function that compares
        element's keys

            a) cmphash () compares a key against a key associated with
            a hash table element

            b) static int (*cmphash) (void *data, HASH *element)

            c) the arguments:

                i) data is a reference to a key

                ii) element is a reference to a HASH structure to
                which the key should be compared

            d) the return value:

                i) returns 0 if data and the key associated with the
                element key are equal

                ii) returns non-zero if data and the key associated
                with the element are not equal

        5) rmhash, which is a reference to the function that
        deallocates hash elements and data space

            a) rmhash () is called to delete a hash table element and
            its allocated data from a hash table-given the address of
            the element created by mkhash (), it should reverse
            operations of mkhash ()

            b) void (*rmhash) (HASH *element)

            c) the single argument:

                i) element is a reference to the hash table element to
                delete

            d) there is no return value

        6) comphash, which is reference to the function that computes
        the hash value of a key

            a) comphash () is called to compute the hash value of a
            key

            b) int (*comphash) (void *key)

            c) the arguments:

                i) hash_table is a reference to the hash table
                descriptor

                ii) key is a reference to the key to be hashed

            d) the return value is the key's hash

VIII) performance issues:

    A) note that the number of comparisons, ch, required for a key
    search in the hash table is (on average):

        ch = (1 / 2) * (n / hash_size);

    where n is the number of keys in the hash table, and hash_size is
    the size of the hash table

    B) by comparison, the number of comparisons, cb, required for a
    key search using a binary search routine is (on average):

        cb = (1 / 2) * (log (n) / log (2));

    where it is assumed that an each key compared has an equal
    probability of being the key that is being searched for (probably
    an overly optimistic assumption)

    C) for a similar number of comparisons:

        hash_size = ((n * log (2)) / log (n));

    D) in powers of 2, the hash table size, hash_size, should be:

        N               C               I               P

        2               1               2               2
        4               2               2               2
        8               3               2               2
        16              4               4               5
        32              5               6               7
        64              6               10              11
        128             7               18              19
        256             8               32              37
        512             9               56              59
        1024            10              102             103
        2048            11              186             191
        4096            12              341             347
        8192            13              630             631
        16384           14              1170            1171
        32768           15              2184            2203
        65536           16              4096            4099
        131072          17              7710            7717
        262144          18              14563           14563
        524288          19              27594           27611
        1048576         20              52428           52433
        2097152         21              99864           99871
        4194304         22              190650          190657
        8388608         23              364722          364739
        16777216        24              699050          699053
        33554432        25              1342177         1342177
        67108864        26              2581110         2581121
        134217728       27              4971026         4971037
        268435456       28              9586980         9586981
        536870912       29              18512790        18512807
        1073741824      30              35791394        35791397
        2147483648      31              69273666        69273719

        where:

        1) N is the number of keys in the lookup table

        2) C is the maximum number of key comparisons (minus 1)
        required to locate a key, eg., floor (log (N) / log (2)

        3) I is the size of the hash table index, hash_size,
        floor (N / (log (N) / log (2)))

        4) P is the next larger prime than I

IX) data structure (with exactly one data space object allocated,
referenced by the HASH structure element, hash_data:)

    hash index array, an array of HASHTABLE->hash_size many HASH structures
+--------------------------------------------------------------------------
|
|
|
+-->typedef struct hash_struct
    {
        struct hash_struct *previous;
        void *hash_data;
        struct hash_struct *next;
    } HASH;
    typedef struct hash_struct
    {
        struct hash_struct *previous;
        void *hash_data;
        struct hash_struct *next;
    } HASH;
                      .
                      .
                      .
+-----------------------------------------------------------------------------+
|                                                                             |
|                                                                             |
+-->typedef struct hash_struct         +-->typedef struct hash_struct         |
|   {                                  |   {                                  |
|       struct hash_struct *previous;  |       struct hash_struct *previous;--+
|       void *hash_data; (null)        |       void *hash_data;--------------+
|       struct hash_struct *next;------+       struct hash_struct *next;--+  |
|   } HASH;                                } HASH;                        |  |
|                                                                         |  |
|                                                                         |  |
+-------------------------------------------------------------------------+  |
                      .                                                      |
                      .                                                      |
                      .                                                      |
    typedef struct hash_struct         +-------------------------------------+
    {                                  |
        struct hash_struct *previous;  |
        void *hash_data;               +-->[object data area]
        struct hash_struct *next;
    } HASH;
    typedef struct hash_struct
    {
        struct hash_struct *previous;
        void *hash_data;
        struct hash_struct *next;
    } HASH;

XII) the hash table index size, hash_size, should be a prime number,
(for best operation,); the following program will list the prime
numbers from 2 to the value of the command line argument:

    #include <stdio.h>
    #include <stdlib.h>

    int main (int argc,char *argv[])
    {
        int i,
            j,
            prime_flag,
            max_num;

        if (argc < 2)
        {
            (void) printf ("no maximum number\n");
            exit (1);
        }

        if ((max_num = atoi (argv[1])) < 2)
        {
            (void) printf ("maximum number must be greater than 2");
            exit (2);
        }

        for (i = 2; i <= max_num; i ++)
        {
            prime_flag = 1;

            for (j = 2; j < i; j ++)
            {

                if (i % j == 0)
                {
                    prime_flag = 0;
                    break;
                }

            }

            if (prime_flag == 1)
            {
                (void) printf ("%d\n",i);
            }

        }

        exit (0);
    }

XI) compile time issues:

    A) if UNIQUE_KEYS is defined, multiple keys will not be permitted
    in the hash table

    B) if UNIQUE_KEYS is not defined, multiple keys will be permitted
    in the hash table, and hash_find (), and hash_delete () will
    always operate on the most recent key inserted by hash_insert (),
    (thus operating like a "stack")

*/

static int hash_error = 0; /* hash success/failure error code */

static HASH *temp_hash; /* temporary HASH reference in SWAP() */

#define SWAP(x,y) temp_hash = (x);(x) = (y);(y) = temp_hash /* linked list link element manipulator */

/*

static int hash_init (HASHTABLE *hash_table);

initialize the hash table: if this function has not been executed for
this HASHTABLE structure, allocate hash_table->hash_size many HASH
structures for the hash table index, initialize each element in the
hash table index's next and previous references to reference the
element itself

the required argument is a reference to the hash table descriptor
structure, hash_table

returns NOERROROR if successful, EALLOC if a memory
allocation error, and HASH_INI_ERR if the hash table was already
initialized

*/

#ifdef __STDC__

static int hash_init (HASHTABLE *hash_table)

#else

static int hash_init (hash_table)
HASHTABLE *hash_table;

#endif

{
    size_t i; /* hash table element counter */

    HASH *hash_ref; /* reference to element in hash table index */

    hash_error = HASH_INI_ERR; /* assume hash table already initilaized error */

    if (hash_table->table == (HASH *) 0) /* protect from repeated calls by testing if hash table index has been allocated */
    {
        hash_error = EALLOC; /* assume memory error */

        if ((hash_table->table = (HASH *) malloc (hash_table->hash_size * sizeof (HASH))) != (HASH *) 0) /* allocate the hash table index */
        {
            hash_error = NOERROR; /* assume no error */

            for (i = 0; i < hash_table->hash_size; i ++) /* initialize the hash table doubly, circular linked list */
            {
                hash_ref = &hash_table->table[i]; /* reference the element in the hash table index */
                hash_ref->next = hash_ref; /* references its own element */
                hash_ref->previous = hash_ref; /* references its own element */
            }

        }


    }

    return (hash_error); /* return any error */
}

#ifdef UNIQUE_KEYS

/*

int hash_insert (HASHTABLE *hash_table, void *data);

insert a key and data into the hash table: calculate the key's hash,
search the hash table index implicitly addressed by the key's hash for
a matching key, if the key is not found, insert the key and data at
the end of the doubly linked list for this hash table index element

the required arguments are a reference to the hash table descriptor
structure, hash_table, and a reference to the key value, data

returns NOERROROR if successful, HASH_DUP_ERR if a duplicate key
was found, HASH_MK_ERR if an error occured in mkhash () and hash_error
was not set by mkhash (), else the value of hash_error that was set by
mkhash ()

*/

#ifdef __STDC__

static int hash_insert (HASHTABLE *hash_table, void *data)

#else

static int hash_insert (hash_table, data)
HASHTABLE *hash_table;
void *data;

#endif

{
    int hindex; /* index of element in hash table */

    HASH *element, /* new element to be inserted in hash table */
         *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *last_element, /* reference to last element in hash index's doubly, circular linked list */
         *head_element; /* reference to head element in hash index's doubly, circular linked list */

    hash_error= NOERROR; /* assume no error */

    hindex = (*hash_table->comphash) (hash_table, data); /* get the data's hash */
    head_element = next_element =&hash_table->table[hindex]; /* reference the head element in hash index's doubly, circular linked list */

    while ((next_element = next_element->next) != head_element) /* while the element in the hash index's doubly, circular linked list is not the list's head */
    {

        if ((*hash_table->cmphash) (data, next_element) == 0) /* element compare with the data ? */
        {
            hash_error = HASH_DUP_ERR; /* yes, data is already in hash table */
            break;
        }

    }

    if (hash_error == NOERROR) /* any errors? */
    {
        hash_error = HASH_MK_ERR; /* no, assume mkhash () error */

        if ((element = (*hash_table->mkhash) (data)) != (HASH *) 0) /* no, get the hash element and data space */
        {
            last_element = next_element->previous; /* reference the last element in the doubly, circular linked list */
            SWAP (last_element->next, element->next); /* the new element is last in the doubly, circular linked list */
            SWAP (next_element->previous, element->previous);
            hash_error = NOERROR; /* assume no error */
        }

    }

    return (hash_error); /* return any error */
}

#else

/*

static int hash_insert (HASHTABLE *hash_table, void *data);

insert a key into the hash table: calculate the key's hash, insert the
key and data at the beginning of the doubly linked list for this hash
table index element, which is implicitly addressed by the key's hash

the required arguments are a reference to the hash table descriptor
structure, hash_table, and a reference to the key value, data

returns NOERROROR if successful, HASH_MK_ERR if an error occured
in mkhash () and hash_error was not set by mkhash (), else the value
of hash_error that was set by mkhash ()

*/

#ifdef __STDC__

static int hash_insert (HASHTABLE *hash_table, void *data)

#else

static int hash_insert (hash_table, data)
HASHTABLE *hash_table;
void *data;

#endif

{
    int hindex; /* index of element in hash table */

    HASH *element, /* reference to new element to be inserted in hash table */
         *next_element, /* reference to next element in hash's doubly, circular linked list */
         *previous_element; /* reference to indexed element in the hash table */

    hash_error = HASH_MK_ERR; /* assume mkhash () error */

    hindex = (*hash_table->comphash) (hash_table, data); /* get the data's hash */
    previous_element = &hash_table->table[hindex]; /* reference the element in the hash table index */
    next_element = previous_element->next; /* start with the first element in the hash's doubly, circular linked list */

    if ((element = (*hash_table->mkhash) (data)) != (HASH *) 0) /* get the hash element */
    {
        SWAP (previous_element->next, element->next); /* the new element is first in the doubly, circular linked list */
        SWAP (next_element->previous, element->previous);
        hash_error = NOERROR; /* assume no error */
    }

    return (hash_error); /* return any error */
}

#endif

/*

static HASH *hash_find (HASHTABLE *hash_table, data *data);

find data in the hash table: calculate the key's hash, search for the
key and data, starting at the beginning of the doubly linked list for
this hash table index element, which is implicitly addressed by the
key's hash

the required arguments are a reference to the hash table descriptor
structure, hash_table, and a reference to the key value, data

returns a reference to the hash table's element referencing the data,
0 if key is not found

*/

#ifdef __STDC__

static HASH *hash_find (HASHTABLE *hash_table, void *data)

#else

static HASH *hash_find (hash_table, data)
HASHTABLE *hash_table;
void *data;

#endif

{
    int hindex; /* index of element in hash table */

    HASH *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *head_element, /* reference to head element in hash index's doubly, circular linked list */
         *element = (HASH *) 0; /* return value, assume error */

    hash_error = HASH_KEY_ERR; /* assume key not found error */

    hindex = (*hash_table->comphash) (hash_table, data); /* get the data's hash */
    head_element = next_element = &hash_table->table[hindex]; /* reference the head element in hash index's doubly, circular linked list */

    while ((next_element = next_element->next) != head_element) /* while the element in the hash index's doubly, circular linked list is not the list's head */
    {

        if ((*hash_table->cmphash) (data,next_element) == 0) /* element compare with the data ? */
        {
            element = next_element; /* yes, found it, return a reference to the element */
            hash_error = NOERROR; /* assume no error */
            break;
        }

    }

    return (element); /* return any error */
}

#ifdef NOT_USED

/*

static int hash_delete (HASHTABLE *hash_table, void *data);

delete data from the hash table: calculate the key's hash, search for
the key and data, starting at the beginning of the doubly linked list
for this hash table index element, which is implicitly addressed by
the key's hash, if found, remove the key's HASH element, and data

the required arguments are a reference to the hash table descriptor
structure, hash_table, and a reference to the key value, data

returns NOERROROR if successful, HASH_DEL_ERR if key not found

*/

#ifdef __ANSI__

static int hash_delete (HASHTABLE *hash_table,void *data)

#else

static int hash_delete (hash_table,data)
    HASHTABLE *hash_table;
    void *data;

#endif

{
    int hindex; /* index of element in hash table */

    HASH *element, /* reference to element to be deleted from hash table */
         *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *previous_element, /* reference to previous element in hash index's doubly, circular linked list */
         *head_element; /* reference to head element in hash index's doubly, circular linked list */

    hash_error = HASH_KEY_ERR; /* assume key not found error */

    hindex = (*hash_table->comphash) (hash_table, data); /* get the data's hash */
    head_element = element = &hash_table->table[hindex]; /* reference the head element in hash index's doubly, circular linked list */

    while ((element = element->next) != head_element) /* while the element in the hash index's doubly, circular linked list is not the list's head */
    {

        if ((*hash_table->cmphash) (data,element) == 0) /* element compare with the data ? */
        {
            next_element = element->next; /* yes, reference the next element in the hash's doubly, circular linked list */
            previous_element = element->previous; /* reference the previous element in the hash's doubly, circular linked list */
            SWAP (next_element->previous,element->previous); /* delete the element from the hash's doubly, circular linked list */
            SWAP (previous_element->next,element->next);
            (*hash_table->rmhash) (element); /* delete the element, and data referenced by the element */
            hash_error = NOERROR; /* found it, assume no error */
            break;
        }

    }

    return (hash_error); /* return any error */
}

#endif

/*

static void hash_term (HASHTABLE *hash_table);

remove a hash table: for each element in the hash table's index, for
each element in the doubly linked lest referenced by the hash table's
index element, remove the key's HASH element

the required argument is a reference to the hash table descriptor
structure, hash_table

always returns NOERROR

*/

#ifdef __STDC__

static void hash_term (HASHTABLE *hash_table)

#else

static void hash_term (hash_table)
    HASHTABLE *hash_table;

#endif

{
    size_t i;

    HASH *element, /* reference to element being deleted from hash index's doubly, circular linked list */
         *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *head_element; /* reference to head element in hash index's doubly, circular linked list */

    hash_error = NOERROR; /* assume no error */

    for (i = 0; i < hash_table->hash_size && hash_error == NOERROR; i ++) /* for each of the hash table's index's (while there are no errors) */
    {
        head_element = &hash_table->table[i]; /* reference the head element in hash index's doubly, circular linked list */
        next_element = head_element->next; /* start with the first element in the hash index's doubly, circular linked list */

        while (next_element != head_element) /* while the element in the hash's doubly, circular linked list is not the list's head */
        {
            element = next_element; /* reference the element to be deleted from the doubly, circular linked list */
            next_element = next_element->next; /* reference next element in the hash's doubly, circular linked list */
            SWAP (next_element->previous,element->previous); /* delete the element from the hash's doubly, circular linked list */
            SWAP (head_element->next,element->next);
            (*hash_table->rmhash) (element); /* delete the element, and data referenced by the element */
        }

    }

    free (hash_table->table); /* no, free the hash table index */
}

/*

static int hash_text (HASHTABLE *hash_table, void *key);

compute the hash value for a text key

the routine requires a long to be 32 bits, which represents some
portability issues

the required arguments are a reference to the hash table descriptor
structure, hash_table, and a reference to the key value, key

returns the hash value for the key

*/

#ifdef __STDC__

static int hash_text (HASHTABLE *hash_table, void *key)

#else

static int hash_text (hash_table, key)
    HASHTABLE *hash_table;
    void *key;

#endif

{
    char *char_ref; /* byte reference into data */

    unsigned num,
             hash_num = (unsigned) 0; /* hash number is initially zero */

    for (char_ref = key; *char_ref != (char) '\0'; char_ref ++) /* for each character in the key */
    {
        hash_num = (hash_num << 4) + (unsigned char) *char_ref; /* multiply the hash by 16, and add the value of the character */

        /*

        In 32 bit machines, 4026531840 = 0xf0000000, for ANSI unsigned
        v.s. K&R signed declaration compatibility; the equivalent
        construct is:

        if ((num = hash_num & 0xf0000000) > 0)
        {
        }

        where the value being shifted must be unsigned, as per ANSI
        specification on bit-wise shift operators-it should work on
        any machine of 32 bit, or greater, unsigned integer values.

        */

        if ((num = ((hash_num >> 28) & 0xf) << 28) > 0) /* any high order bits set? */
        {
            hash_num = hash_num ^ (num >> 24); /* yes, divide by 16777216, and exor on itself */
            hash_num = hash_num ^ num; /* reset the high order bits to 0 */
        }

    }

    return ((int) (hash_num % (unsigned int) hash_table->hash_size)); /* return the hash number */
}

#ifdef NOT_USED

/*

static int hash_integer (HASHTABLE *hash_table, void *key);

compute the hash value for a integer key

the required arguments are a reference to the hash table descriptor
structure, hash_table, and a reference to the key value, key

returns the hash value for the key

*/

#ifdef __STDC__

static int hash_integer (HASHTABLE *hash_table, void *key)

#else

static int hash_integer (hash_table, key)
    HASHTABLE *hash_table;
    void *key;

#endif

{
    return (*((int *) key) % (int) hash_table->hash_size); /* return the hash number */
}

#endif

#ifdef QA

/*

simple routines to exercise a hash table, with keys of character data

the synopsis is:

    hash first_file [second_file] [third_file]

where the files consist of lists of words (unique if UNIQUE_KEYS is
defined,) for example /usr/lib/dict/words

if less than 3 files are specified, then the last file name will be
replicated to make 3 files

the program simply loads the words from first_file into a hash table,
then searches the hash table for the words in the second file, then
deletes the words from the hash table in the third file, then removes
the hash table and all associated data-the key and the data stored are
the same

then the program, using the count of records from the first file,
loads integers into a hash table, then searches the hash table for the
integers, then deletes the integers from the hash table, then removes
the hash table

finally, the program, using the count of records from the first file,
allocates data, of less than MAX_PTR_ARRAY bytes each, and loads
pointers to the allocated data into a hash table, then searches the
hash table for the pointers, then deletes the pointers from the hash
table, then removes that hash table

the last two tests make sparse arrays of integers and pointers,
respectively

*/

#ifdef __STDC__

static int text_cmphash (void *data, HASH *element); /* function to compare a text key with a hash table's element key */
static HASH *text_mkhash (void *data); /* function to allocate a text hash table element and data */
static void text_rmhash (HASH *element); /* function to deallocate a text hash table element and data */

static int int_cmphash (void *data, HASH *element); /* function to compare an integer key with a hash table's element key */
static HASH *int_mkhash (void *data); /* function to allocate an integer hash table element */
static void int_rmhash (HASH *element); /* function to deallocate an integer hash table element */

static int ptr_cmphash (void *data, HASH *element); /* function to compare an pointer key with a hash table's element key */
static HASH *ptr_mkhash (void *data); /* function to allocate an pointer hash table element */
static void ptr_rmhash (HASH *element); /* function to deallocate an pointer hash table element */
static int hash_pointer (HASHTABLE *hash_table, void *key); /* function to compute the hash value of an pointer key */

#else

static int text_cmphash (); /* function to compare a text key with a hash table's element key */
static HASH *text_mkhash (); /* function to allocate a text hash table element and data */
static void text_rmhash (); /* function to deallocate a text hash table element and data */

static int int_cmphash (); /* function to compare an integer key with a hash table's element key */
static HASH *int_mkhash (); /* function to allocate an integer hash table element */
static void int_rmhash (); /* function to deallocate an integer hash table element */

static int ptr_cmphash (); /* function to compare an pointer key with a hash table's element key */
static HASH *ptr_mkhash (); /* function to allocate an pointer hash table element */
static void ptr_rmhash (); /* function to deallocate an pointer hash table element */
static int hash_pointer (); /* function to compute the hash value of an pointer key */

#endif

static HASHTABLE text_table = {2729, (HASH *) 0, text_mkhash, text_cmphash, text_rmhash, hash_text}; /* declare the hash table descriptor for text keys*/
static HASHTABLE integer_table = {2729, (HASH *) 0, int_mkhash, int_cmphash, int_rmhash, hash_integer}; /* declare the hash table descriptor for integer keys */
static HASHTABLE pointer_table = {2729, (HASH *) 0, ptr_mkhash, ptr_cmphash, ptr_rmhash, hash_pointer}; /* declare the hash table descriptor for pointer keys */

#define MAX_PTR_ARRAY 100 /* maximum size of arrays allocated in test of pointer keys */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char data[BUFSIZ], /* I/O buffer for fgets () */
         *file_name[3]; /* references to file names */

    int ctr, /* record counter */
        number; /* number of records in input file */

    void *mem_ref, /* memory reference */
         **ptr_array; /* reference to array of pointers */

    HASH *element; /* reference to hash table element */

    FILE *infile; /* reference to input file */

    switch (argc) /* how many arguments? */
    {

        case 1: /* not enough */

            (void) fprintf (stderr, "usage: %s first_file [second_file] [third_file]\n", argv[0]); /* print help, and exit */
            exit (1);
            break;

        case 2: /* one, use if for the second and third file namem, also */

            file_name[0] = argv[1];
            file_name[1] = argv[1];
            file_name[2] = argv[1];
            break;

        case 3: /* two, use the last for the third file name, also */

            file_name[0] = argv[1];
            file_name[1] = argv[2];
            file_name[2] = argv[2];
            break;

        default: /* three, or more */

            file_name[0] = argv[1];
            file_name[1] = argv[2];
            file_name[2] = argv[3];
            break;

    }

    if (hash_init (&text_table) != NOERROR) /* initialize the hash table */
    {
        (void) fprintf (stderr, "error initializing text hash table (error code %d)\n", hash_error);
        exit (1);
    }

    if ((infile = fopen (file_name[0], "r")) == (FILE *) 0) /* open the first input file */
    {
        (void) fprintf (stderr, "error opening file, %s\n", file_name[0]);
        exit (1);
    }

    ctr = 0; /* reset the record counter */
    (void) printf ("loading text hash array\n");

    while (fgets (data, BUFSIZ, infile) == data) /* for each record in the file */
    {
        ctr ++; /* increment the record counter */

        if (hash_insert (&text_table, data) != NOERROR) /* add a key, data, to the hash table-the key and the data stored are the same */
        {
            (void) fprintf (stderr, "error inserting text key into hash table, %s (error code %d)\n", data, hash_error); /* could not add a key, data, to the hash table, print the error */
        }

    }

    if (fclose (infile) != 0) /* close the first input file */
    {
        (void) fprintf (stderr, "error closing file, %s\n", file_name[0]); /* couldn't close the input file, print the error and exit */
        exit (1);
    }

    number = ctr; /* save the record count */
    (void) printf ("finished loading text hash array with %d elements\n", ctr);

    if ((infile = fopen (file_name[1], "r")) == (FILE *) 0) /* open the second input file */
    {
        (void) fprintf (stderr, "error opening file, %s\n", file_name[1]);
        exit (1);
    }

    ctr = 0; /* reset the record counter */
    (void) printf ("finding elements in text hash array\n");

    while (fgets (data, BUFSIZ, infile) == data) /* for each record in the file */
    {
        ctr ++; /* increment the record counter */

        if ((element = hash_find (&text_table, data)) == NOERROR) /* find a hash table element with a key, data, in the hash table */
        {
            (void) fprintf (stderr, "error finding text key, %s (error code %d)\n", data, hash_error); /* couldn't find a hash table element with a key, data, in the hash table, print the error */
        }

        else
        {

            if (strcmp (data, element->hash_data) != 0) /* compare the key with the stored data */
            {
                (void) fprintf (stderr, "error comparing text key %s with %s", data, (char *) element->hash_data); /* couldn't compare the key with the stored data, print the error */
            }

        }

    }

    if (fclose (infile) != 0) /* close the second input file */
    {
        (void) fprintf (stderr, "error closing file, %s\n", file_name[1]); /* couldn't close the input file, print the error and exit */
        exit (1);
    }

    (void) printf ("finished finding elements in text hash array with %d elements\n", ctr);

    if ((infile = fopen (file_name[2], "r")) == (FILE *) 0) /* open the third input file */
    {
        (void) fprintf (stderr, "error opening file, %s\n", file_name[2]);
        exit (1);
    }

    ctr = 0; /* reset the record counter */
    (void) printf ("deleting elements in text hash array\n");

    while (fgets (data, BUFSIZ, infile) == data) /* for each record in the file */
    {
        ctr ++; /* increment the record counter */

        if (hash_delete (&text_table, data) != NOERROR) /* delete a key and data from the hash table */
        {
            (void) fprintf (stderr, "error deleting text key %s (error code %d)\n", data, hash_error); /* couldn't delete a key and data from the hash table, print the error */
        }

    }

    (void) printf ("finished deleting elements in text hash array with %d elements\n", ctr);

    if (fclose (infile) != 0) /* close the third input file */
    {
        (void) fprintf (stderr, "error closing file, %s\n", file_name[2]); /* couldn't close the input file, print the error and exit */
        exit (1);
    }

    hash_term (&text_table); /* terminate the hash table */

    if (hash_init (&integer_table) != NOERROR) /* initialize the hash table */
    {
        (void) fprintf (stderr, "error initializing integer hash table (error code %d)\n", hash_error);
        exit (1);
    }

    (void) printf ("loading integer hash array\n");

    for (ctr = 0; ctr < number; ctr ++) /* for each number*/
    {

        if (hash_insert (&integer_table, &ctr) != NOERROR) /* add a key to the hash table */
        {
            (void) fprintf (stderr, "error inserting integer into hash table, %d (error code %d)\n", ctr, hash_error); /* could not add a key to the hash table, print the error */
        }

    }

    (void) printf ("finished loading integer hash array with %d elements\n", ctr);
    (void) printf ("finding elements in integer hash array\n");

    for (ctr = 0; ctr < number; ctr ++) /* for each number*/
    {

        if ((element = hash_find (&integer_table, &ctr)) == NOERROR) /* find a hash table element with a key in the hash table */
        {
            (void) fprintf (stderr, "error finding integer key, %d (error code %d)\n", ctr, hash_error); /* couldn't find a hash table element with a key in the hash table, print the error */
        }

        else
        {

            if (ctr != element->integer) /* compare the key with the stored data */
            {
                (void) fprintf (stderr, "error comparing integer key %d with %d", ctr, element->integer); /* couldn't compare the key with the stored data, print the error */
            }

        }

    }

    (void) printf ("finished finding elements in integer hash array with %d elements\n", ctr);
    (void) printf ("deleting elements in integer hash array\n");

    for (ctr = 0; ctr < number; ctr ++) /* for each number*/
    {

        if (hash_delete (&integer_table, &ctr) != NOERROR) /* delete a key from the hash table */
        {
            (void) fprintf (stderr, "error deleting integer key %d (error code %d)\n", ctr, hash_error); /* couldn't delete a key from the hash table, print the error */
        }

    }

    (void) printf ("finished deleting elements in integer hash array with %d elements\n", ctr);
    hash_term (&integer_table); /* terminate the hash table */

    if ((ptr_array = (void **) malloc (number * sizeof (void *))) == (void **) 0) /* allocate the array of pointers */
    {
        (void) fprintf (stderr, "error allocating memory\n"); /* couldn't allocate the array of pointers, print the error and exit */
        exit (1);
    }

    if (hash_init (&pointer_table) != NOERROR) /* initialize the hash table */
    {
        (void) fprintf (stderr, "error initializing pointer hash table (error code %d)\n", hash_error);
        exit (1);
    }

    (void) printf ("loading pointer hash array\n");

    for (ctr = 0; ctr < number; ctr ++) /* for each number*/
    {

        if ((mem_ref = malloc ((ctr % MAX_PTR_ARRAY) * sizeof (char))) == (void *) 0) /* allocate the data area */
        {
            (void) fprintf (stderr, "error allocating memory\n"); /* couldn't allocate the data area, print the error and exit */
            exit (1);
        }

        if (hash_insert (&pointer_table, mem_ref) != NOERROR) /* add a key to the hash table */
        {
            (void) fprintf (stderr, "error inserting pointer into hash table, %d (error code %d)\n", (int) mem_ref, hash_error); /* could not add a key to the hash table, print the error */
        }

        ptr_array[ctr] = mem_ref; /* save a copy of the key */
    }

    (void) printf ("finished loading pointer hash array with %d elements\n", ctr);
    (void) printf ("finding elements in pointer hash array\n");

    for (ctr = 0; ctr < number; ctr ++) /* for each number*/
    {

        if ((element = hash_find (&pointer_table, ptr_array[ctr])) == NOERROR) /* find a hash table element with a key in the hash table */
        {
            (void) fprintf (stderr, "error finding pointer key, %d (error code %d)\n", (int) ptr_array[ctr], hash_error); /* couldn't find a hash table element with a key in the hash table, print the error */
        }

        else
        {

            if (ptr_array[ctr] != element->pointer) /* compare the key with the stored data */
            {
                (void) fprintf (stderr, "error comparing pointer key %d with %d", (int) ptr_array[ctr], (int) element->pointer); /* couldn't compare the key with the stored data, print the error */
            }

            free (ptr_array[ctr]); /* free the allocated data area */
        }

    }

    (void) printf ("finished finding elements in pointer hash array with %d elements\n", ctr);
    (void) printf ("deleting elements in pointer hash array\n");

    for (ctr = 0; ctr < number; ctr ++) /* for each number*/
    {

        if (hash_delete (&pointer_table, ptr_array[ctr]) != NOERROR) /* delete a key from the hash table */
        {
            (void) fprintf (stderr, "error deleting pointer key %d (error code %d)\n", (int) ptr_array[ctr], hash_error); /* couldn't delete a key from the hash table, print the error */
        }

    }

    (void) printf ("finished deleting elements in pointer hash array with %d elements\n", ctr);
    hash_term (&pointer_table); /* terminate the hash table */

    exit (0);
    return (0); /* for formality */
}

/*

static int text_cmphash (void *data, HASH *element);

function that compares a text key with a hash table element's text key

(note that the data reference in HASH is a void reference, and
requires a cast to the appropriate data type)

returns 0 if data and the key associated with the element key are
equal

returns non-zero if data and the key associated with the element are
not equal

*/

#ifdef __STDC__

static int text_cmphash (void *data, HASH *element)

#else

static int text_cmphash (data, element)
void *data;
HASH *element;

#endif

{
    return (strcmp (data, ((char *) element->hash_data))); /* return the value of the comparison */
}

/*

static HASH *text_mkhash (void *data);

function that allocates and loads a hash table element, and allocates
a data (which in this simple case, is simply a copy of the key)

returns a reference to the element constructed if successful

returns NULL if no element was constructed

Note: hash_error, is set to EALLOC if a memory allocation error
occured

*/

#ifdef __STDC__

static HASH *text_mkhash (void *data)

#else

static HASH *text_mkhash (data)
void *data;

#endif

{
    void *obj_ref; /* reference to data */

    HASH *element = (HASH *) 0; /* reference to the hash table element */

    hash_error = EALLOC; /* assume memory error */

    if ((obj_ref = (char *) malloc ((1 + strlen (data)) * sizeof (char))) != (char *) 0) /* allocate the hash table element key's data area */
    {

        if ((element = (HASH *) malloc (sizeof (HASH))) != (HASH *) 0) /* allocate the hash table element */
        {
            element->next = element; /* all of the elements link references reference the element itself */
            element->previous = element;
            element->hash_data = obj_ref; /* reference to element's data */
            (void) strcpy (obj_ref, data); /* save the hash table element's key as its data */
            hash_error = NOERROR; /* assume no error */
        }

    }

    return (element); /* return a reference to the hash table element */
}

/*

static void text_rmhash (HASH *element);

function to deallocate a text hash table element

returns nothing

*/

#ifdef __STDC__

static void text_rmhash (HASH *element)

#else

static void text_rmhash (element)
HASH *element;

#endif

{
    free (element->hash_data); /* free the key's data area allocated in text_mkhash () */
    free (element); /* free the hash table element allocated in text_mkhash () */
}

/*

static int int_cmphash (void *data, HASH *element);

function that compares an integer key with a hash table element's
integer key

(note that the data reference in HASH is a void reference, and
requires a cast to the appropriate data type)

returns 0 if data and the key associated with the element key are
equal

returns non-zero if data and the key associated with the element are
not equal

*/

#ifdef __STDC__

static int int_cmphash (void *data, HASH *element)

#else

static int int_cmphash (data, element)
void *data;
HASH *element;

#endif

{
    return (element->integer - *((int *) data)); /* return the value of the comparison */
}

/*

static HASH *int_mkhash (void *data);

function that allocates and loads a hash table element with an integer
key value

returns a reference to the element constructed if successful

returns NULL if no element was constructed

Note: hash_error, is set to EALLOC if a memory allocation error
occured

*/

#ifdef __STDC__

static HASH *int_mkhash (void *data)

#else

static HASH *int_mkhash (data)
void *data;

#endif

{
    HASH *element = (HASH *) 0; /* reference to the hash table element */

    hash_error = EALLOC; /* assume memory error */

    if ((element = (HASH *) malloc (sizeof (HASH))) != (HASH *) 0) /* allocate the hash table element */
    {
        element->next = element; /* all of the elements link references reference the element itself */
        element->previous = element;
        element->integer = *((int *) data); /* save the integer */
        hash_error = NOERROR; /* assume no error */
    }

    return (element); /* return a reference to the hash table element */
}

/*

static void int_rmhash (HASH *element);

function to deallocate an integer hash table element

returns nothing

*/

#ifdef __STDC__

static void int_rmhash (HASH *element)

#else

static void int_rmhash (element)
HASH *element;

#endif
{
    free (element); /* free the hash table element allocated in int_mkhash () */
}

/*

static int ptr_cmphash (void *data, HASH *element);

function that compares a pointer key with a hash table element's
pointer key

returns 0 if data and the key associated with the element key are
equal

returns non-zero if data and the key associated with the element are
not equal

*/

#ifdef __STDC__

static int ptr_cmphash (void *data, HASH *element)

#else

static int ptr_cmphash (data, element)
void *data;
HASH *element;

#endif

{
    return ((char *) element->pointer - (char *) data); /* return the value of the comparison */
}

/*

static HASH *ptr_mkhash (void *data);

function that allocates and loads a hash table element with a pointer
key value

returns a reference to the element constructed if successful

returns NULL if no element was constructed

Note: hash_error, is set to EALLOC if a memory allocation error
occured

*/

#ifdef __STDC__

static HASH *ptr_mkhash (void *data)

#else

static HASH *ptr_mkhash (data)
void *data;

#endif

{
    HASH *element = (HASH *) 0; /* reference to the hash table element */

    hash_error = EALLOC; /* assume memory error */

    if ((element = (HASH *) malloc (sizeof (HASH))) != (HASH *) 0) /* allocate the hash table element */
    {
        element->next = element; /* all of the elements link references reference the element itself */
        element->previous = element;
        element->pointer = data; /* save the pointer */
        hash_error = NOERROR; /* assume no error */
    }

    return (element); /* return a reference to the hash table element */
}

/*

static void ptr_rmhash (HASH *element);

function to deallocate an pointer hash table element

returns nothing

*/

#ifdef __STDC__

static void ptr_rmhash (HASH *element)

#else

static void ptr_rmhash (element)
HASH *element;

#endif
{
    free (element); /* free the hash table element allocated in ptr_mkhash () */
}

/*

static int hash_pointer (HASHTABLE *hash_table, void *key);

compute the hash value for a pointer key

the required arguments are a reference to the hash table descriptor
structure, hash_table, and a reference to the key value, key

returns the hash value for the key

*/

#ifdef __STDC__

static int hash_pointer (HASHTABLE *hash_table, void *key)

#else

static int hash_pointer (hash_table, key)
    HASHTABLE *hash_table;
    void *key;

#endif

{
    return (((char *) key - (char *) 0) % (int) hash_table->hash_size); /* return the hash number */
}

#endif

/*

static int text_cmphash (void *data, HASH *element);

function that compares a text key with a hash table element's text key

(note that the data reference in HASH is a void reference, and
requires a cast to the appropriate data type)

returns 0 if data and the key associated with the element key are
equal

returns non-zero if data and the key associated with the element are
not equal

*/

#ifdef __STDC__

static int text_cmphash (void *data, HASH *element)

#else

static int text_cmphash (data, element)
void *data;
HASH *element;

#endif

{
    return (strcmp (data, ((char *) element->hash_data))); /* return the value of the comparison */
}

/*

static HASH *text_mkhash (void *data);

function that allocates and loads a hash table element, and allocates
a data (which in this simple case, is simply a copy of the key)

returns a reference to the element constructed if successful

returns NULL if no element was constructed

Note: hash_error, is set to EALLOC if a memory allocation error
occured

*/

#ifdef __STDC__

static HASH *text_mkhash (void *data)

#else

static HASH *text_mkhash (data)
void *data;

#endif

{
    void *obj_ref; /* reference to data */

    HASH *element = (HASH *) 0; /* reference to the hash table element */

    hash_error = EALLOC; /* assume memory error */

    if ((obj_ref = (char *) malloc ((1 + strlen (data)) * sizeof (char))) != (char *) 0) /* allocate the hash table element key's data area */
    {

        if ((element = (HASH *) malloc (sizeof (HASH))) != (HASH *) 0) /* allocate the hash table element */
        {
            element->next = element; /* all of the elements link references reference the element itself */
            element->previous = element;
            element->hash_data = obj_ref; /* reference to element's data */
            element->hash_data_printed = 0; /* reset the hash data printed, 0 = no, 1 = yes */
            (void) strcpy (obj_ref, data); /* save the hash table element's key as its data */
            hash_error = NOERROR; /* assume no error */
        }

    }

    return (element); /* return a reference to the hash table element */
}

/*

static void text_rmhash (HASH *element);

function to deallocate a text hash table element

returns nothing

*/

#ifdef __STDC__

static void text_rmhash (HASH *element)

#else

static void text_rmhash (element)
HASH *element;

#endif

{
    free (element->hash_data); /* free the key's data area allocated in text_mkhash () */
    free (element); /* free the hash table element allocated in text_mkhash () */
}

/*

Parse a record based on sequential delimiters.

int strtoken (char *string, char *parse_array, char **parse, const char *delim);

I) Parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field.

Returns the number of fields parsed.

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, const char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string;
char *parse_array;
char **parse;
char *delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}
