/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 2001-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

receivedIP.c is for listing the IP addresses in "Received: " records
in the headers of e-mail files to stdout. The order is the same as
encountered in the e-mail header.

The program requires a Posix compatible regex(3) library to parse the
IP addresses. The IP addresses from the e-mail headers are printed to
stout.

The input e-mail file name(s) may be supplied as additional optional
command line arguments, or redirected to the program via stdin for
compatibility with procmail(1), and other e-mail scripting agents.

The program contains less than 200 lines of declarations and
statements, all of which are documented with in line comments.

The program has been compiled and tested on SunOS, Solaris, and Linux,
and may work on other brands of Unix.

The program returns 0 if no error and IP addresses were found in any
"Received: " header records, 1 if no error and no IP addresses were
found; else returns a unique error code greater than 1 representing
the error encountered-which will, also, print an error diagnostic to
stderr.

John Conover
January 4, 2005
john@email.johncon.com

$Revision: 0.0 $
$Date: 2005/01/04 19:36:00 $
$Id: receivedIP.c,v 0.0 2005/01/04 19:36:00 john Exp $
$Log: receivedIP.c,v $
Revision 0.0  2005/01/04 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h> /* for toupper(), isspace() */
#include <unistd.h> /* for getopt() */
#include <regex.h> /* for regular expression; used for dotted quad notation IP address detection */

static char rcsid[] = "$Id: receivedIP.c,v 0.0 2005/01/04 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 2001-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define REGEXMATCH "([[(])(([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3}))([])])" /* the regular expression to detect a dotted quad IP address in the "Received: " record of an e-mail header */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "List the IP addresses in \"Received: \" records in the header of e-mail:\n",
    "Usage: receivedIP [-v] [filename(s)]\n",
    "    -v, print the program's version information\n",
    "    filename(s), e-mail file name(s), (defaults to stdin)\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error, match found\n",
    "No error, match not found\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error in input file record length\n",
    "Error compiling regular expression for dotted quad notation\n",
    "Error in e-mail format or structure\n"
};

#define MATCH 0 /* error values, one for each index in the error message array */
#define NOMATCH 1
#define EARGS 2
#define EOPEN 3
#define ECLOSE 4
#define ERECL 5
#define ECREGEX 6
#define ECONTINUE 7

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */

#else

static void print_message (); /* print any error messages */

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         temp[BUFLEN], /* temporary buffer space */
         *cptr; /* temporary character reference */

    int retval = NOMATCH, /* return value, assume no error */
        c, /* command line switch */
        match_count = 0, /* count of the number of matches-the number of IP addresses found in received records */
        received_active = 0; /* received active switch, 0 = no, 1 = received record, and/or its continuation(s) */

    size_t num_match = 8; /* REGEXMATCH contains 8 substrings */

    FILE *infile = stdin; /* reference to input file; defaults to stdin */

    regex_t reg; /* dotted quad regular expression compile */
    regmatch_t matches[8]; /* REGEXMATCH contains 8 substrings; matches[] contains their offsets */

    while ((c = getopt (argc, argv, "v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'v': /* request for program's version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOMATCH) /* enough arguments? */
    {
        retval = ECREGEX; /* assume error compiling regular expression for a dotted quad notation IP address */

        if (regcomp (&reg, REGEXMATCH, REG_EXTENDED) == 0) /* compile the dotted quad regular expression */
        {

            do /* while there are input file(s), if any, listed on the command line */
            {
                retval = EOPEN; /* assume error opening file */

                if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* open the input file; if none specified, default to stdin */
                {
                    retval = ECONTINUE; /* assume error in e-mail format or structure */

                    while ((fgets (buffer, BUFLEN, infile) != (char *) 0) && (retval == ECONTINUE)) /* read the records from the input file until EOF, or an error in processing, or a null line, (e.g., start of message body) */
                    {
                        retval = ERECL; /* assume error in record length */

                        if (strchr (buffer, '\n') != (char *) 0) /* a record with no carriage return was too long for the fgets () buffer, and an overflow condition exists; line contain a carriage return? */
                        {
                            retval = NOMATCH; /* assume no error */

                            if (buffer[0] != '\n') /* a record with no characters is first line of message body, quit */
                            {
                                retval = ECONTINUE; /* re-assume error in e-mail format or structure */
                                cptr = buffer; /* reference the first character in the record from the input file */

                                while (*cptr != '\n') /* capitalize the record from the input file; for each character in the record from the input file */
                                {
                                    *cptr = (char) toupper (*cptr); /* convert the character to upper case */
                                    cptr ++; /* next character */
                                }

                                if (strncmp ("RECEIVED: ", buffer, 10) == 0) /* first line of received record? */
                                {
                                    received_active = 1; /* set the received active switch, 0 = no, 1 = received record, and/or its continuation(s) */
                                }

                                else if (! isspace (buffer[0])) /* a record from the input file in the header having white space as the first character is a continuation record; if so leave the received active switch in its current state, else, reset it */
                                {
                                    received_active = 0; /* reset the received active switch, 0 = no, 1 = received record, and/or its continuation(s) */
                                }

                                if (received_active == 1) /* received active switch, 0 = no, 1 = received record, and/or its continuation(s), set? */
                                {
                                    cptr = buffer; /* reference the first character in the record from the input file */

                                    while (regexec (&reg, cptr, num_match, matches, 0) == 0) /* search the "received: " record, or its continuation(s), from the input file for dotted quad notation IP addresses, one after the other */
                                    {

                                        /*

                                        found a dotted quad notation IP address in the record from the input file:

                                        matches[0] contains the offsets of the entire match, e.g., [1.2.3.4], or (1.2.3.4)
                                        matches[1] contains the offsets of the leading '(' or '[' character of the entire match
                                        matches[2] contains the offsets of the dotted quad notation IP address, e.g., 1.2.3.4
                                        matches[3] contains the offsets of the first element in dotted quad notation IP address, e.g., 1
                                        matches[4] contains the offsets of the second element in dotted quad notation IP address, e.g., 2
                                        matches[5] contains the offsets of the third element in dotted quad notation IP address, e.g., 3
                                        matches[6] contains the offsets of the forth element in dotted quad notation IP address, e.g., 4
                                        matches[7] contains the offsets of the trailing ')' or ']' character of the entire match

                                        */

                                        match_count ++; /* increment the count of the number of matches-the number of IP addresses found in received records */

                                        temp[0] = '\0'; /* null the temporary buffer space */
                                        (void) strncat (temp, cptr + matches[3].rm_so, (unsigned int) (matches[3].rm_eo - matches[3].rm_so)); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                        (void) strcat (temp, "."); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                        (void) strncat (temp, cptr + matches[4].rm_so, (unsigned int) (matches[4].rm_eo - matches[4].rm_so)); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                        (void) strcat (temp, "."); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                        (void) strncat (temp, cptr + matches[5].rm_so, (unsigned int) (matches[5].rm_eo - matches[5].rm_so)); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                        (void) strcat (temp, "."); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                        (void) strncat (temp, cptr + matches[6].rm_so, (unsigned int) (matches[6].rm_eo - matches[6].rm_so)); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                        (void) strcat (temp, "\n"); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                        (void) printf ("%s", temp); /* print the dotted quad IP address in the "received: " record to stdout */
                                        cptr = &cptr[matches[0].rm_eo]; /* reference the next character after the entire match of the dotted quad IP address in the record from the input file to continue the search for dotted quad notation IP addresses, one after the other */
                                    }

                                }

                            }

                        }

                    }

                    if (argc > optind) /* using stdin as input? */
                    {

                        if (fclose (infile) == EOF) /* no, close the input file */
                        {
                            retval = ECLOSE; /* error closing file */
                        }

                    }

                }

            }
            while ((argc > ++ optind) && ((retval == MATCH) || (retval == NOMATCH))); /* while there are any remaining input file(s), if any, listed on the command line */

            regfree (&reg); /* free the dotted quad regular expression compile */
        }

    }

    retval = (((retval == MATCH) || (retval == NOMATCH)) && (match_count > 0)) ? MATCH : retval; /* if any matches, but no errors, set the return value to MATCH */
    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit-match found.

        2) Else if retval is unity, print nothing-normal/successful
           program exit-match not found.

        3) Else if retval is two, print help.

        4) Else retval is not zero, unity, or two, it is an error
           code, print the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, match found, print nothing */
        case 1: /* program ended without errors, match not found, print nothing */

            break;

        case 2: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}
