/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 2001-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

receivedIPdbrm.c is for removing the IP addresses in database files(s)
from a database file that is a sorted list of IP addresses that are
not permitted in database files. The database with IP addresses
removed is printed to stdout.

The program requires mmap(2) to map the database file into the Unix VM
system. The database file name that is a sorted list of IP addresses
that are not permitted in database file is a required command line
argument.

The database is a standard Unix text file, one IP address per line, in
lexical order, constructed with "sort -u infile > outfile", or
equivalent. An IP address range can be represented as a Class A, B, or
C range.  For example, the IP address "123.210." in the database file
would match "123.210.1.0".

The database mechanism is conservative with machine resources,
requiring about 12.5 micro-seconds of machine time to lookup a word in
the Unix system dictionary, (2.5 MB, quarter of a million words,
single 466 MHz., Pentium, lightly loaded, Linux 2.2, time(1) command
to lookup every word in the dictionary, divided by the number of
words.) Conceptually, the database mechanism is implemented similar to
the the technique used in the look(1) command, but requires exact
matches, as opposed to partial key matches.

The input database file name(s) may be supplied as additional optional
command line arguments, or redirected to the program via stdin.

The most common usage is where the input database, rmexample.db, is a
list of IP addresses of hosts that are up stream mail providers.

    receivedIPdbrm example.db < rmexample.db > example.db

which besides removing records that contain IP addresses of up stream
mail providers from example.db, also, removes records that would be
matched by a superset of IP addresses-for example, if the database
file, example.db, contained the IP addresses "123.210.1.0", and the
rmexample.db file contained the Class B address "123.210", the
"123.210.1.0" address would be removed from the output, since the
Class B address is a superset of the IP address.

The program contains less than 300 lines of declarations and
statements, all of which are documented with in line comments.

The program has been compiled and tested on SunOS, Solaris, and Linux,
and may work on other brands of Unix.

The program returns 0 if no error and a match was found in the
database file for the IP addresses, 1 if no error and no match found;
else returns a unique error code greater than 1 representing the error
encountered-which will, also, print an error diagnostic to stderr.

John Conover
January 8, 2005
john@email.johncon.com

$Revision: 0.0 $
$Date: 2005/01/08 19:36:00 $
$Id: receivedIPdbrm.c,v 0.0 2005/01/08 19:36:00 john Exp $
$Log: receivedIPdbrm.c,v $
Revision 0.0  2005/01/08 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* for getopt() */
#include <sys/types.h> /* for open() */
#include <sys/stat.h> /* for open() */
#include <fcntl.h> /* for open() */
#include <unistd.h> /* for mmap() */
#include <sys/mman.h> /* for mmap() */

static char rcsid[] = "$Id: receivedIPdbrm.c,v 0.0 2005/01/08 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 2001-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define TOKEN_SEPARATORS " \t\n\r\b,." /* file record field separators for dotted quad notation */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Remove the IP addresses in database files(s) against a database file:\n",
    "Usage: receivedIPdbrm [-v] rmdbfilename [dbfilename(s)]\n",
    "    -v, print the program's version information\n",
    "    rmdbfilename, remove database file name\n",
    "    dbfilename(s), database file names(s), (defaults to stdin)\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error, match found\n",
    "No error, match not found\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error in input file record length\n",
    "Error in status of database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error in database file\n"
};

#define MATCH 0 /* error values, one for each index in the error message array */
#define NOMATCH 1
#define EARGS 2
#define EOPEN 3
#define ECLOSE 4
#define ERECL 5
#define ESTAT 6
#define EMAP 7
#define EUNMAP 8
#define EDB 9

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap);
static int strtoken (char *string, char *parse_array, char **parse, const char *delim);

#else

static void print_message (); /* print any error messages */
static char *binarysearch ();
static int strtoken ();

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         temp[BUFLEN], /* temporary buffer space */
         *begin, /* begining of memory mapped file */
         *end; /* end of memory mapped file */

    int retval = NOMATCH, /* return value, assume no error */
        fd = -1, /* input file descriptor */
        fields, /* number of fields in a record */
        field_ctr, /* field counter in number of fields in a record */
        c, /* command line switch */
        match_count = 0; /* count of the number of matches-the number of IP addresses found in the database */

    FILE *infile = stdin; /* reference to input file; defaults to stdin */

    size_t filesize; /* size of file */

    struct stat fileinfo; /* structure for file status */

    caddr_t mapbegin; /* reference to beginning of mmap'ed file */

    while ((c = getopt (argc, argv, "v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'v': /* request for program's version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOMATCH) /* any errors? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (argc > optind) /* enough arguments? */
        {
            retval = ESTAT; /* assume error in status of file */

            if (stat (argv[optind], &fileinfo) == 0) /* get the status of the file */
            {
                retval = EDB;  /* assume error in database file */

                if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of file */
                {
                    retval = EOPEN; /* assume error opening file */

                    if ((fd = open (argv[optind], O_RDONLY)) != -1) /* open the file */
                    {
                        retval = EMAP; /* assume error mapping file */

                        if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, fd, (off_t) 0)) != (caddr_t) 0) /* map the file */
                        {
                            retval = EDB;  /* assume error in database file */

                            if (mapbegin[filesize - 1] == '\n') /* verify that the last record of the memory mapped file has a terminating '\n' character as required for binarysearch() */
                            {
                                retval = NOMATCH; /* yes, return value, assume no error */
                                optind ++; /* file argument handled */
                                begin = (char *) mapbegin; /* reference the begining of memory mapped file */
                                end = &mapbegin[filesize - 1]; /* reference the end of memory mapped file */

                                do /* while there are input file(s), if any, listed on the command line */
                                {
                                    retval = EOPEN; /* assume error opening file */

                                    if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* open the input file; if none specified, default to stdin */
                                    {

                                        while (fgets (buffer, BUFLEN, infile) != (char *) 0) /* read the records from the input file until EOF */
                                        {
                                            retval = ERECL; /* assume error in record length */

                                            if (strchr (buffer, '\n') != (char *) 0) /* a record with no carriage return was too long for the fgets () buffer, and an overflow condition exists; line contain a carriage return? */
                                            {
                                                fields = strtoken (buffer, parsebuffer, token, TOKEN_SEPARATORS); /* parse the record into quad fields, skip the record if there are no fields */
                                                temp[0] = '\0'; /* null the temporary buffer space */
                                                retval = NOMATCH; /* return value, assume no match */

                                                for (field_ctr = 0; (field_ctr < fields) && (retval == NOMATCH); field_ctr ++) /* for each field in number of fields in a record */
                                                {
                                                    (void) strcat (temp, token[field_ctr]); /* concatinate the quads into an IP address */

                                                    if (field_ctr < 3) /* last field of a dotted quad? */
                                                    {
                                                        (void) strcat (temp, "."); /* add a period field separator to the dotted quad */
                                                    }

                                                    if (binarysearch (begin, end, temp, strlen (temp), begin, end) != (char *) 0) /* search the database for the IP address */
                                                    {
                                                        retval = MATCH; /* assume a match will be found in the database */
                                                        match_count ++; /* increment the count of the number of matches-the number of IP addresses found in the database */
                                                        break;
                                                    }

                                                }

                                                if (retval != MATCH) /* no unique match found in the database? */
                                                {
                                                    (void) printf ("%s\n", temp); /* yes, the IP address was not found in the database, print the IP address */
                                                }

                                            }

                                        }

                                        if (argc > optind) /* using stdin as input? */
                                        {

                                            if (fclose (infile) == EOF) /* no, close the input file */
                                            {
                                                retval = ECLOSE; /* error closing file */
                                            }

                                        }

                                    }

                                }
                                while ((argc > ++ optind) && ((retval == MATCH) || (retval == NOMATCH))); /* while there are any remaining input file(s), if any, listed on the command line */

                            }

                            if (munmap ((void *) mapbegin, filesize) != 0) /* unmap the file */
                            {
                                retval = EUNMAP; /* error unmapping file */
                            }

                        }

                        if (close (fd) != 0) /* close the file */
                        {
                            retval = ECLOSE; /* error closing file */
                        }

                    }

                }

            }

        }

    }

    retval = (((retval == MATCH) || (retval == NOMATCH)) && (match_count > 0)) ? MATCH : retval; /* if any matches, but no errors, set the return value to MATCH */
    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit-match found.

        2) Else if retval is unity, print nothing-normal/successful
           program exit-match not found.

        3) Else if retval is two, print help.

        4) Else retval is not zero, unity, or two, it is an error
           code, print the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, match found, print nothing */
        case 1: /* program ended without errors, match not found, print nothing */

            break;

        case 2: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Recursively binary search a memory area.

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap);

I) Binary search a memory area, referenced by the begin and end
variables, for string. The variable begin references the beginning of
the area to be searched; the variable end references the end of the
area to be searched, and will always reference the '\n' terminating
character of the last record of the memory area, (which is a
requirment); the variable string is the null terminated string to
search for, and has a length represented by the variable string_len.

II) Requirements:

    A) The memory area size must be greater than zero, (this is a
    requirement for mmap(1), also.)

    B) The last character of the memory area must be a '\n' text file
    record terminator, e.g., the last record, (which, minimally, may
    be the only record,) must be terminated.

    C) The records in the memory area must be in lexical sequence,
    (consistent with strcmp() and strncmp().)

III) The general strategy is:

    A) For each iteration in the recursion:

        1) Divide the memory area exactly in two.

        2) Work from this middle, in both directions, to find the end
        of the preceding record, (or beginning of file, whichever
        comes first) and, the end of the current record:

            a) The current record contains the first instance of the
            the terminating character at, or past, the middle.

            b) The current record's beginning is the next character
            after the preceding records's terminating character, or
            the beginning of the memory area.

        3) Compare this record with the string for a match.

        4) If not a match, decide which direction to move for the next
        iteration, and call this function, again.

        5) The recursive binary search terminates when a match is
        found, or when there is no more memory area to search.

    B) Return a reference to the matched area in memory, or null if no
    match.

Returns a reference to the matched string, null if no match.

*/

#ifdef __STDC__

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap)

#else

static char *binarysearch (begin, end, string, string_len, BOFmap, EOFmap)
char *begin;
char *end;
char *string;
size_t string_len;
char *BOFmap;
char *EOFmap;

#endif

{
    int retval; /* return for string comparisons */

    char *middle, /* reference to the middle between begin and end */
         *BOFmiddle_record, /* reference to the beginning of the record in the middle */
         *EOFmiddle_record, /* reference to the end of the record in the middle */
         *BOFnext_record, /* beginning of next record */
         *EOFprevious_record, /* end of previous record */
         *match = (char *) 0; /* reference to the beginning of the record in the middle if a match is found; null if no match */

    size_t middle_record_len; /* length record in the middle */

/* #define QA */ /* comment out to disable QA/Debuging */

#ifdef QA

    /*

    The Unix dictionary file, "sort -u /usr/local/lib/dict/websters >
    testfile", is a good QA test candidate.

    Also "thisprogram testfile < testfile > out" without QA #defined,
    and "cmp testfile out" is a good QA test candidate, (to verify
    that all words are successfully found,) where thisprogram is the
    name of the program containing this function.

    Test for the pathological cases, too; duplicate records, null
    records, single null records, etc.

    */

    char QAbuf[BUFSIZ], /* QA print buffer */
         *QAbegin, /* beginning of record to be printed */
         *QAend; /* end of record to be printed */

    size_t QAreclen; /* QA record length in print buffer */

#endif

    middle = BOFmiddle_record = EOFmiddle_record = begin + ((end - begin) / 2); /* reference the middle between begin and end */

    while ((BOFmap < BOFmiddle_record) && (*(BOFmiddle_record - 1) != '\n')) /* step backwards to the beginning of the record in the middle */
    {
        BOFmiddle_record --;
    }

    while ((EOFmiddle_record < EOFmap) && (*EOFmiddle_record != '\n')) /* step forwards to the end of the record in the middle */
    {
        EOFmiddle_record ++;
    }

    begin = (BOFmiddle_record < begin) ? BOFmiddle_record : begin; /* beginning of record in the middle outside of begin? */
    end = (end < EOFmiddle_record) ? EOFmiddle_record : end; /* end of record in the middle outside of end? */
    BOFnext_record = (EOFmiddle_record < EOFmap) ? EOFmiddle_record + 1 : EOFmap; /* save the beginning of next record */
    EOFprevious_record = (BOFmap < BOFmiddle_record) ? BOFmiddle_record - 1 : BOFmap; /* end of previous record */
    middle_record_len = (size_t) (EOFmiddle_record - BOFmiddle_record); /* save the length of the record in the middle, minus the terminating '\n' */
    retval = strncmp (string, BOFmiddle_record, string_len < middle_record_len ? string_len : middle_record_len); /* compare the record in the middle with the search string */

#ifdef QA

    (void) printf ("%s, ", string); /* print the string being searched for */

    QAreclen = strchr (BOFmiddle_record, '\n') - BOFmiddle_record; /* save the length of the record being searched */
    (void) strncpy (QAbuf, BOFmiddle_record, QAreclen); /* copy the record being searched to the print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the record being searched in the print buffer */
    (void) printf ("%s, ", QAbuf); /* print the copy of the record being searched */

    QAreclen = strchr (begin, '\n') - begin; /* length of first record in begin/end */
    (void) strncpy (QAbuf, begin, QAreclen); /* copy first record in begin/end to QA print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the first record in begin/end */
    (void) printf ("(%s - ", QAbuf); /* print the copy of the first record in begin/end */

    QAend = QAbegin = end; /* end of last record in begin/end */

    while ((BOFmap < QAbegin) && (*(QAbegin - 1) != '\n')) /* step backwards to the beginning of the last record in begin/end */
    {
        QAbegin --;
    }

    (void) strncpy (QAbuf, QAbegin, (size_t) (QAend - QAbegin)); /* copy last record in begin/end to QA print buffer */
    QAbuf[QAend - QAbegin] = '\0'; /* terminate the copy of the last record in begin/end */
    (void) printf ("%s)\n", QAbuf); /* print the copy of the last record in begin/end */

#endif

    if (retval < 0) /* string < record? */
    {

        if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
        {
            match = binarysearch (begin, EOFprevious_record, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
        }

    }

    else if (retval > 0) /* string > record? */
    {

        if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
        {
            match = binarysearch (BOFnext_record, end, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
        }

    }

    else /* match, string = record? */
    {

        if (string_len < middle_record_len) /* partial key match; string shorter than record? */
        {

            if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
            {
                match = binarysearch (begin, EOFprevious_record, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
            }

        }

        else if (string_len > middle_record_len) /* partial key match; string longer than record? */
        {

            if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
            {
                match = binarysearch (BOFnext_record, end, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
            }

        }

        else /* exact match */
        {
            match = BOFmiddle_record; /* return a reference to this record */
        }

    }

    return (match); /* return a reference to the match found between begin and end; null if no match */
}

/*

Parse a record based on sequential delimiters.

int strtoken (char *string, char *parse_array, char **parse, const char *delim);

I) Parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field.

Returns the number of fields parsed.

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, const char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string;
char *parse_array;
char **parse;
char *delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}
