/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 2001-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

receivedIPdbusort.c is for sorting a database of IP addresses-each IP
address in the output will be unique.

The program requires mmap(2) to map the database file of IP addresses
into the Unix VM system. The database file name is a required command
line argument.

The database is a standard Unix text file, one IP address per line, in
lexical order, constructed with "sort -u infile > outfile", or
equivalent. An IP address range can be represented as a Class A, B, or
C range.  For example, the IP address "123.210." in the database file
would match "123.210.1.0".

The most common usage is where the input database, example.db, is a
list of IP addresses, and the output file is a uniquely sorted list of
IP addresses, out.db.

    receivedIPdbusort example.db > out.db.

The sort mechanism is implemented as a quicksort.

The program, also, eliminates duplicate records, (i.e., records that
are lexically equal,) and removes leading and trailing white space
in each record.

The program contains less than 300 lines of declarations and
statements, all of which are documented with in line comments.

The program has been compiled and tested on SunOS, Solaris, and Linux,
and may work on other brands of Unix.

The program returns 0 if no error; else returns a unique error code
greater than 1 representing the error encountered-which will, also,
print an error diagnostic to stderr.

John Conover
January 18, 2005
john@email.johncon.com

$Revision: 0.0 $
$Date: 2005/01/18 19:36:00 $
$Id: receivedIPdbusort.c,v 0.0 2005/01/18 19:36:00 john Exp $
$Log: receivedIPdbusort.c,v $
Revision 0.0  2005/01/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* for getopt() */
#include <sys/types.h> /* for open() */
#include <sys/stat.h> /* for open() */
#include <fcntl.h> /* for open() */
#include <unistd.h> /* for mmap() */
#include <sys/mman.h> /* for mmap() */

static char rcsid[] = "$Id: receivedIPusort.c,v 0.0 2005/01/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 2001-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Sort the IP addresses in a database file:\n",
    "Usage: receivedIPdbusort [-v] dbfilename\n",
    "    -v, print the program's version information\n",
    "    dbfilename, database file name\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error in status of database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error allocating memory\n",
    "Error in database file\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define ECLOSE 3
#define ESTAT 4
#define EMAP 5
#define EUNMAP 6
#define EALLOC 7
#define EDB 8

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static void str_sort (char *array[],int bottom,int top);
static int strcmpeol (char *s, char *t);

#else

static void print_message (); /* print any error messages */
static void str_sort ();
static int strcmpeol ();

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         eof = '\n', /* an EOF terminated null record */
         *cptr, /* temporary character reference */
         *begin, /* begining of memory mapped file */
         *end, /* end of memory mapped file */
         *last_record, /* reference to the last record output */
         **ref_array; /* reference to array of pointers to records in memory mapped file */

    int retval = NOERROR, /* return value, assume no error */
        fd = -1, /* input file descriptor */
        c, /* command line switch */
        line_counter, /* the count of lines in the file */
        i; /* line counter */

    size_t filesize; /* size of file */

    struct stat fileinfo; /* structure for file status */

    caddr_t mapbegin; /* reference to beginning of mmap'ed file */

    while ((c = getopt (argc, argv, "v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'v': /* request for program's version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* any errors? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (argc > optind) /* enough arguments? */
        {
            retval = ESTAT; /* assume error in status of file */

            if (stat (argv[optind], &fileinfo) == 0) /* get the status of the file */
            {
                retval = EDB;  /* assume error in database file */

                if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of file */
                {
                    retval = EOPEN; /* assume error opening file */

                    if ((fd = open (argv[optind], O_RDONLY)) != -1) /* open the file */
                    {
                        retval = EMAP; /* assume error mapping file */

                        if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, fd, (off_t) 0)) != (caddr_t) 0) /* map the file */
                        {
                            retval = EDB;  /* assume error in database file */

                            if (mapbegin[filesize - 1] == '\n') /* verify that the last record of the memory mapped file has a terminating '\n' character as required for binarysearch() */
                            {
                                retval = NOERROR; /* assume no error */
                                optind ++; /* file argument handled */
                                cptr = begin = mapbegin; /* reference the beginning of the map file */
                                end = &mapbegin[filesize - 1]; /* reference the end of the map file */
                                line_counter = 0; /* zero the count of lines in the file */

                                while (cptr < end) /* for each EOL in the file */
                                {
                                    cptr = strchr (cptr, '\n'); /* next EOL in the file */
                                    cptr ++; /* step over the EOL */
                                    line_counter ++; /* increment the count of lines in the file */
                                }

                                if (line_counter > 0) /* any records? */
                                {
                                    retval = EALLOC; /* assume allocating memory */

                                    if ((ref_array = (char **) malloc (((size_t) line_counter) * sizeof (char *))) != (char **) 0) /* allocate space for the references to the array of pointers to records in memory mapped file */
                                    {
                                        retval = NOERROR; /* assume no error */
                                        cptr = begin = mapbegin; /* reference the beginning of the map file */
                                        line_counter = 0; /* zero the count of lines in the file */

                                        while (cptr < end) /* for each EOL in the file */
                                        {

                                            while (*cptr == ' ' || *cptr == '\t' || *cptr == '\b') /* skip any leading white space in the record */
                                            {
                                                cptr ++;
                                            }

                                            ref_array[line_counter] = cptr; /* save the reference to the record in the memory mapped file */
                                            cptr = strchr (cptr, '\n'); /* next EOL in the file */
                                            cptr ++; /* step over the EOL */
                                            line_counter ++; /* increment the count of lines in the file */
                                        }

                                        str_sort (ref_array, 0, line_counter - 1); /* sort the array of pointers to records in memory mapped file */
                                        last_record = &eof; /* set the reference to the last record output to an EOF null terminated null record */

                                        for (i = 0; i < line_counter; i ++) /* for each record in the file */
                                        {

                                            if (*ref_array[i] != '\n') /* null record? */
                                            {

                                                if (strcmpeol (ref_array[i], last_record) != 0) /* this record the same as the last record output? */
                                                {
                                                    buffer[0] = '\0'; /* null the i/o buffer */
                                                    (void) strncat (buffer, ref_array[i], (size_t) (strchr (ref_array[i], '\n') - ref_array[i])); /* make a copy the record */

                                                    if ((cptr = strchr (buffer, ' ')) != (char *) 0) /* reference any trailing space */
                                                    {
                                                        *cptr = '\0'; /* if it exists, null it */
                                                    }

                                                    if ((cptr = strchr (buffer, '\t')) != (char *) 0) /* reference any trailing tab */
                                                    {
                                                        *cptr = '\0'; /* if it exists, null it */
                                                    }

                                                    if ((cptr = strchr (buffer, '\b')) != (char *) 0) /* reference any trailing backspace */
                                                    {
                                                        *cptr = '\0'; /* if it exists, null it */
                                                    }

                                                    (void) printf ("%s\n", buffer); /* and print the record */
                                                    last_record = ref_array[i]; /* save the reference to the last record output */
                                                }

                                            }

                                        }

                                        free (ref_array); /* free the space for the references to the array of pointers to records in memory mapped file */
                                    }

                                }

                            }

                            if (munmap ((void *) mapbegin, filesize) != 0) /* unmap the file */
                            {
                                retval = EUNMAP; /* error unmapping file */
                            }

                        }

                        if (close (fd) != 0) /* close the file */
                        {
                            retval = ECLOSE; /* error closing file */
                        }

                    }

                }

            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Quicksort an EOL, ('\n'), terminated array of character strings.

static void str_sort (char *array[], int bottom, int top);

Note: this is a recursive proceedure, requiring three references, and
4 integers to be allocated on the stack for each recursion.

Note: on entry, argument "bottom" should be zero, and argument "top"
should be the index of the last array element, ie., if the array
contains n many elements, then the call would be:

    str_sort (array, 0, n - 1);

Returns nothing.

*/

#define MIDWAY(a, b) (int)(((unsigned int)(a)+(unsigned int)(b))/(unsigned int)2)

#ifdef __STDC__

static void str_sort (char *array[], int bottom, int top)

#else

static void str_sort (array, bottom, top)
char *arrag[];
int bottom;
int top;

#endif

{
    int i = bottom, /* bottom index of the array */
        j = top; /* top index of the array */

    char *x,
         *y;

    x = array[MIDWAY(bottom, top)]; /* starting in the middle of the array */

    do /* while the bottom moving index of the array is less than or equal to the top moving index of the array */
    {

        while (strcmpeol (array[i], x) < 0 && i < top) /* increment the bottom moving index of the array while the elements referenced are lexically "less" */
        {
            i ++;
        }

        while (strcmpeol (array[j], x) > 0 && j > bottom) /* decrement the top moving index of the array while the elements referenced are lexically "greater" */
        {
            j --;
        }

        if (i <= j) /* if the bottom moving index is less than or equal to the top moving index of the array */
        {
            y = array[i]; /* swap the element references, increment the bottom moving index, decrement the top moving index */
            array[i] = array[j];
            array[j] = y;
            i ++;
            j --;
        }

    }
    while (i <= j);

    if (bottom < j) /* if the top moving index is not at the bottom of the array elements, sort the remaining elements */
    {
        str_sort (array, bottom, j);
    }

    if (i < top) /* if the bottom moving index is not at the top of the array elements, sort the remaining elements */
    {
        str_sort (array, i, top);
    }

}

/*

Compare two EOL, ('\n'), terminated character strings.

static int strcmpeol (char *s, char *t);

Exactly the same as strcmp (), but the character strings are EOL,
'\n', instead of null, '\0', terminated.

Returns an integer less than, equal to, or greater than zero to
indicate whether string, s, is less than, equal to, or greater than
string, t.

*/

#ifdef __STDC__

static int strcmpeol (char *s, char *t)

#else

static int strcmpeol (s, t)
char *s;
char *t;

#endif

{

    for ( ; *s == *t; s ++, t ++) /* for each character */
    {

        if (*s == '\n') /* if the source is at EOL */
        {
            return (0); /* match, return zero */
        }

    }

    return ((int) (*s - *t)); /* no match, return the difference between the last two characters compared */
}
