/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 2001-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

receivedIPforgedb.c is for validating the IP addresses in "Received: "
records in the headers of e-mail files against IP addresses in a
database file. The fully qualified domain name is a required argument,
and only "Received: " records from that domain will be validated.

The program requires a Posix compatible regex(3) library to parse the
IP addresses, and mmap(2) to map the database file of IP addresses
into the Unix VM system. The IP addresses from the e-mail header are
validated against the database file using a binary search. The
database file name is a required command line argument.

The database is a standard Unix text file, one IP address per line, in
lexical order, constructed with "sort -u infile > outfile", or
equivalent. An IP address range can be represented as a Class A, B, or
C range.  For example, the IP address "123.210." in the database file
would match "123.210.1.0" in a "Received: " e-mail header record.

The database mechanism is conservative with machine resources,
requiring about 12.5 micro-seconds of machine time to lookup a word in
the Unix system dictionary, (2.5 MB, quarter of a million words,
single 466 MHz., Pentium, lightly loaded, Linux 2.2, time(1) command
to lookup every word in the dictionary, divided by the number of
words.) Conceptually, the database mechanism is implemented similar to
the the technique used in the look(1) command, but requires exact
matches, as opposed to partial key matches.

The program has implicit IP addresses that do not have to be included
in the database-those with invalid "dotted quad" element values, (such
as greater that 255, for example.) Such IP addresses will be
rejected-if the -p, or -P arguments are used, such values will be
denoted by not having a trailing "dot".

The input e-mail file name(s) may be supplied as additional optional
command line arguments, or redirected to the program via stdin for
compatibility with procmail(1), and other e-mail scripting agents.

A suitable procmail(1) recipe example might be:

    :0 wfh
    * ? receivedIPforgedb domain.com accept.db
    | formail -A "X-Notice: Message in accept.db database"

which could be, if necessary, overridden, on a case-by-case basis,
with the example recipe:

    :0 wfh
    * ^X-Notice: +Message +in +accept.db +database
    * ? receivedIPforgedb domain.com accept.db
    | formail -I "X-Notice: Message in accept.db database"

or similar construct.

The program contains less than 300 lines of declarations and
statements, all of which are documented with in line comments.

The program has been compiled and tested on SunOS, Solaris, and Linux,
and may work on other brands of Unix.

The program returns 0 if no error and all IP addresses were found in
the database file, that were found in all "Received: " header records,
containing the domain name, 1 if no error and the domain name found in
a "Received: " header record, but no corresponding IP address match
found; else returns a unique error code greater than 1 representing
the error encountered-which will, also, print an error diagnostic to
stderr.

The -r option is useful for controlling the return value under error
conditions-for example, the program return can be preempted if the
database file can not be opened, (or read,) with a return value of
match, or no match, depending on environmental requirements.

John Conover
January 17, 2005
john@email.johncon.com

$Revision: 0.0 $
$Date: 2005/01/17 19:36:00 $
$Id: receivedIPforgedb.c,v 0.0 2005/01/17 19:36:00 john Exp $
$Log: receivedIPforgedb.c,v $
Revision 0.0  2005/01/17 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* for getopt() */
#include <ctype.h> /* for toupper(), isspace() */
#include <regex.h> /* for regular expression; used for dotted quad notation IP address detection */
#include <sys/types.h> /* for open() */
#include <sys/stat.h> /* for open() */
#include <fcntl.h> /* for open() */
#include <unistd.h> /* for mmap() */
#include <sys/mman.h> /* for mmap() */

static char rcsid[] = "$Id: receivedIPforgedb.c,v 0.0 2005/01/17 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 2001-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define REGEXMATCH "([[(])(([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3}))([])])" /* the regular expression to detect a dotted quad IP address in the "Received: " record of an e-mail header */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Validate the domain IP addresses in \"Received: \" records in the header of\n",
    "e-mail against IP addresses in a database file:\n",
    "Usage: receivedIPforgedb [-p | -P] [-r m|n] [-v] fqdn dbfilename [filename(s)]\n",
    "    -p, print the IP address match from the database\n",
    "    -P, print the IP address if it is not in the database\n",
    "    -r m|n, on file error, exit return = match for m, no match for n\n",
    "    -v, print the program's version information\n",
    "    fqdn, fully qualified domain name\n",
    "    dbfilename, database file name\n",
    "    filename(s), e-mail file name(s), (defaults to stdin)\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error, match found\n",
    "No error, match not found\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error in input file record length\n",
    "Error compiling regular expression\n",
    "Error in e-mail format or structure\n",
    "Error in status of database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error in database file\n"
};

#define MATCH 0 /* error values, one for each index in the error message array */
#define NOMATCH 1
#define EARGS 2
#define EOPEN 3
#define ECLOSE 4
#define ERECL 5
#define ECREGEX 6
#define ECONTINUE 7
#define ESTAT 8
#define EMAP 9
#define EUNMAP 10
#define EDB 11

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap);

#else

static void print_message (); /* print any error messages */
static char *binarysearch ();

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         temp[BUFLEN], /* temporary buffer space */
         *cptr, /* temporary character reference */
         *cptr1, /* temporary character reference */
         *quad, /* reference to element in dotted quad */
         *begin, /* begining of memory mapped file */
         *end; /* end of memory mapped file */

    int retval = NOMATCH, /* return value, assume no error */
        fd = -1, /* input file descriptor */
        c, /* command line switch */
        quad_number, /* numerical value of an element in dotted quad */
        match_count = 0, /* count of the number of matches-the number of IP addresses found in the database */
        received_active = 0, /* received active switch, 0 = no, 1 = received record, and/or its continuation(s) */
        print_flag = 0, /* verbose print flag, 0 = no print, 1 = print if IP address is in database, 2 = print if it isn't */
        error_return_change = -1, /* change exit return on error, -1 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
        non_match = MATCH; /* domain found with invalid IP address, 0 = no, 1 = yes */

    size_t num_match = 8; /* REGEXMATCH contains 8 substrings */

    FILE *infile = stdin; /* reference to input file; defaults to stdin */

    size_t filesize; /* size of file */

    struct stat fileinfo; /* structure for file status */

    caddr_t mapbegin; /* reference to beginning of mmap'ed file */

    regex_t reg; /* dotted quad regular expression compile */
    regmatch_t matches[8]; /* REGEXMATCH contains 8 substrings; matches[] contains their offsets */

    regex_t reg1; /* received domain regular expression compile */
    regmatch_t matches1[8]; /* domain regular expression contains 8 substrings; matches1[] contains their offsets */

    while ((c = getopt (argc, argv, "Ppr:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'p': /* request for verbose print? */

                print_flag = 1; /* set the verbose print flag, 0 = no print, 1 = print if IP address is in database, 2 = print if it isn't */
                break;

            case 'P': /* request for verbose print? */

                print_flag = 2; /* set the verbose print flag, 0 = no print, 1 = print if IP address is in database, 2 = print if it isn't */
                break;

            case 'r': /* request for change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH? */

                switch (*optarg) /* first character of argument object, only */
                {

                    case 'm':

                        error_return_change = MATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    case 'n':

                        error_return_change = NOMATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    default:

                        optind = argc; /* force argument error */
                        retval = EARGS; /* assume not enough arguments */
                        break;
                }

                break;

            case 'v': /* request for program's version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOMATCH) /* any errors? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (argc > optind) /* enough arguments? */
        {
            (void) strcpy (temp, "RECEIVED: +FROM +([-0-9A-Z_+=]+\\.)*"); /* searching for something like "Received: from somemachine.somedomain.com" */
            cptr = strchr (temp, '\0'); /* append to end of "Received: " preamble */
            cptr1 = argv[optind]; /* save the reference to domain name */

            do /* for each character in the domain name */
            {
                *cptr ++ = (char) toupper (*cptr1); /* capitalize the character, and add to the "Received: " preamble */
            }
            while (*cptr1 ++ != '\0');

            (void) strcat (temp, "[^-0-9a-z._+=]"); /* add non-domain characters */

            retval = ECREGEX; /* assume error compiling regular expression for a dotted quad notation IP address */

            if (regcomp (&reg1, temp, REG_EXTENDED) == 0) /* compile the received domain regular expression */
            {
                optind ++; /* domain name handled */
                retval = EARGS; /* assume not enough arguments */

                if (argc > optind) /* enough arguments? */
                {
                    retval = ESTAT; /* assume error in status of file */

                    if (stat (argv[optind], &fileinfo) == 0) /* get the status of the file */
                    {
                        retval = EDB;  /* assume error in database file */

                        if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of file */
                        {
                            retval = EOPEN; /* assume error opening file */

                            if ((fd = open (argv[optind], O_RDONLY)) != -1) /* open the file */
                            {
                                retval = EMAP; /* assume error mapping file */

                                if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, fd, (off_t) 0)) != (caddr_t) 0) /* map the file */
                                {
                                    retval = EDB;  /* assume error in database file */

                                    if (mapbegin[filesize - 1] == '\n') /* verify that the last record of the memory mapped file has a terminating '\n' character as required for binarysearch() */
                                    {
                                        retval = NOMATCH; /* yes, return value, assume no error */
                                        optind ++; /* file argument handled */
                                        begin = (char *) mapbegin; /* reference the begining of memory mapped file */
                                        end = &mapbegin[filesize - 1]; /* reference the end of memory mapped file */
                                        retval = ECREGEX; /* assume error compiling regular expression for a dotted quad notation IP address */

                                        if (regcomp (&reg, REGEXMATCH, REG_EXTENDED) == 0) /* compile the dotted quad regular expression */
                                        {

                                            do /* while there are input file(s), if any, listed on the command line */
                                            {
                                                retval = EOPEN; /* assume error opening file */

                                                if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* open the input file; if none specified, default to stdin */
                                                {
                                                    retval = ECONTINUE; /* assume error in e-mail format or structure */

                                                    while ((fgets (buffer, BUFLEN, infile) != (char *) 0) && (retval == ECONTINUE)) /* read the records from the input file until EOF, or an error in processing, or a null line, (e.g., start of message body) */
                                                    {
                                                        retval = ERECL; /* assume error in record length */

                                                        if (strchr (buffer, '\n') != (char *) 0) /* a record with no carriage return was too long for the fgets () buffer, and an overflow condition exists; line contain a carriage return? */
                                                        {
                                                            retval = NOMATCH; /* assume no error */

                                                            if (buffer[0] != '\n') /* a record with no characters is first line of message body, quit */
                                                            {
                                                                retval = ECONTINUE; /* re-assume error in e-mail format or structure */
                                                                cptr = buffer; /* reference the first character in the record from the input file */

                                                                while (*cptr != '\n') /* capitalize the record from the input file; for each character in the record from the input file */
                                                                {
                                                                    *cptr = (char) toupper (*cptr); /* convert the character to upper case */
                                                                    cptr ++; /* next character */
                                                                }

                                                                if (strncmp ("RECEIVED: ", buffer, 10) == 0) /* first line of received record? */
                                                                {
                                                                    cptr = buffer; /* reference the first character in the record from the input file */

                                                                    if (regexec (&reg1, cptr, num_match, matches1, 0) == 0) /* search the "received: " from the input file for the received domain */
                                                                    {

                                                                        /*

                                                                        found received domain in the record from the input file:

                                                                        matches1[0] contains the offsets of the entire match, e.g., Received: from somemachine.somedomain.com

                                                                        */

                                                                        received_active = 1; /* set the received active switch, 0 = no, 1 = received record, and/or its continuation(s) */
                                                                    }

                                                                }

                                                                else if (! isspace (buffer[0])) /* a record from the input file in the header having white space as the first character is a continuation record; if so leave the received active switch in its current state, else, reset it */
                                                                {
                                                                    received_active = 0; /* reset the received active switch, 0 = no, 1 = received record, and/or its continuation(s) */
                                                                }

                                                                if (received_active == 1) /* received active switch, 0 = no, 1 = received record, and/or its continuation(s), set? */
                                                                {
                                                                    cptr = buffer; /* reference the first character in the record from the input file */

                                                                    while ((regexec (&reg, cptr, num_match, matches, 0) == 0) && (retval == ECONTINUE)) /* search the "received: " record, or its continuation(s), from the input file for dotted quad notation IP addresses, one after the other */
                                                                    {

                                                                        /*

                                                                        found a dotted quad notation IP address in the record from the input file:

                                                                        matches[0] contains the offsets of the entire match, e.g., [1.2.3.4], or (1.2.3.4)
                                                                        matches[1] contains the offsets of the leading '(' or '[' character of the entire match
                                                                        matches[2] contains the offsets of the dotted quad notation IP address, e.g., 1.2.3.4
                                                                        matches[3] contains the offsets of the first element in dotted quad notation IP address, e.g., 1
                                                                        matches[4] contains the offsets of the second element in dotted quad notation IP address, e.g., 2
                                                                        matches[5] contains the offsets of the third element in dotted quad notation IP address, e.g., 3
                                                                        matches[6] contains the offsets of the forth element in dotted quad notation IP address, e.g., 4
                                                                        matches[7] contains the offsets of the trailing ')' or ']' character of the entire match

                                                                        */

                                                                        retval = MATCH; /* assume a match will be found in the database */
                                                                        match_count ++; /* assume a match will be found in the database, increment the count of the number of matches-the number of IP addresses found in the database */

                                                                        temp[0] = '\0'; /* null the temporary buffer space */
                                                                        quad = temp; /* reference the first element in the dotted quad */
                                                                        (void) strncat (temp, cptr + matches[3].rm_so, (unsigned int) (matches[3].rm_eo - matches[3].rm_so)); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */

                                                                        if (((quad_number = atoi (quad)) >= 0) && (quad_number < 256)) /* evaulate the numerical value of element in dotted quad */
                                                                        {
                                                                            (void) strcat (temp, "."); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */

                                                                            if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                            {
                                                                                quad = strchr (quad, '\0'); /* reference the second element in the dotted quad */
                                                                                (void) strncat (temp, cptr + matches[4].rm_so, (unsigned int) (matches[4].rm_eo - matches[4].rm_so)); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */

                                                                                if (((quad_number = atoi (quad)) >= 0) && (quad_number <= 255)) /* evaulate the numerical value of element in dotted quad */
                                                                                {
                                                                                    (void) strcat (temp, "."); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */

                                                                                    if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                                    {
                                                                                        quad = strchr (quad, '\0'); /* reference the third element in the dotted quad */
                                                                                        (void) strncat (temp, cptr + matches[5].rm_so, (unsigned int) (matches[5].rm_eo - matches[5].rm_so)); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */

                                                                                        if (((quad_number = atoi (quad)) >= 0) && (quad_number <= 255)) /* evaulate the numerical value of element in dotted quad */
                                                                                        {
                                                                                            (void) strcat (temp, "."); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */

                                                                                            if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                                            {
                                                                                                quad = strchr (quad, '\0'); /* reference the fourth element in the dotted quad */
                                                                                                (void) strncat (temp, cptr + matches[6].rm_so, (unsigned int) (matches[6].rm_eo - matches[6].rm_so)); /* re-assemble a copy of the match which is a dotted quad IP address in the "received: " record from the input file */

                                                                                                if (((quad_number = atoi (quad)) >= 0) && (quad_number <= 255)) /* evaulate the numerical value of element in dotted quad */
                                                                                                {

                                                                                                    if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                                                    {
                                                                                                        retval = ECONTINUE; /* re-assume error in e-mail format or structure */
                                                                                                        match_count --; /* assumption that a match would be found in the database was incorrect, decrement the count of the number of matches-the number of IP addresses found in the database */
                                                                                                    }

                                                                                                }

                                                                                            }

                                                                                        }

                                                                                    }

                                                                                }

                                                                            }

                                                                        }

                                                                        if (retval != MATCH) /* match not found in the database? */
                                                                        {
                                                                            non_match = NOMATCH; /* set the domain found with invalid IP address, 0 = no, 1 = yes */
                                                                        }

                                                                        if (print_flag == 1) /* verbose print flag, 0 = no print, 1 = print if IP address is in database, 2 = print if it isn't, equal to 1? */
                                                                        {

                                                                            if (retval == MATCH) /* match found in the database? */
                                                                            {
                                                                                (void) printf ("%s\t", temp); /* yes, print the match */
                                                                                temp[0] = '\0'; /* null the temporary buffer space */
                                                                                (void) strncat (temp, cptr + matches[2].rm_so, (unsigned int) (matches[2].rm_eo - matches[2].rm_so)); /* copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                                                                (void) printf ("%s\n", temp); /* yes, print the IP address */
                                                                            }

                                                                            retval = ECONTINUE; /* continue, re-assume error in e-mail format or structure */
                                                                        }

                                                                        if (print_flag == 2) /* verbose print flag, 0 = no print, 1 = print if IP address is in database, 2 = print if it isn't, equal to 2? */
                                                                        {

                                                                            if (retval != MATCH) /* match not found in the database? */
                                                                            {
                                                                                temp[0] = '\0'; /* null the temporary buffer space */
                                                                                (void) strncat (temp, cptr + matches[2].rm_so, (unsigned int) (matches[2].rm_eo - matches[2].rm_so)); /* copy of the match which is a dotted quad IP address in the "received: " record from the input file */
                                                                                (void) printf ("%s\n", temp); /* print the IP address */
                                                                            }

                                                                            retval = ECONTINUE; /* continue, re-assume error in e-mail format or structure */
                                                                        }

                                                                        cptr = &cptr[matches[0].rm_eo]; /* reference the next character after the entire match of the dotted quad IP address in the record from the input file to continue the search for dotted quad notation IP addresses, one after the other */
                                                                    }

                                                                }

                                                            }

                                                        }

                                                    }

                                                    if (argc > optind) /* using stdin as input? */
                                                    {

                                                        if (fclose (infile) == EOF) /* no, close the input file */
                                                        {
                                                            retval = ECLOSE; /* error closing file */
                                                        }

                                                    }

                                                }

                                            }
                                            while ((argc > ++ optind) && ((retval == MATCH) || (retval == NOMATCH))); /* while there are any remaining input file(s), if any, listed on the command line */

                                            regfree (&reg); /* free the dotted quad regular expression compile */
                                        }

                                    }

                                    if (munmap ((void *) mapbegin, filesize) != 0) /* unmap the file */
                                    {
                                        retval = EUNMAP; /* error unmapping file */
                                    }

                                }

                                if (close (fd) != 0) /* close the file */
                                {
                                    retval = ECLOSE; /* error closing file */
                                }

                            }

                        }

                    }

                }

            }

        }

    }

    if ((retval == MATCH) || (retval == NOMATCH)) /* if any nonmatches, but no errors, set the return value to the domain found with invalid IP address, 0 = no, 1 = yes */
    {
        retval = non_match;
    }

    print_message (retval); /* print any error messages */

    if ((error_return_change != -1) && (retval >= EARGS)) /* change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH, and error? */
    {
        retval = error_return_change; /* yes, set the error value to MATCH or NOMATCH */
    }

    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit-match found.

        2) Else if retval is unity, print nothing-normal/successful
           program exit-match not found.

        3) Else if retval is two, print help.

        4) Else retval is not zero, unity, or two, it is an error
           code, print the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, match found, print nothing */
        case 1: /* program ended without errors, match not found, print nothing */

            break;

        case 2: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Recursively binary search a memory area.

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap);

I) Binary search a memory area, referenced by the begin and end
variables, for string. The variable begin references the beginning of
the area to be searched; the variable end references the end of the
area to be searched, and will always reference the '\n' terminating
character of the last record of the memory area, (which is a
requirment); the variable string is the null terminated string to
search for, and has a length represented by the variable string_len.

II) Requirements:

    A) The memory area size must be greater than zero, (this is a
    requirement for mmap(1), also.)

    B) The last character of the memory area must be a '\n' text file
    record terminator, e.g., the last record, (which, minimally, may
    be the only record,) must be terminated.

    C) The records in the memory area must be in lexical sequence,
    (consistent with strcmp() and strncmp().)

III) The general strategy is:

    A) For each iteration in the recursion:

        1) Divide the memory area exactly in two.

        2) Work from this middle, in both directions, to find the end
        of the preceding record, (or beginning of file, whichever
        comes first) and, the end of the current record:

            a) The current record contains the first instance of the
            the terminating character at, or past, the middle.

            b) The current record's beginning is the next character
            after the preceding records's terminating character, or
            the beginning of the memory area.

        3) Compare this record with the string for a match.

        4) If not a match, decide which direction to move for the next
        iteration, and call this function, again.

        5) The recursive binary search terminates when a match is
        found, or when there is no more memory area to search.

    B) Return a reference to the matched area in memory, or null if no
    match.

Returns a reference to the matched string, null if no match.

*/

#ifdef __STDC__

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap)

#else

static char *binarysearch (begin, end, string, string_len, BOFmap, EOFmap)
char *begin;
char *end;
char *string;
size_t string_len;
char *BOFmap;
char *EOFmap;

#endif

{
    int retval; /* return for string comparisons */

    char *middle, /* reference to the middle between begin and end */
         *BOFmiddle_record, /* reference to the beginning of the record in the middle */
         *EOFmiddle_record, /* reference to the end of the record in the middle */
         *BOFnext_record, /* beginning of next record */
         *EOFprevious_record, /* end of previous record */
         *match = (char *) 0; /* reference to the beginning of the record in the middle if a match is found; null if no match */

    size_t middle_record_len; /* length record in the middle */

/* #define QA */ /* comment out to disable QA/Debuging */

#ifdef QA

    /*

    The Unix dictionary file, "sort -u /usr/local/lib/dict/websters >
    testfile", is a good QA test candidate.

    Also "thisprogram testfile < testfile > out" without QA #defined,
    and "cmp testfile out" is a good QA test candidate, (to verify
    that all words are successfully found,) where thisprogram is the
    name of the program containing this function.

    Test for the pathological cases, too; duplicate records, null
    records, single null records, etc.

    */

    char QAbuf[BUFSIZ], /* QA print buffer */
         *QAbegin, /* beginning of record to be printed */
         *QAend; /* end of record to be printed */

    size_t QAreclen; /* QA record length in print buffer */

#endif

    middle = BOFmiddle_record = EOFmiddle_record = begin + ((end - begin) / 2); /* reference the middle between begin and end */

    while ((BOFmap < BOFmiddle_record) && (*(BOFmiddle_record - 1) != '\n')) /* step backwards to the beginning of the record in the middle */
    {
        BOFmiddle_record --;
    }

    while ((EOFmiddle_record < EOFmap) && (*EOFmiddle_record != '\n')) /* step forwards to the end of the record in the middle */
    {
        EOFmiddle_record ++;
    }

    begin = (BOFmiddle_record < begin) ? BOFmiddle_record : begin; /* beginning of record in the middle outside of begin? */
    end = (end < EOFmiddle_record) ? EOFmiddle_record : end; /* end of record in the middle outside of end? */
    BOFnext_record = (EOFmiddle_record < EOFmap) ? EOFmiddle_record + 1 : EOFmap; /* save the beginning of next record */
    EOFprevious_record = (BOFmap < BOFmiddle_record) ? BOFmiddle_record - 1 : BOFmap; /* end of previous record */
    middle_record_len = (size_t) (EOFmiddle_record - BOFmiddle_record); /* save the length of the record in the middle, minus the terminating '\n' */
    retval = strncmp (string, BOFmiddle_record, string_len < middle_record_len ? string_len : middle_record_len); /* compare the record in the middle with the search string */

#ifdef QA

    (void) printf ("%s, ", string); /* print the string being searched for */

    QAreclen = strchr (BOFmiddle_record, '\n') - BOFmiddle_record; /* save the length of the record being searched */
    (void) strncpy (QAbuf, BOFmiddle_record, QAreclen); /* copy the record being searched to the print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the record being searched in the print buffer */
    (void) printf ("%s, ", QAbuf); /* print the copy of the record being searched */

    QAreclen = strchr (begin, '\n') - begin; /* length of first record in begin/end */
    (void) strncpy (QAbuf, begin, QAreclen); /* copy first record in begin/end to QA print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the first record in begin/end */
    (void) printf ("(%s - ", QAbuf); /* print the copy of the first record in begin/end */

    QAend = QAbegin = end; /* end of last record in begin/end */

    while ((BOFmap < QAbegin) && (*(QAbegin - 1) != '\n')) /* step backwards to the beginning of the last record in begin/end */
    {
        QAbegin --;
    }

    (void) strncpy (QAbuf, QAbegin, (size_t) (QAend - QAbegin)); /* copy last record in begin/end to QA print buffer */
    QAbuf[QAend - QAbegin] = '\0'; /* terminate the copy of the last record in begin/end */
    (void) printf ("%s)\n", QAbuf); /* print the copy of the last record in begin/end */

#endif

    if (retval < 0) /* string < record? */
    {

        if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
        {
            match = binarysearch (begin, EOFprevious_record, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
        }

    }

    else if (retval > 0) /* string > record? */
    {

        if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
        {
            match = binarysearch (BOFnext_record, end, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
        }

    }

    else /* match, string = record? */
    {

        if (string_len < middle_record_len) /* partial key match; string shorter than record? */
        {

            if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
            {
                match = binarysearch (begin, EOFprevious_record, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
            }

        }

        else if (string_len > middle_record_len) /* partial key match; string longer than record? */
        {

            if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
            {
                match = binarysearch (BOFnext_record, end, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
            }

        }

        else /* exact match */
        {
            match = BOFmiddle_record; /* return a reference to this record */
        }

    }

    return (match); /* return a reference to the match found between begin and end; null if no match */
}
