/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 2001-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

receivedMSGIDdb.c is for validating the "Message-ID: " records in the
headers of e-mail files against domain names in a database file.

The program requires a Posix compatible regex(3) library to parse the
e-mail addresses, and mmap(2) to map the database file of e-mail
addresses into the Unix VM system. The e-mail addresses from the
e-mail header are validated against the database file using a binary
search. The database file name is a required command line argument.

The database is a standard Unix text file, one e-mail address per
line, in lexical order, constructed with "sort -u infile > outfile",
or equivalent.

The database mechanism is conservative with machine resources,
requiring about 12.5 micro-seconds of machine time to lookup a word in
the Unix system dictionary, (2.5 MB, quarter of a million words,
single 466 MHz., Pentium, lightly loaded, Linux 2.2, time(1) command
to lookup every word in the dictionary, divided by the number of
words.) Conceptually, the database mechanism is implemented similar to
the the technique used in the look(1) command, but requires exact
matches, as opposed to partial key matches.

The input e-mail file name(s) may be supplied as additional optional
command line arguments, or redirected to the program via stdin for
compatibility with procmail(1), and other e-mail scripting agents.

A suitable procmail(1) recipe example might be:

    :0 wfh
    * ? receivedMSGIDdb address.db
    | formail -A "X-Notice: Message in address.db database"

The -f from_address option can be used to validate that the
"Message-ID: " and the "From: " e-mail address, (or any other e-mail
address,) are from the same machine/domain. The -f option can be used
multiple times to validate any number e-mail addresses.

The program contains less than 300 lines of declarations and
statements, all of which are documented with in line comments.

The program has been compiled and tested on SunOS, Solaris, and Linux,
and may work on other brands of Unix.

The program returns 0 if no error and a match was found in the
database file for the domain name in any "Message-ID: " header
records, 1 if no error and no match found; else returns a unique error
code greater than 1 representing the error encountered-which will,
also, print an error diagnostic to stderr.

However, if one or more e-mail addresses are specified on the command
line with the -f option, then if all e-mail addresses have the same
domain name as found in the database file from the "Message-ID: "
header record, then the program returns 1. If any e-mail address does
not have the same domain name as found in the database file from the
"Message-ID: " record, then the program returns 0. If the domain name
in the "Message-ID: " record is not found in the database, the program
returns 1, as before. The return values for the -f option are for
compatibility with procmail scripts.

The -r option is useful for controlling the return value under error
conditions-for example, the program return can be preempted if the
database file can not be opened, (or read,) with a return value of
match, or no match, depending on environmental requirements.

John Conover
January 17, 2005
john@email.johncon.com

$Revision: 0.0 $
$Date: 2005/01/17 19:36:00 $
$Id: receivedMSGIDdb.c,v 0.0 2005/01/17 19:36:00 john Exp $
$Log: receivedMSGIDdb.c,v $
Revision 0.0  2005/01/17 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* for getopt() */
#include <ctype.h> /* for tolower(), isspace() */
#include <regex.h> /* for regular expression; used for e-mail address detection */
#include <sys/types.h> /* for open() */
#include <sys/stat.h> /* for open() */
#include <fcntl.h> /* for open() */
#include <unistd.h> /* for mmap() */
#include <sys/mman.h> /* for mmap() */

static char rcsid[] = "$Id: receivedMSGIDdb.c,v 0.0 2005/01/17 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 2001-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define REGEXMATCH "@([-0-9a-z._+=]+)" /* the regular expression to detect an e-mail address in the "Message-ID: " record of an e-mail header */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Validate the domain name in \"Message-ID: \" records in the header of e-mail\n",
    "against domain names in a database file:\n",
    "Usage: receivedMSGIDdb [-c] [-f from_address] [-r m|n] [-v]\n",
    "                       dbfilename [filename(s)]\n",
    "    -c, print all matching e-mail addresses, not just the first\n",
    "    -f from_address, validate the e-mail address, from_address\n",
    "    -r m|n, on file error, exit return = match for m, no match for n\n",
    "    -v, print the program's version information\n",
    "    dbfilename, database file name\n",
    "    filename(s), e-mail file name(s), (defaults to stdin)\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error, match found\n",
    "No error, match not found\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error in input file record length\n",
    "Error compiling regular expression for e-mail address notation\n",
    "Error in e-mail format or structure\n",
    "Error in status of database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error in database file\n"
};

#define MATCH 0 /* error values, one for each index in the error message array */
#define NOMATCH 1
#define EARGS 2
#define EOPEN 3
#define ECLOSE 4
#define ERECL 5
#define ECREGEX 6
#define ECONTINUE 7
#define ESTAT 8
#define EMAP 9
#define EUNMAP 10
#define EDB 11

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap);

#else

static void print_message (); /* print any error messages */
static char *binarysearch ();

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         temp[BUFLEN], /* temporary buffer space */
         *cptr, /* temporary character reference */
         *mptr, /* reference to a match in domain name */
         *begin, /* begining of memory mapped file */
         *end, /* end of memory mapped file */
         *from_address[BUFLEN]; /* reference to e-mail address to be validated, if any-BUFLEN many since BUFLEN is the buffer length of stdin for overflow protection */

    int retval = NOMATCH, /* return value, assume no error */
        fd = -1, /* input file descriptor */
        c, /* command line switch */
        i, /* general counter index */
        j, /* general integer storage */
        match_count = 0, /* count of the number of matches-the number of domain names found in the database */
        email_match_count = 0, /* count of the number of matches in e-mail addresses, if any */
        old_email_match_count = 0, /* previous count of the number of matches in e-mail addresses, if any */
        msgid_active = 0, /* message-id active switch, 0 = no, 1 = message-id record, and/or its continuation(s) */
        continue_flag = 0, /* continue to print all matching e-mail addresses, not just the first, 0 = no, 1 = yes */
        from_address_ctr = 0, /* count of e-mail address to be validated, if any */
        error_return_change = -1; /* change exit return on error, -1 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */

    size_t num_match = 8; /* REGEXMATCH contains 8 substrings */

    FILE *infile = stdin; /* reference to input file; defaults to stdin */

    size_t filesize; /* size of file */

    struct stat fileinfo; /* structure for file status */

    caddr_t mapbegin; /* reference to beginning of mmap'ed file */

    regex_t reg; /* e-mail address regular expression compile */
    regmatch_t matches[8]; /* REGEXMATCH contains 8 substrings; matches[] contains their offsets */

    while ((c = getopt (argc, argv, "cf:r:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'c': /* request for continue to print all matching domain names, not just the first? */

                continue_flag = 1; /* set the continue to print all matching domain names, not just the first, 0 = no, 1 = yes */
                break;

            case 'f': /* request for reference to e-mail address to be validated, if any? */

                if (from_address_ctr < BUFLEN) /* too many references to e-mail address to be validated? */
                {

                    if ((cptr = strchr (optarg, (int) '@')) == (char *) 0) /* e-mail address to be validated have an '@' character? */
                    {
                        cptr = optarg; /* no, assume its a domain name */
                    }

                    else
                    {
                        cptr ++; /* yes, step over that '@' character */
                    }

                    from_address[from_address_ctr] = cptr; /* save the reference to e-mail address' sub-domain */
                    from_address_ctr ++; /* increment the count of e-mail address to be validated, if any */
                }

                break;

            case 'r': /* request for change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH? */

                switch (*optarg) /* first character of argument object, only */
                {

                    case 'm':

                        error_return_change = MATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    case 'n':

                        error_return_change = NOMATCH; /* set the change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH */
                        break;

                    default:

                        optind = argc; /* force argument error */
                        retval = EARGS; /* assume not enough arguments */
                        break;
                }

                break;

            case 'v': /* request for program's version information? */

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOMATCH) /* any errors? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (argc > optind) /* enough arguments? */
        {
            retval = ESTAT; /* assume error in status of file */

            if (stat (argv[optind], &fileinfo) == 0) /* get the status of the file */
            {
                retval = EDB;  /* assume error in database file */

                if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of file */
                {
                    retval = EOPEN; /* assume error opening file */

                    if ((fd = open (argv[optind], O_RDONLY)) != -1) /* open the file */
                    {
                        retval = EMAP; /* assume error mapping file */

                        if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, fd, (off_t) 0)) != (caddr_t) 0) /* map the file */
                        {
                            retval = EDB;  /* assume error in database file */

                            if (mapbegin[filesize - 1] == '\n') /* verify that the last record of the memory mapped file has a terminating '\n' character as required for binarysearch() */
                            {
                                retval = NOMATCH; /* yes, return value, assume no error */
                                optind ++; /* file argument handled */
                                begin = (char *) mapbegin; /* reference the begining of memory mapped file */
                                end = &mapbegin[filesize - 1]; /* reference the end of memory mapped file */
                                retval = ECREGEX; /* assume error compiling regular expression for an email address */

                                if (regcomp (&reg, REGEXMATCH, REG_EXTENDED | REG_ICASE) == 0) /* compile the domain name regular expression */
                                {

                                    do /* while there are input file(s), if any, listed on the command line */
                                    {
                                        retval = EOPEN; /* assume error opening file */

                                        if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* open the input file; if none specified, default to stdin */
                                        {
                                            retval = ECONTINUE; /* assume error in e-mail format or structure */

                                            while ((fgets (buffer, BUFLEN, infile) != (char *) 0) && (retval == ECONTINUE)) /* read the records from the input file until EOF, or an error in processing, or a null line, (e.g., start of message body) */
                                            {
                                                retval = ERECL; /* assume error in record length */

                                                if (strchr (buffer, '\n') != (char *) 0) /* a record with no carriage return was too long for the fgets () buffer, and an overflow condition exists; line contain a carriage return? */
                                                {
                                                    retval = NOMATCH; /* assume no error */

                                                    if (buffer[0] != '\n') /* a record with no characters is first line of message body, quit */
                                                    {
                                                        retval = ECONTINUE; /* re-assume error in e-mail format or structure */
                                                        cptr = buffer; /* reference the first character in the record from the input file */

                                                        while (*cptr != '\n') /* lowercase the record from the input file; for each character in the record from the input file */
                                                        {
                                                            *cptr = (char) tolower (*cptr); /* convert the character to lower case */
                                                            cptr ++; /* next character */
                                                        }

                                                        if (strncmp ("message-id: ", buffer, 12) == 0) /* first line of message-id record? */
                                                        {
                                                            msgid_active = 1; /* set the message-id active switch, 0 = no, 1 = message-id record, and/or its continuation(s) */
                                                        }

                                                        else if (! isspace (buffer[0])) /* a record from the input file in the header having white space as the first character is a continuation record; if so leave the message-id active switch in its current state, else, reset it */
                                                        {
                                                            msgid_active = 0; /* reset the message-id active switch, 0 = no, 1 = message-id record, and/or its continuation(s) */
                                                        }

                                                        if (msgid_active == 1) /* message-id active switch, 0 = no, 1 = message-id record, and/or its continuation(s), set? */
                                                        {
                                                            cptr = buffer; /* reference the first character in the record from the input file */

                                                            while ((regexec (&reg, cptr, num_match, matches, 0) == 0) && (retval == ECONTINUE)) /* search the "message-id: " record, or its continuation(s), from the input file for domain names, one after the other */
                                                            {

                                                                /*

                                                                found a domain name in the record from the input file:

                                                                matches[0] contains the offsets of the entire match, e.g., "@email.johncon.com"
                                                                matches[1] contains the offsets of the for sub-domain, e.g., "email.johncon.com"

                                                                */

                                                                retval = MATCH; /* assume a match will be found in the database */
                                                                match_count ++; /* assume a match will be found in the database, increment the count of the number of matches-the number of domain names found in the database */

                                                                temp[0] = '\0'; /* null the temporary buffer space */
                                                                (void) strncat (temp, cptr + matches[1].rm_so, (unsigned int) (matches[1].rm_eo - matches[1].rm_so)); /* save the sub-domain name from the "message-id: " record from the input file */

                                                                if (binarysearch (begin, end, temp, strlen (temp), begin, end) == (char *) 0)
                                                                {
                                                                    retval = ECONTINUE; /* re-assume error in e-mail format or structure */
                                                                    match_count --; /* assumption that a match would be found in the database was incorrect, decrement the count of the number of matches-the number of domain names found in the database */
                                                                }

                                                                if (retval == MATCH) /* match found in the database? */
                                                                {
                                                                    old_email_match_count = email_match_count; /* save the previous count of the number of matches in e-mail addresses, if any */

                                                                    for (i = 0; i < from_address_ctr; i ++) /* for each e-mail address to be validated, if any */
                                                                    {
                                                                        j = (int) (strlen (temp) - strlen (from_address[i])); /* save the difference in lengths of the e-mail address to be validated and the sub-domain name from the "message-id: " record from the input file */

                                                                        if (j < 0) /* length of the e-mail address to be validated > than the sub-domain name */
                                                                        {

                                                                            if ((mptr = strstr (from_address[i], temp)) == (char *) 0) /* match not found? */
                                                                            {
                                                                                break;
                                                                            }

                                                                            else if (*(mptr - 1) != '.') /* match of previous character at a sub-domain marker? */
                                                                            {
                                                                                break;
                                                                            }

                                                                            else
                                                                            {
                                                                                email_match_count ++; /* match, increment the count of the number of matches in e-mail addresses, if any */
                                                                            }

                                                                        }

                                                                        else if (j > 0) /* length of the e-mail address to be validated < than the sub-domain name */
                                                                        {

                                                                            if ((mptr = strstr (temp, from_address[i])) == (char *) 0)  /* match not found? */
                                                                            {
                                                                                break;
                                                                            }

                                                                            else if (*(mptr - 1) != '.') /* match of previous character at a sub-domain marker? */
                                                                            {
                                                                                break;
                                                                            }

                                                                            else
                                                                            {
                                                                                email_match_count ++; /* match, increment the count of the number of matches in e-mail addresses, if any */
                                                                            }

                                                                        }

                                                                        else /* if j = 0, length of the e-mail address to be validated = than the sub-domain name */
                                                                        {

                                                                            if (strcmp (temp, from_address[i]) != 0)  /* match not found? */
                                                                            {
                                                                                break;
                                                                            }

                                                                            else
                                                                            {
                                                                                email_match_count ++; /* match, increment the count of the number of matches in e-mail addresses, if any */
                                                                            }

                                                                        }

                                                                    }

                                                                    if (retval == MATCH) /* if still match found in the database? */
                                                                    {

                                                                        if (email_match_count > old_email_match_count) /* count of the number of matches in e-mail addresses, if any, greater than the previous count of the number of matches in e-mail addresses means a match was found */
                                                                        {
                                                                            (void) printf ("%s\n", temp); /* print the match */
                                                                        }

                                                                    }

                                                                }

                                                                if (continue_flag != 0) /* continue to print all matching domain names, not just the first, 0 = no, 1 = yes, set? */
                                                                {
                                                                    retval = ECONTINUE; /* continue, re-assume error in e-mail format or structure */
                                                                }

                                                                cptr = &cptr[matches[0].rm_eo]; /* reference the next character after the entire match of domain names in the record from the input file to continue the search domain names, one after the other */
                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                            if (argc > optind) /* using stdin as input? */
                                            {

                                                if (fclose (infile) == EOF) /* no, close the input file */
                                                {
                                                    retval = ECLOSE; /* error closing file */
                                                }

                                            }

                                        }

                                    }
                                    while ((argc > ++ optind) && ((retval == MATCH) || (retval == NOMATCH))); /* while there are any remaining input file(s), if any, listed on the command line */

                                    regfree (&reg); /* free the domain name regular expression compile */
                                }

                            }

                            if (munmap ((void *) mapbegin, filesize) != 0) /* unmap the file */
                            {
                                retval = EUNMAP; /* error unmapping file */
                            }

                        }

                        if (close (fd) != 0) /* close the file */
                        {
                            retval = ECLOSE; /* error closing file */
                        }

                    }

                }

            }

        }

    }

    retval = (((retval == MATCH) || (retval == NOMATCH)) && (match_count > 0)) ? MATCH : retval; /* if any matches, but no errors, set the return value to MATCH */
    print_message (retval); /* print any error messages */

    if ((error_return_change != -1) && (retval >= EARGS)) /* change exit return on error, 0 = no, error changed to match = MATCH, error changed to nomatch = NOMATCH, and error? */
    {
        retval = error_return_change; /* yes, set the error value to MATCH or NOMATCH */
    }

    else if (from_address_ctr > 0) /* count of e-mail address to be validated, if any, greater than zero? */
    {

        if (match_count > 0) /* any count of the number of matches-the number of domain names found in the database? */
        {

            if (match_count * from_address_ctr == email_match_count) /* if all email addresses were found with the same domain name as the message id record */
            {
                retval = NOMATCH; /* return NOMATCH */
            }

            else
            {
                retval = MATCH; /* return MATCH */
            }

        }

    }

    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit-match found.

        2) Else if retval is unity, print nothing-normal/successful
           program exit-match not found.

        3) Else if retval is two, print help.

        4) Else retval is not zero, unity, or two, it is an error
           code, print the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, match found, print nothing */
        case 1: /* program ended without errors, match not found, print nothing */

            break;

        case 2: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Recursively binary search a memory area.

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap);

I) Binary search a memory area, referenced by the begin and end
variables, for string. The variable begin references the beginning of
the area to be searched; the variable end references the end of the
area to be searched, and will always reference the '\n' terminating
character of the last record of the memory area, (which is a
requirment); the variable string is the null terminated string to
search for, and has a length represented by the variable string_len.

II) Requirements:

    A) The memory area size must be greater than zero, (this is a
    requirement for mmap(1), also.)

    B) The last character of the memory area must be a '\n' text file
    record terminator, e.g., the last record, (which, minimally, may
    be the only record,) must be terminated.

    C) The records in the memory area must be in lexical sequence,
    (consistent with strcmp() and strncmp().)

III) The general strategy is:

    A) For each iteration in the recursion:

        1) Divide the memory area exactly in two.

        2) Work from this middle, in both directions, to find the end
        of the preceding record, (or beginning of file, whichever
        comes first) and, the end of the current record:

            a) The current record contains the first instance of the
            the terminating character at, or past, the middle.

            b) The current record's beginning is the next character
            after the preceding records's terminating character, or
            the beginning of the memory area.

        3) Compare this record with the string for a match.

        4) If not a match, decide which direction to move for the next
        iteration, and call this function, again.

        5) The recursive binary search terminates when a match is
        found, or when there is no more memory area to search.

    B) Return a reference to the matched area in memory, or null if no
    match.

Returns a reference to the matched string, null if no match.

*/

#ifdef __STDC__

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, char *BOFmap, char *EOFmap)

#else

static char *binarysearch (begin, end, string, string_len, BOFmap, EOFmap)
char *begin;
char *end;
char *string;
size_t string_len;
char *BOFmap;
char *EOFmap;

#endif

{
    int retval; /* return for string comparisons */

    char *middle, /* reference to the middle between begin and end */
         *BOFmiddle_record, /* reference to the beginning of the record in the middle */
         *EOFmiddle_record, /* reference to the end of the record in the middle */
         *BOFnext_record, /* beginning of next record */
         *EOFprevious_record, /* end of previous record */
         *match = (char *) 0; /* reference to the beginning of the record in the middle if a match is found; null if no match */

    size_t middle_record_len; /* length record in the middle */

/* #define QA */ /* comment out to disable QA/Debuging */

#ifdef QA

    /*

    The Unix dictionary file, "sort -u /usr/local/lib/dict/websters >
    testfile", is a good QA test candidate.

    Also "thisprogram testfile < testfile > out" without QA #defined,
    and "cmp testfile out" is a good QA test candidate, (to verify
    that all words are successfully found,) where thisprogram is the
    name of the program containing this function.

    Test for the pathological cases, too; duplicate records, null
    records, single null records, etc.

    */

    char QAbuf[BUFSIZ], /* QA print buffer */
         *QAbegin, /* beginning of record to be printed */
         *QAend; /* end of record to be printed */

    size_t QAreclen; /* QA record length in print buffer */

#endif

    middle = BOFmiddle_record = EOFmiddle_record = begin + ((end - begin) / 2); /* reference the middle between begin and end */

    while ((BOFmap < BOFmiddle_record) && (*(BOFmiddle_record - 1) != '\n')) /* step backwards to the beginning of the record in the middle */
    {
        BOFmiddle_record --;
    }

    while ((EOFmiddle_record < EOFmap) && (*EOFmiddle_record != '\n')) /* step forwards to the end of the record in the middle */
    {
        EOFmiddle_record ++;
    }

    begin = (BOFmiddle_record < begin) ? BOFmiddle_record : begin; /* beginning of record in the middle outside of begin? */
    end = (end < EOFmiddle_record) ? EOFmiddle_record : end; /* end of record in the middle outside of end? */
    BOFnext_record = (EOFmiddle_record < EOFmap) ? EOFmiddle_record + 1 : EOFmap; /* save the beginning of next record */
    EOFprevious_record = (BOFmap < BOFmiddle_record) ? BOFmiddle_record - 1 : BOFmap; /* end of previous record */
    middle_record_len = (size_t) (EOFmiddle_record - BOFmiddle_record); /* save the length of the record in the middle, minus the terminating '\n' */
    retval = strncmp (string, BOFmiddle_record, string_len < middle_record_len ? string_len : middle_record_len); /* compare the record in the middle with the search string */

#ifdef QA

    (void) printf ("%s, ", string); /* print the string being searched for */

    QAreclen = strchr (BOFmiddle_record, '\n') - BOFmiddle_record; /* save the length of the record being searched */
    (void) strncpy (QAbuf, BOFmiddle_record, QAreclen); /* copy the record being searched to the print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the record being searched in the print buffer */
    (void) printf ("%s, ", QAbuf); /* print the copy of the record being searched */

    QAreclen = strchr (begin, '\n') - begin; /* length of first record in begin/end */
    (void) strncpy (QAbuf, begin, QAreclen); /* copy first record in begin/end to QA print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the first record in begin/end */
    (void) printf ("(%s - ", QAbuf); /* print the copy of the first record in begin/end */

    QAend = QAbegin = end; /* end of last record in begin/end */

    while ((BOFmap < QAbegin) && (*(QAbegin - 1) != '\n')) /* step backwards to the beginning of the last record in begin/end */
    {
        QAbegin --;
    }

    (void) strncpy (QAbuf, QAbegin, (size_t) (QAend - QAbegin)); /* copy last record in begin/end to QA print buffer */
    QAbuf[QAend - QAbegin] = '\0'; /* terminate the copy of the last record in begin/end */
    (void) printf ("%s)\n", QAbuf); /* print the copy of the last record in begin/end */

#endif

    if (retval < 0) /* string < record? */
    {

        if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
        {
            match = binarysearch (begin, EOFprevious_record, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
        }

    }

    else if (retval > 0) /* string > record? */
    {

        if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
        {
            match = binarysearch (BOFnext_record, end, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
        }

    }

    else /* match, string = record? */
    {

        if (string_len < middle_record_len) /* partial key match; string shorter than record? */
        {

            if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
            {
                match = binarysearch (begin, EOFprevious_record, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
            }

        }

        else if (string_len > middle_record_len) /* partial key match; string longer than record? */
        {

            if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
            {
                match = binarysearch (BOFnext_record, end, string, string_len, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
            }

        }

        else /* exact match */
        {
            match = BOFmiddle_record; /* return a reference to this record */
        }

    }

    return (match); /* return a reference to the match found between begin and end; null if no match */
}
